/*
 * TempBufferRepository.h
 *
 *  Created on: Sep 19, 2011
 *      Author: troth22s
 */

#ifndef TEMPBUFFERREPOSITORY_H_
#define TEMPBUFFERREPOSITORY_H_


#include <Data/DataTypes.h>
#include <Utils/Common.h>
#include <Data/EnumComputeEngine.h>
#include <list>
#include <vector>
#include <map>
#include <float.h>
#include <math.h>
#include <assert.h>


namespace grip {





/**
 * @class TempBufferWrapper
 * @brief This class is responsible for managing information about whether a temporary buffer is currently occupied or not.
 */

#undef CLASSNAME
#define CLASSNAME TempBufferWrapper

class CLASSNAME {

private:
	void* 				__m_data;
	bool 				__m_occupied;
	unsigned int 		__m_size;
	enum_computeEngine __m_bufferType;

	unsigned int __m_numReaders;
	
public:

	// @brief a wrapper has to be initialized with its associated data, width, height and occupation status
	CLASSNAME(void* p_data, unsigned int p_size, bool p_occupied, enum_computeEngine p_bufferType, unsigned int p_numReaders = 0) {
		DEBUG_MSG_FUNCTIONCALL("TempBufferWrapper(" << p_data << ")");

		__m_data = p_data;
		__m_occupied = p_occupied;

		__m_size = p_size;

		__m_bufferType = p_bufferType;
		
		__m_numReaders = p_numReaders;
	}

	// @brief take care for deleting the referenced data manually, this is not done here as we do not even know what data it is!
	~CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("~TempBufferWrapper()");
	}

	/** buffer can not be requested if it is already held for writing or reading elsewhere
	 *  multiple readers are usually present when one output connector is connected to multiple
	 *  input connectors
	 */
	bool isOccupied() {
		DEBUG_MSG_FUNCTIONCALL("isAvailableForWriting()");
		
		return (__m_occupied || __m_numReaders > 0);
	}

	void setOccupied() {
		DEBUG_MSG_FUNCTIONCALL("setOccupied()");

		assert(!isOccupied());
		__m_occupied = true;
	}

	void unsetOccupied() {
		DEBUG_MSG_FUNCTIONCALL("unsetOccupied()");

		assert(isOccupied());
		__m_occupied = false;
	}

	void* getData() {
		DEBUG_MSG_FUNCTIONCALL("getData()");

		return __m_data;
	}
	
	void addReader() {
	
		++__m_numReaders;
		
	}
	
	void removeReader() {
	
		assert(__m_numReaders > 0);
		
		--__m_numReaders;
		
	}
	
	unsigned int getNumReaders() {
	
		return __m_numReaders;
		
	}

	unsigned int getSize() {
		DEBUG_MSG_FUNCTIONCALL("getSize()");

		return __m_size;
	}

	enum_computeEngine getBufferType() {
		DEBUG_MSG_FUNCTIONCALL("getBufferType()");

		return __m_bufferType;
	}

};

#undef CLASSNAME
#define CLASSNAME TempBufferRepository

class CLASSNAME {

private:

	std::vector<TempBufferWrapper*> __m_bufferGrid;

	bool __m_isInitialized;

	grip::dt_uint32_t __m_computeEngineUsage;

public:


	// @brief maximum size and number of cells can bet set here
	CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("TempBufferRepository()");

		__m_isInitialized = false;
		__m_computeEngineUsage = 0;

		init();

	}


	
	/* @brief initialize the repository with the given maximum extent and cell size */
	void init() {
		DEBUG_MSG_FUNCTIONCALL("init()");

		assert(!__m_isInitialized);
		__m_computeEngineUsage = 0;

		__m_isInitialized = true;
	}

	// @brief delete ALL the stuff (just iterate over the vector and the contained lists and delete the buffers)
	~CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("~TempBufferRepository()");

		for (std::vector<TempBufferWrapper*>::iterator it = __m_bufferGrid.begin(); it != __m_bufferGrid.end(); ++it) {
			delete *it;
		}

	}

	/* @brief get a complete list of registered buffers
	 * @return complete list of registered buffers
	 * This is mainly necessary for the cleanup operation at the end of graph execution
	 */
	std::list<TempBufferWrapper*> getCompleteBufferList(enum_computeEngine p_bufferType) {
		DEBUG_MSG_FUNCTIONCALL("getCompleteBufferList()");

		std::list<TempBufferWrapper*> bufferList;

		for (std::vector<TempBufferWrapper*>::iterator it = __m_bufferGrid.begin(); it != __m_bufferGrid.end(); ++it) {
			
			if ((*it)->getBufferType() == p_bufferType) {
				bufferList.push_back(*it);
			}
			
		}

		return bufferList;
	}


	/* @brief get a buffer which is not necessarily the closest one
	 * @return the TempBufferWrapper for the first usable buffer found if there is one or NULL if there is no such buffer
	 */
	TempBufferWrapper* getUsableBuffer(unsigned int p_size, enum_computeEngine p_bufferType) {
		DEBUG_MSG_FUNCTIONCALL("getUsableBuffer(" << p_widthInBytes << ", " << p_heightInRows << ")");

		assert(p_bufferType != ENUM_COMPENG_ARBITRARY);

//		std::cout << "getUsableBuffer: " << p_widthInBytes << ", " << p_heightInRows << std::endl;

		assert(__m_isInitialized);


		// check for availability
		for (std::vector<TempBufferWrapper*>::iterator it = __m_bufferGrid.begin(); it != __m_bufferGrid.end(); ++it) {



			if ((*it)->getBufferType() == p_bufferType && !(*it)->isOccupied() && (*it)->getSize() >= p_size) {
				(*it)->setOccupied();

				return *it;
			}
			
			
		}

		return 0;
	}




	/* @brief store a buffer of the given type and size and set its occupation status to p_occupied */
	template<class FTYPE> TempBufferWrapper* addBuffer(FTYPE p_theBuffer, unsigned int p_size, bool p_occupied, enum_computeEngine p_bufferType) {
		DEBUG_MSG_FUNCTIONCALL("addBuffer(" << p_theBuffer << ", " << p_size << ")");

//		std::cout << "addBuffer: " << p_widthInBytes << ", " << p_heightInRows << std::endl;

		assert(p_bufferType != ENUM_COMPENG_ARBITRARY);
		assert(__m_isInitialized);
		// buffers can only be stored if their width and height do not exceed the maximum size predetermined for the grid
		
		//		std::cout << "adding at index " << index << " (" << cellX << ", " << cellY << ")" << std::endl;

		void* theData = reinterpret_cast<void*> (p_theBuffer);

		TempBufferWrapper* tbw = new TempBufferWrapper(theData, p_size, p_occupied, p_bufferType);

		__m_bufferGrid.push_back(tbw);

		__m_computeEngineUsage |= p_bufferType;

		return tbw;
	}

	/* @brief add an already wrapped buffer */
	void addTempBufferWrapper(TempBufferWrapper* p_theTempBufferWrapper) {
		DEBUG_MSG_FUNCTIONCALL("addTempBufferWrapper(" << p_theTempBufferWrapper << ")");

		assert(__m_isInitialized);

		__m_bufferGrid.push_back(p_theTempBufferWrapper);

		__m_computeEngineUsage |= p_theTempBufferWrapper->getBufferType();

	}


	bool usedComputeEngine(enum_computeEngine p_bufferType) {
		DEBUG_MSG_FUNCTIONCALL("usedComputeDevice(" << p_bufferType << ")");

		return __m_computeEngineUsage & p_bufferType;
	}

};

#undef CLASSNAME
#define CLASSNAME TempBufferRepositoryDirectory

/* @class RepositoryDirectory
 * @brief this class implements a directory of repositories, so that single processing nodes may register their own buffers easily (ideally by their own node name) */
class CLASSNAME {
private:

	std::map<std::string, TempBufferRepository*> __m_directory;

public:

	CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("RepositoryDirectory()");
	}


	~CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("~RepositoryDirectory()");
	}


	/* @brief get a list of all TempBufferWrappers* stored in all repositories */
	std::list<TempBufferWrapper*> getCompleteBufferList(enum_computeEngine p_bufferType) {
		DEBUG_MSG_FUNCTIONCALL("getCompleteBufferList()");

		std::list<TempBufferWrapper*> bufferList;

		for (std::map<std::string, TempBufferRepository*>::iterator it = __m_directory.begin(); it != __m_directory.end(); ++it) {
			std::list<TempBufferWrapper*> repositoryBufferList = it->second->getCompleteBufferList(p_bufferType);
			bufferList.merge(repositoryBufferList);
		}

		return bufferList;
	}


	/* @brief check if a repository of the given name exists */
	bool repositoryExists(const char* p_repName) {
		DEBUG_MSG_FUNCTIONCALL("repositoryExists(" << p_repName << ")");

		return __m_directory.find(p_repName) != __m_directory.end();
	}


	/* @brief get a repository by the given name */
	TempBufferRepository* getRepositoryByName(const char* p_repName) {
		DEBUG_MSG_FUNCTIONCALL("getRepositoryByName(" << p_repName << ")");

		assert(repositoryExists(p_repName));

		return __m_directory[p_repName];

	}

	/* @brief add a repository with the given size and number of cells by the given name */
	TempBufferRepository* addRepositoryByName(const char* p_repName) {
		DEBUG_MSG_FUNCTIONCALL("addRepositoryByName(" << p_repName << ")");

//		assert(!repositoryExists(p_repName));


		if (!repositoryExists(p_repName)) {

			TempBufferRepository* tbr = new TempBufferRepository();

			__m_directory[p_repName] = tbr;

			return tbr;

		}

	}


	bool usedComputeEngine(enum_computeEngine p_bufferType) {
		DEBUG_MSG_FUNCTIONCALL("usedComputeEngine(" << p_bufferType << ")");

		for (std::map<std::string, TempBufferRepository*>::iterator it = __m_directory.begin(); it != __m_directory.end(); ++it) {
			if (it->second->usedComputeEngine(p_bufferType)) {
				return true;
			}
		}

		return false;
	}

};



};

#endif /* TEMPBUFFERREPOSITORY_H_ */
