/*
 * UniversalContainer.h
 *
 *  Created on: 15.06.2010
 *      Author: troth
 */

#ifndef UNIVERSALCONTAINER_H_
#define UNIVERSALCONTAINER_H_

#include <Data/UniversalDataElement.h>
#include <Data/DataManager.h>
#include <Data/EnumComputeEngine.h>
#include <Data/ParameterSet.h>
#include <Utils/Debug.h>
#include <Utils/Common.h>



#include <map>
#include <list>
#include <string>
#include <iostream>

#include <GL/gl.h>

namespace grip {

	#undef	CLASSNAME
	#define	CLASSNAME	UniversalContainer

	using namespace std;

	typedef struct __ADDITIONAL_INFO {
		int32_t typeId;
		enum_computeEngine dirty;	// relies on enum_computeEngine
		void* devicePtr;
	} element_info;

	template<class CTYPE> class MapRegistry {
	public:
		typedef map<string, UniversalDataElement<CTYPE> > type;
	};

	/**
	 * @class UniversalContainer
	 *
	 * @brief This is a heterogeneous container for specified types
	 *
	 * Memory allocation/deallocation is handeled by std::tr1::shared_ptr
	 */
	class CLASSNAME {
	private:

		___GENERATE_MAP_REGISTRY

		map<string, element_info> __m_elementInfo;
		
		map<string, bool> __m_nodeActivation;

		dt_uint32_t __m_computeEngineUsage;

		/** @brief this serves for getting the right map for FTYPE and is specialized for all these types */
		template<class FTYPE> __inline__ typename MapRegistry<FTYPE>::type& getMap();

		map<string, list<ParameterSet*> > __m_parameterSetLinks;
		map<string, ParameterSet*> __m_parameterSets;

	public:

		/** @brief default constructor */
		INLINE CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("UniversalContainer()");

			__m_computeEngineUsage = 0;
		}

		/** @brief default destructor */
		INLINE ~CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("~UniversalContainer()");

		}

		void addParameterSet(const string& p_setName) {
			DEBUG_MSG_FUNCTIONCALL("addParameterSet(" << p_setName << ")");

			if (__m_parameterSets.find(p_setName) == __m_parameterSets.end()) {
				__m_parameterSets[p_setName] = new ParameterSet(p_setName);
			}
		}

		void addSetLinkToParameter(const string& p_parameterName, const string& p_setName) {
			DEBUG_MSG_FUNCTIONCALL("addSetLinkToParameter(" << p_parameterName << ", " << p_setName << ")");

			std::cout << "adding " << p_parameterName << " to " << p_setName << " ###" << std::endl;

			__m_parameterSetLinks[p_parameterName].push_back(__m_parameterSets[p_setName]);
		}


		void setParameterChanged(const string& p_parameterName) {
			DEBUG_MSG_FUNCTIONCALL("setParameterChanged(" << p_parameterName << ")");

			std::list<ParameterSet*> theList = __m_parameterSetLinks[p_parameterName];
			for (std::list<ParameterSet*>::iterator it2 = theList.begin(); it2 != theList.end(); ++it2) {
				(*it2)->setChanged(true);
			}

		}


		// when this has been checked once, it is reset automatically
		bool parameterSetChanged(const string& p_setName) {
//			if (__m_parameterSets.find(p_setName) == __m_parameterSets.end()) {
//				std::cout << p_setName << " not found!" << std::endl;
//				return false;
//			}

			bool parameterSetChanged = __m_parameterSets[p_setName]->getChanged();

			__m_parameterSets[p_setName]->setChanged(false);

			return parameterSetChanged;
		}



		/** @brief initialize an object with default values 0 and ENUM_COMPENG_NONE */
		INLINE void initObject(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("initObject(" << p_key << ")");

			__m_computeEngineUsage = 0;
			__m_elementInfo[p_key].devicePtr = 0;
			__m_elementInfo[p_key].dirty = ENUM_COMPENG_NONE;
		}



		/** @brief get iterator for element info pointing to first element */
		map<string, element_info>::iterator getElementInfoBegin() {
			DEBUG_MSG_FUNCTIONCALL("getElementInfoBegin()");

			return __m_elementInfo.begin();
		}

		/** @brief get iterator for element info pointing to the end */
		map<string, element_info>::iterator getElementInfoEnd() {
			DEBUG_MSG_FUNCTIONCALL("getElementInfoEnd()");

			return __m_elementInfo.end();
		}


		/** @brief set object for a given key and with size p_size */
		template<class FTYPE> INLINE void setObject(const string& p_key, FTYPE p_object, uint32_t p_size) {
			DEBUG_MSG_FUNCTIONCALL("setObject(" << p_key << ", " << p_object << ")");

			UniversalDataElement<FTYPE> ude = DataManager::createDataElement<FTYPE>(p_object, p_size);

			getMap<FTYPE>()[p_key] = ude;
			__m_elementInfo[p_key].typeId = ude.getTypeId();
		}

		/** @brief get object stored for key p_key, also get size in p_size */
		template<class FTYPE> INLINE FTYPE getObject(const string& p_key, uint32_t& p_size) {
			DEBUG_MSG_FUNCTIONCALL("getObject(" << p_key << ", " << p_size << ")");

			/* object not in container || object in container, but not of needed type */
			if (!objectExists(p_key) || DataTypes::getIdByType<FTYPE>() != getTypeId(p_key)) {
				std::cout << "*** ERROR: UniversalContainer does not contain object \"" << p_key << "\" of type \"" << DataTypes::getNameByType<FTYPE>() << "\"" << std::endl;
			}

			UniversalDataElement<FTYPE> ude = (getMap<FTYPE>())[p_key];

			p_size = ude.getSize();
			return ude.getData();
		}


		/** @brief get object stored for key p_key */
		template<class FTYPE> INLINE FTYPE getObject(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getObject(" << p_key << ")");

			uint32_t size;
			return getObject<FTYPE>(p_key, size);
		}


		/** @brief returns the number of elements a container element has. checks for existence. */
		template<class FTYPE> INLINE uint32_t getObjectSize(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getObjectSize(" << p_key << ")");

			return objectExists(p_key) ? getMap<FTYPE>()[p_key].getSize() : 0;
		}


		/** @brief get number of elements of a given type FTYPE */
		template<class FTYPE> INLINE uint32_t getSize() {
			DEBUG_MSG_FUNCTIONCALL("getSize()");

			return getMap<FTYPE>().size();
		}


		/**
		 * @brief The data element associated with p_key is just deleted from the map,
		 * but memory is not deallocated. Accordingly, the UniversalDataElement is returned.
		 * Existence is checked. Returns NULL element if p_key is not in container
		 */
		template<class FTYPE> INLINE UniversalDataElement<FTYPE> removeObject(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("removeObject(" << p_key << ")");

			UniversalDataElement<FTYPE> ude = getMap<FTYPE>()[p_key];
			getMap<FTYPE>().erase(p_key);

			return ude;
		}


		/** @brief check if an object has been inserted into the map before */
		INLINE bool objectExists(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("objectExists(" << p_key << ")");

			return (__m_elementInfo.find(p_key) != __m_elementInfo.end());
		}


		/** @brief check if data has changed */
		INLINE enum_computeEngine isDirty(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("isDirty(" << p_key << ")");

			return objectExists(p_key) ? __m_elementInfo[p_key].dirty : ENUM_COMPENG_NONE;
		}


		/** @brief set if data has changed */
		INLINE void setDirty(const string& p_key, enum_computeEngine p_dirty) {
			DEBUG_MSG_FUNCTIONCALL("setDirty(" << p_key << ")");

			__m_computeEngineUsage |= p_dirty;
			__m_elementInfo[p_key].dirty = p_dirty;
		}


		/** @brief by passing an enum_dirty, it can be checked if the various device types have been used */
		INLINE bool usedComputeEngine(enum_computeEngine p_dirty) {
			DEBUG_MSG_FUNCTIONCALL("usedComputeDevice(" << p_dirty << ")");

			return __m_computeEngineUsage & p_dirty;
		}


		/** @brief unset when data has been copied back */
		INLINE void unsetDirty(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("unsetDirty(" << p_key << ")");

			setDirty(p_key, ENUM_COMPENG_NONE);
		}

		/** @brief set (CUDA or other) device pointer */
		INLINE void setDevicePtr(const string& p_key, dt_void_t* p_ptr) {
			DEBUG_MSG_FUNCTIONCALL("setDevicePtr(" << p_key << ", " << p_ptr << ")");

			//std::cout << "setting devicePtr for " << p_key << ": " << p_ptr << std::endl;
			if (p_ptr != 0) {
				__m_computeEngineUsage |= ENUM_COMPENG_CUDA;
			}

			__m_elementInfo[p_key].devicePtr = p_ptr;
		}


		/** @brief set (CUDA or other) device pointer of given type */
		template<class FTYPE> INLINE void setDevicePtr(const string& p_key, FTYPE p_ptr) {
			DEBUG_MSG_FUNCTIONCALL("setDevicePtr(" << p_key << ", " << p_ptr << ")");

			//std::cout << "setting devicePtr for " << p_key << ": " << p_ptr << std::endl;
			if (p_ptr != 0) {
				__m_computeEngineUsage |= ENUM_COMPENG_CUDA;
			}


			__m_elementInfo[p_key].devicePtr = (void*) p_ptr;
		}

		/** @brief get (CUDA or other) device pointer as given type */
		template<class FTYPE> INLINE FTYPE getDevicePtr(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getDevicePtr<>(" << p_key << ")");

//			std::cout << "getting devicePtr for " << p_key << ": " << __m_elementInfo[p_key].devicePtr << std::endl;

			return reinterpret_cast<FTYPE>(__m_elementInfo[p_key].devicePtr);
		}


		/** @brief returns device ptr as void, helpful if it's not necessary to know the type but only size (e.g. copying) */
		INLINE void* getDevicePtr(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getDevicePtr(" << p_key << ")");

			return __m_elementInfo[p_key].devicePtr;
		}


		/** @brief returns type id of the stored object, as given in class DataTypes */
		INLINE int32_t getTypeId(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getTypeId(" << p_key << ")");

			return __m_elementInfo[p_key].typeId;
		}


		/** @brief returns type name of the stored object, as given in class DataTypes */
		INLINE string getTypeName(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getTypeName(" << p_key << ")");

			return DataTypes::getNameById(getTypeId(p_key));
		}
		
		
		INLINE void setNodeActivation(const string& p_key, bool p_active) {
			DEBUG_MSG_FUNCTIONCALL("setNodeActivation(" << p_key << ", " << p_active << ")");
			
			__m_nodeActivation[p_key] = p_active;			
		}
		
		
		INLINE bool getNodeActivation(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getNodeActivation(" << p_key << ")");
			
			return __m_nodeActivation[p_key];			
		}
		
		

	};

	___GENERATE_GETMAP_FUNCTION;

}

#endif /* UNIVERSALCONTAINER_H_ */
