/*
 * GraphReader.h
 *
 *  Created on: Oct 29, 2010
 *      Author: troth22s
 */

#ifndef GRAPHREADER_H_
#define GRAPHREADER_H_

#include <list>
#include <string>
#include <algorithm>
#include <vector>

#include <Graph/GraphWrapper.h>

namespace grip {

#undef CLASSNAME
#define CLASSNAME GraphReader

using namespace std;

/**
 * @class GraphReader
 * @brief this class reads a graph and generates a sequential processing list from it
 */
class CLASSNAME {

private:

	GraphWrapper __m_graphWrapper;

public:

	/** @brief Create uninitialized GraphReader, use init method afterwards */
	CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("GraphReader()");

	}

	/** @brief Create GraphReader for a given graph file, automatically init this (create graph wrapper) */
	CLASSNAME(const string& p_graphFile) {
		DEBUG_MSG_FUNCTIONCALL("GraphReader(" << p_graphFile << ")");

		__m_graphWrapper.init(p_graphFile);
	}

	/** @brief Create GraphReader for a given GraphWrapper */
	CLASSNAME(GraphWrapper& p_graphWrapper) {
		DEBUG_MSG_FUNCTIONCALL("GraphReader(...)");

		__m_graphWrapper = p_graphWrapper;
	}

	/** @brief Default destructor */
	~CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("~GraphReader()");

	}

	/** @brief Initialize GraphReader with a given graph file */
	void init(const string& p_graphFile) {
		DEBUG_MSG_FUNCTIONCALL("init(" << p_graphFile << ")");

		__m_graphWrapper.init(p_graphFile);
	}

	/** @brief Initialize GraphReader with a given GraphWrapper */
	void init(GraphWrapper& p_graphWrapper) {
		__m_graphWrapper = p_graphWrapper;
	}

	/** @brief generate a correctly ordered sequential traversal for the supplied graph */
	vector<string> generateOrderedTraversal(list<string>& p_nodeNames) {
		DEBUG_MSG_FUNCTIONCALL("generateOrderedTraversal()");

		vector<string> orderedNodeNames;

		list<string>::iterator itAll = p_nodeNames.begin();

		while (!p_nodeNames.empty()) {
			// iterate over all nodes in list nodeNames
			for (list<string>::iterator itAll = p_nodeNames.begin(); itAll != p_nodeNames.end(); itAll++) {

				string currName = *itAll;

				DEBUG_MSG_RUNTIMEINFO("trying node " << currName);

				//check if any of the predecessors for the current node *it is still in list nodeNames
				bool stillHasPredecessor = false;

				list<string> pred = __m_graphWrapper.getPredecessorNames(currName);

				for (list<string>::iterator itPred = pred.begin(); itPred != pred.end(); itPred++) {
					// still in node names? => current node is not next to be processed!
					if (find(p_nodeNames.begin(), p_nodeNames.end(), *itPred) != p_nodeNames.end()) {
						DEBUG_MSG_RUNTIMEINFO("node " << currName << " still has predecessor left (" << *itPred << "), TRYING NEXT!");
						stillHasPredecessor = true;
						break;
					}
				}

				// node has no predecessor left, so it can be processed next in list
				if (!stillHasPredecessor) {
					DEBUG_MSG_RUNTIMEINFO("node " << currName << " has no predecessor left or is a loop with only loop-internal predecessors, ADDING!");

					// the differentiation between special nodes and filter nodes has to be made here
					/** how should loop nodes be handeled?
					 * - add loop start node to ordered names
					 * - make an ordered traversal of loop node successors until loop node is reached again
					 * - add this ordered traversal to ordered names
					 * - add loop end node to ordered names
					 */

					string currType;
					__m_graphWrapper.getNodeTypeByName(currName, currType);

					DEBUG_MSG_RUNTIMEINFO("!stillHasPredecessor: " << currType);

					if (currType == "node_loop_start") {
						DEBUG_MSG_RUNTIMEINFO("node_loop_start!");
						string loopName = currName.substr(0, currName.length() - string("_start").length());
						string loopStartName = currName;
						string loopEndName = loopName + "_end";

						list<string> currentSuccessorList = __m_graphWrapper.getLoopSuccessorNames(loopStartName);
						set<string> currentSuccessorSet;
						for (list<string>::iterator it = currentSuccessorList.begin(); it != currentSuccessorList.end(); it++) {
							currentSuccessorSet.insert(*it);
						}

						int oldSize = currentSuccessorSet.size();
						int newSize = oldSize;

						do {
							oldSize = newSize;

							list<string> successorsToAppend;

							for (set<string>::iterator it = currentSuccessorSet.begin(); it != currentSuccessorSet.end(); it++) {
								list<string> theSuccessors;

								if (*it != loopEndName) {
									theSuccessors = __m_graphWrapper.getSuccessorNames(*it);
								}

								for (list<string>::iterator innerIt = theSuccessors.begin(); innerIt != theSuccessors.end(); innerIt++) {
									successorsToAppend.push_back(*innerIt);
								}

							}

							for (list<string>::iterator it = successorsToAppend.begin(); it != successorsToAppend.end(); it++) {
								currentSuccessorSet.insert(*it);
							}

							newSize = currentSuccessorSet.size();

						} while (newSize > oldSize);

						currentSuccessorList.clear();

						for (set<string>::iterator it = currentSuccessorSet.begin(); it != currentSuccessorSet.end(); it++) {
							currentSuccessorList.push_back(*it);
							DEBUG_MSG_RUNTIMEINFO("### " << *it);
						}

						vector<string> orderedSubList = generateOrderedTraversal(currentSuccessorList);

						orderedNodeNames.push_back(loopStartName);
						p_nodeNames.remove(loopStartName);
						DEBUG_MSG_RUNTIMEINFO("### removed " << loopStartName);

						for (vector<string>::iterator it = orderedSubList.begin(); it != orderedSubList.end(); it++) {
							orderedNodeNames.push_back(*it);

							p_nodeNames.remove(*it);
							DEBUG_MSG_RUNTIMEINFO("### removed " << *it);
						}

						itAll = p_nodeNames.begin(); // keep the iterator valid

					} else {
						orderedNodeNames.push_back(currName);
						p_nodeNames.remove(currName);
						itAll = p_nodeNames.begin(); // keep the iterator valid
						break;
					}
				}
			}
		}

		return orderedNodeNames;

	}

};

}

#endif /* GRAPHREADER_H_ */
