/*
 * GraphRunner.h
 *
 *  Created on: 08.11.2010
 *      Author: troth
 */

#ifndef GRAPHRUNNER_H_
#define GRAPHRUNNER_H_

#include <Utils/Common.h>
#include <Graph/GraphReader.h>
#include <Graph/NodeLoader.h>
#include <Graph/NodeHelper.h>
#include <Graph/GraphVisualizer.h>
#include <Data/UniversalContainer.h>
#include <Data/EnumComputeEngine.h>
#include <Data/TempBufferRepository.h>
#include <Image/UniversalImage.h>
#include <Gui/ParameterizerGui.h>
#include <Utils/TimerMacros.h>

//#include <OpenThreads/Mutex>
#include <SFML/System.hpp>


#include <vector>
#include <stack>

namespace grip {

	#undef CLASSNAME
	#define CLASSNAME GraphRunner

	using namespace std;

	/**
	 * @class GraphRunner
	 *
	 * This class utilizes the GraphWrapper and GraphReader class to organize an ordered
	 * graph traversal with actual node calls. It also uses the NodeLoader to dynamically
	 * load the needed symbols at runtime.
	 */
	class CLASSNAME {

	private:

		GraphReader* 			__m_graphReader;
		GraphWrapper* 			__m_graphWrapper;

		GraphVisualizer 		__m_graphVisualizer;

		TempBufferRepositoryDirectory __m_tbRepoDir; 

		NodeLoader 				__m_nodeLoader;
		vector<string> 			__m_orderedNames;
		list<string> 			__m_nodeTypes;
		UniversalContainer 		__m_universalContainer;
		ParameterizerGuiThread*	__m_guiThread;
		sf::Thread*				__m_guiThreadSf;
		
		enum_visualizer_modes 	__m_visMode;

		sf::Mutex* 				__m_parameterMutex;

		___GRIP_INITIALIZE_TIMER(0);


		map<string, FunctionTypes::f_run_node_t> __m_runFunctions;
		map<string, bool> __m_nodeDeactivationInfo;

		bool __m_visualize;
		bool __m_parameterizerGui;

		#define PATH_PREFIX "../" + p_nodeName + "/Debug/"
		#define PATH_SUFFIX	"lib" + p_nodeName + ".so"

		std::string 			__m_path_prefix;
		std::string				__m_path_directory;
		std::string				__m_lib_prefix;
		std::string				__m_lib_suffix;


		/** @brief load the graph, generate the sequential traversal, load all needed symbols */
		void load() {
			DEBUG_MSG_FUNCTIONCALL("load()");

			list<string> nodeNames = __m_graphWrapper->getAllNodeNames();

			__m_orderedNames = __m_graphReader->generateOrderedTraversal(nodeNames);
			__m_nodeTypes = __m_graphWrapper->getAllNodeTypes();

			// load run functions and store them in a map
			for (list<string>::iterator it = __m_nodeTypes.begin(); it != __m_nodeTypes.end(); it++) {
				
				std::cout << *it << std::endl;
				
				// load .so for node type *it, if it's a filter and no loop or other type of node
				if (*it != "node_loop_start" && *it != "node_loop_end") {
					__m_runFunctions[*it] = __m_nodeLoader.loadNodeFunction<FunctionTypes::f_run_node_t> (*it, "run_node");
					FunctionTypes::f_can_be_deactivated_t f = __m_nodeLoader.loadNodeFunction<FunctionTypes::f_can_be_deactivated_t> (*it, "can_be_deactivated");
					__m_nodeDeactivationInfo[*it] = f != 0 ? f() : false;
				}
			}

		}

		/** @brief Initializes the dynamic parameterizer gui. Ensure that __m_graphWrapper is already initialized here! */
		void initParameterizerGui() {
			DEBUG_MSG_FUNCTIONCALL("initParameterizerGui()");

			__m_parameterMutex = new sf::Mutex();
			
			if (__m_parameterizerGui) {
				
				__m_guiThread = new grip::ParameterizerGuiThread(__m_graphWrapper, &__m_universalContainer, __m_nodeDeactivationInfo, __m_parameterMutex);

				__m_guiThreadSf = new sf::Thread(&grip::ParameterizerGuiThread::Run, __m_guiThread);
				
				__m_guiThreadSf->launch();
				
				// wait until initialization is finished
				std::cout << "WAITING UNTIL READY..." << std::flush;
				while(!(__m_guiThread->isReady()));
				std::cout << "READY!" << std::endl;

			} else {
				__m_guiThread = 0;
			}

		}






		void cleanupMemory() {
			DEBUG_MSG_FUNCTIONCALL("cleanupMemory()");

			DEBUG_MSG_RUNTIMEINFO("CUDA used in uniCon? " << __m_universalContainer.usedComputeEngine(ENUM_COMPENG_CUDA));
			DEBUG_MSG_RUNTIMEINFO("HOST used in uniCon? " << __m_universalContainer.usedComputeEngine(ENUM_COMPENG_HOST));
			DEBUG_MSG_RUNTIMEINFO("CUDA used in tbRepoDir? " << __m_tbRepoDir.usedComputeEngine(ENUM_COMPENG_CUDA));
			DEBUG_MSG_RUNTIMEINFO("HOST used in tbRepoDir? " << __m_tbRepoDir.usedComputeEngine(ENUM_COMPENG_HOST));

			if (__m_universalContainer.usedComputeEngine(ENUM_COMPENG_CUDA) || __m_tbRepoDir.usedComputeEngine(ENUM_COMPENG_CUDA)) {
				if (__m_runFunctions.find("node_cleanup_cuda") == __m_runFunctions.end()) {
					DEBUG_MSG_RUNTIMEINFO("LOADING node_cleanup_cuda");
					__m_runFunctions["node_cleanup_cuda"] = __m_nodeLoader.loadNodeFunction<FunctionTypes::f_run_node_t> ("node_cleanup_cuda", "run_node");
				}

				DEBUG_MSG_RUNTIMEINFO("RUNNING node_cleanup_cuda");

				__m_runFunctions["node_cleanup_cuda"]("node_cleanup_cuda", *__m_graphWrapper, __m_universalContainer, __m_tbRepoDir);
			}

			if (__m_universalContainer.usedComputeEngine(ENUM_COMPENG_HOST) || __m_tbRepoDir.usedComputeEngine(ENUM_COMPENG_HOST)) {
				if (__m_runFunctions.find("node_cleanup_host") == __m_runFunctions.end()) {
					DEBUG_MSG_RUNTIMEINFO("LOADING node_cleanup_host");
					__m_runFunctions["node_cleanup_host"] = __m_nodeLoader.loadNodeFunction<FunctionTypes::f_run_node_t> ("node_cleanup_host", "run_node");
				}

				DEBUG_MSG_RUNTIMEINFO("RUNNING node_cleanup_host");

				
				
				__m_runFunctions["node_cleanup_host"]("node_cleanup_host", *__m_graphWrapper, __m_universalContainer, __m_tbRepoDir);
			}
		}

	public:

		/** @brief Create uninitialized GraphRunner. Has to be initialized with the init()-method afterwards */
		CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("GraphRunner()");

			__m_graphReader = 0;
			__m_graphWrapper = 0;

			__m_visMode = ENUM_GRAPHVISUALIZER_NONE;

			__m_path_prefix = "";
			__m_path_directory = "";
			__m_lib_prefix = "lib";
			__m_lib_suffix = ".so";
			__m_nodeLoader.setPathAndLibraryInformation(__m_path_prefix, __m_path_directory, __m_lib_prefix, __m_lib_suffix);
		}

		/**
		 * @brief Create GraphRunner for a graph file. We can visualize the graph and provide the parameterizer gui here by setting
		 * the respective bools.
		 */
		CLASSNAME(const string& p_graphFile, bool p_visualize, enum_visualizer_modes p_visMode, bool p_parameterizerGui) {
			DEBUG_MSG_FUNCTIONCALL("GraphRunner(" << p_graphFile << ", " << p_visualize << ")");

			__m_path_prefix = "";
			__m_path_directory = "";
			__m_lib_prefix = "lib";
			__m_lib_suffix = ".so";
			__m_nodeLoader.setPathAndLibraryInformation(__m_path_prefix, __m_path_directory, __m_lib_prefix, __m_lib_suffix);

			init(p_graphFile, p_visualize, p_visMode, p_parameterizerGui);
		}


		/** @brief default destructor, also handles the gui thread */
		~CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("~GraphRunner()");

			cleanupMemory();

			delete __m_graphWrapper;
			delete __m_graphReader;

			if (__m_guiThread != 0) {
				__m_guiThreadSf->terminate();
				__m_guiThreadSf->wait();

				delete __m_guiThread;
			
			}
			
			
			delete __m_parameterMutex;
		}



		grip::ParameterizerGuiThread* getGuiThread() {
			DEBUG_MSG_FUNCTIONCALL("getGuiThread()");

			return __m_guiThread;
		}



		bool parameterSetChanged(const string& p_setName) {
			DEBUG_MSG_FUNCTIONCALL("parameterSetChanged(" << p_setName << ")");

//			std::cout << "parameterSetChanged(" << p_setName << ")" << std::endl;
			bool retVal = __m_universalContainer.parameterSetChanged(p_setName);

//			std::cout << "retVal: " << retVal << std::endl;

			return retVal;

		}


		void setPathPrefixAndSubDirectory(const char* p_pathPrefix, const char* p_subDir) {
			DEBUG_MSG_FUNCTIONCALL("setPathPrefixAndSubDirectory(" << p_pathPrefix << ", " << p_subDir << ")");

			__m_path_prefix = p_pathPrefix;
			__m_path_directory = p_subDir;
			__m_nodeLoader.setPathAndLibraryInformation(__m_path_prefix, __m_path_directory, __m_lib_prefix, __m_lib_suffix);
		}

		void setLibPrefixAndSuffix(const char* p_libPrefix, const char* p_libSuffix) {
			DEBUG_MSG_FUNCTIONCALL("setLibPrefixAndSuffix(" << p_libPrefix << ", " << p_libSuffix << ")");

			__m_lib_prefix = p_libPrefix;
			__m_lib_suffix = p_libSuffix;
			__m_nodeLoader.setPathAndLibraryInformation(__m_path_prefix, __m_path_directory, __m_lib_prefix, __m_lib_suffix);
		}


		/**
		 * @brief Initialize GraphRunner for a graph file. We can visualize the graph and provide the parameterizer gui here by setting
		 * the respective bools.
		 *
		 * The plugin path as well as the library pre- and suffix have to be set BEFORE calling init(...), as this method already relies
		 * on this information.
		 */
		void init(const std::string& p_graphFile, bool p_visualize, enum_visualizer_modes p_visMode, bool p_parameterizerGui) {
			DEBUG_MSG_FUNCTIONCALL("init(" << p_graphFile << ", " << p_visualize << ")");

			std::cout << "Initializing graphWrapper" << std::endl;
			__m_graphWrapper = new grip::GraphWrapper(p_graphFile);
			__m_graphReader = new grip::GraphReader;


			std::cout << "Initializing graphReader" << std::endl;
			__m_graphReader->init(*__m_graphWrapper);

			__m_visualize = p_visualize;

			std::cout << "Loading" << std::endl;
			load();

			__m_visMode = p_visMode;

			std::cout << "updateVisualizer()" << std::endl;
			updateVisualizer();

			__m_parameterizerGui = p_parameterizerGui;

			__m_tbRepoDir.addRepositoryByName("cuda_general");

			std::cout << "Initializing parameterizerGui" << std::endl;
			initParameterizerGui();


			// add all parameter sets and parameters
			std::list<std::string> parameterGroupNames = __m_graphWrapper->getAllParameterGroupNames();
			for (std::list<std::string>::iterator it = parameterGroupNames.begin(); it != parameterGroupNames.end(); ++it) {
				std::cout << "adding parameter group " << *it << std::endl;
				__m_universalContainer.addParameterSet(*it);

				std::list<std::string> groupElementNames = __m_graphWrapper->getAllParameterGroupElements(*it);

				for (std::list<std::string>::iterator it2 = groupElementNames.begin(); it2 != groupElementNames.end(); ++it2) {
					__m_universalContainer.addSetLinkToParameter(*it2, *it);
				}
			}

		}


		/** @brief easy way to reload an updated graph file ;) */
		void reset() {
			DEBUG_MSG_FUNCTIONCALL("reset()");

			cleanupMemory();

			if (__m_guiThread) {
				__m_guiThread->stopWorking();
				__m_guiThreadSf->wait();
				delete __m_guiThread;
				__m_guiThread = 0;
			}

			delete __m_graphWrapper;
			__m_graphWrapper = 0;

			delete __m_graphReader;
			__m_graphReader = 0;

		}


		/** @brief the visualizer is redrawn here */
		void updateVisualizer() {
			DEBUG_MSG_FUNCTIONCALL("updateVisualizer()");

			if (__m_visualize) {
				__m_graphVisualizer.setWrapper(__m_graphWrapper);
				__m_graphVisualizer.generateAndDraw(__m_visMode, __m_orderedNames);
			}
		}


		typedef struct __LOOPOBJECT {
			string loopId;
			vector<string>::iterator jumpTarget;
			uint32_t executedIterations;
			uint32_t iterationLimit;
		} LoopObject;

		/** @brief reset loop object to an uninitialized one */
		inline void resetLoopObject(LoopObject& p_loopObject) {
			p_loopObject.executedIterations = 0;
			p_loopObject.iterationLimit = 0;
		}

		/** @brief check if loop object is uninitialized */
		inline bool loopObjectIsNull(LoopObject& p_loopObject) {
			return p_loopObject.executedIterations == 0 && p_loopObject.iterationLimit == 0;
		}


		/** @brief run the whole processing */
		void run() {
			DEBUG_MSG_FUNCTIONCALL("run()");

			___GRIP_START_TIMER(0);

			__m_parameterMutex->lock();

			stack<LoopObject*> loopStack;

			// iterate over ordered node list and run each filter with given parameters
			for (vector<string>::iterator it = __m_orderedNames.begin(); it != __m_orderedNames.end(); it++) {
//				std::cout << "running " << *it << std::endl;

				string nodeType;
				__m_graphWrapper->getNodeTypeByName(*it, nodeType);

//				std::cout << nodeType << std::endl;

				if (nodeType == "node_loop_start") {
					DEBUG_MSG_RUNTIMEINFO("EXECUTION REACHED node_loop_start");
					// set jump target to start node if we are in the first run

					if (loopStack.empty() || loopStack.top()->loopId != *it) {
						LoopObject* loopObject = new LoopObject;
						loopObject->loopId = *it;
						loopObject->jumpTarget = it;
						loopObject->executedIterations = 0;

						if (__m_graphWrapper->isAttributeDynamic(*it, "iterations")) {
							string parameterName;
							__m_graphWrapper->getDataFromNode(*it, "iterations", parameterName);
							loopObject->iterationLimit = __m_universalContainer.getObject<dt_uint32_t>(parameterName);
						} else {
							__m_graphWrapper->getDataFromNode(*it, "iterations", loopObject->iterationLimit);
						}

						loopStack.push(loopObject);
					}

				} else if (nodeType == "node_loop_end") {
					DEBUG_MSG_RUNTIMEINFO("EXECUTION REACHED node_loop_end");
					loopStack.top()->executedIterations++;
					DEBUG_MSG_RUNTIMEINFO(loopStack.top()->loopId << ": " << loopStack.top()->executedIterations << " of " << loopStack.top()->iterationLimit << " iterations done");

					if (loopStack.top()->executedIterations < loopStack.top()->iterationLimit) {
						DEBUG_MSG_RUNTIMEINFO("jumping to loop start");
						it = loopStack.top()->jumpTarget;
					} else { // reset data and continue
						DEBUG_MSG_RUNTIMEINFO("loop done");
						LoopObject* lo = loopStack.top();
						loopStack.pop();
						delete lo;
					}

				} else {
					DEBUG_MSG_RUNTIMEINFO("RUNNING " << *it << " (type: " << nodeType << ")");

					__m_runFunctions[nodeType](*it, *__m_graphWrapper, __m_universalContainer, __m_tbRepoDir);
				}
			}

			__m_parameterMutex->unlock();

			___GRIP_END_TIMER_AND_PRINT(0, "### graphRunner", "total traversal time", "run()");
		}


		/** @brief Enclosure of the UniversalContainer object, which is automatically set dirty */
		template<class FTYPE> void setParameter(const string& p_key, FTYPE p_param, uint32_t p_size) {
			DEBUG_MSG_FUNCTIONCALL("setParameter(" << p_key << ", " << p_param << ", " << p_size << ")");

			if (!__m_universalContainer.objectExists(p_key)) {
				__m_universalContainer.setObject<FTYPE>(p_key, p_param, p_size);
				__m_universalContainer.initObject(p_key);
			} else {
				__m_universalContainer.setObject<FTYPE>(p_key, p_param, p_size);
			}
			__m_universalContainer.setDirty(p_key, ENUM_COMPENG_HOST);
		}


		/** @brief Set device (graphics memory) pointer for the object belonging to p_key */
		template<class FTYPE> void setDevicePtr(const string& p_key, FTYPE p_ptr) {
			DEBUG_MSG_FUNCTIONCALL("setDevicePtr(" << p_key << ", " << p_ptr << ")");

			if (__m_universalContainer.objectExists(p_key)) {
				__m_universalContainer.setDevicePtr<FTYPE>(p_key, p_ptr);
			}
		}
		
		
		/** @brief Get device (graphics memory) pointer for the object belonging to p_key */
		template<class FTYPE> FTYPE getDevicePtr(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getDevicePtr(" << p_key << ")");
			
			if (__m_universalContainer.objectExists(p_key)) {
			
				return __m_universalContainer.getDevicePtr<FTYPE>(p_key);
				
			} else {
			
				return (FTYPE)0;
				
			}
			
			
		}


		/** @brief set OpenGL buffer id for the object belonging to p_key */
//		void setBufferId(const string& p_key, GLuint p_id) {
//			DEBUG_MSG_FUNCTIONCALL("setBufferId(" << p_key << ", " << p_id << ")");
//
//			if (__m_universalContainer.objectExists(p_key)) {
//				__m_universalContainer.setBufferId(p_key, p_id);
//			}
//		}


		/** @brief set dirty state p_dirty for object belonging to p_key */
		void setDirty(const string& p_key, enum_computeEngine p_dirty) {
			DEBUG_MSG_FUNCTIONCALL("setDirty(" << p_key << ", " << p_dirty << ")");

			if (__m_universalContainer.objectExists(p_key)) {
				__m_universalContainer.setDirty(p_key, p_dirty);
			}
		}


		/** @brief set dirty flag to ENUM_COMPENG_NONE */
		void unsetDirty(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("unsetDirty(" << p_key << ")");

			setDirty(p_key, ENUM_COMPENG_NONE);
		}


		/** @brief get dirty flag */
		enum_computeEngine isDirty(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("isDirty( " << p_key << ")");

			return __m_universalContainer.isDirty(p_key);
		}

		/** @brief get parameter p_key */
		template<class FTYPE> FTYPE getParameter(const string& p_key) {
			DEBUG_MSG_FUNCTIONCALL("getParameter(" << p_key << ")");

			return __m_universalContainer.getObject<FTYPE>(p_key);
		}

		/**
		 * @brief print all necessary input parameters for a graph, i.e. all
		 * input parameters (images) that are not provided by preceeding nodes.
		 */
		void printNecessaryInputParameters() {
			DEBUG_MSG_FUNCTIONCALL("printNecessaryInputParameters()");

			__m_graphWrapper->printNecessaryInputParameters(__m_orderedNames);
		}

		/** @brief get GraphWrapper for this GraphRunner */
		GraphWrapper* getGraphWrapper() {
			DEBUG_MSG_FUNCTIONCALL("getGraphWrapper()");

			return __m_graphWrapper;
		}

	};


}




#endif /* GRAPHRUNNER_H_ */
