/*
 * GraphWrapper.h
 *
 *  Created on: Oct 27, 2010
 *      Author: troth22s
 */

#ifndef GRAPHWRAPPER_H_
#define GRAPHWRAPPER_H_

#include <tinyxml.h>

#include <iostream>
#include <vector>
#include <list>
#include <string>
#include <map>
#include <set>
#include <algorithm>

#include <Utils/StringUtils.h>
#include <Utils/Common.h>

using namespace std;

namespace grip {

#undef CLASSNAME
#define CLASSNAME GraphWrapper

/**
 * @class GraphWrapper
 * @brief This class wraps the access to the concrete graph representation, in this case XML.
 *
 * TODO: Note that this could (and probably should) be made a class hierarchy for different
 * possible representations(could also be different XML formats or parsers!)
 */
class CLASSNAME {

private:
	TiXmlDocument __m_xmlDocument;
	TiXmlElement* __m_rootNode_graph;
	TiXmlElement* __m_rootNode_appParameters;
	TiXmlElement* __m_rootNode_parameterGroups;

	string __m_pathToFile;

	map<string, TiXmlElement*> __m_parameterGroupCacheByName; // access parameter groups directly by name

	map<string, TiXmlElement*> __m_appParameterCacheByName; // access application parameters directly by name

	map<string, TiXmlElement*> __m_nodeCacheByName; // access nodes directly by name
	map<string, list<TiXmlElement*> > __m_nodeCacheByType; // access node lists by type

	bool __m_initialized; // should be set to true when init() is called (and performed)

	/** @brief load the xml file via TinyXML */
	bool load() {
		DEBUG_MSG_FUNCTIONCALL("load()");

		std::cout << "Loading " << __m_pathToFile << "...";
		__m_xmlDocument = TiXmlDocument(__m_pathToFile);
		bool success = __m_xmlDocument.LoadFile();


		__m_rootNode_graph = NULL;
		__m_rootNode_appParameters = NULL;
		__m_rootNode_parameterGroups = NULL;


		if (success) {
			std::cout << "SUCCESS!" << endl;

			for (TiXmlElement* curChild = __m_xmlDocument.FirstChildElement(); curChild; curChild = curChild->NextSiblingElement()) {

				std::string nodeVal = curChild->ValueStr();

				if (nodeVal == "graph") {
					__m_rootNode_graph = curChild;	// first child element should be <graph>

				} else if (nodeVal == "application_parameters") {
					__m_rootNode_appParameters = curChild; // application parameter section, could also be NULL
				} else if (nodeVal == "parameter_groups") {
					__m_rootNode_parameterGroups = curChild; // parameter group section, could also be NULL
				}

			}

		} else {
			std::cout << "FAIL!" << endl;
		}

		return success;
	}


	/** @brief prepare caches for processing nodes by node name and by node type */
	void prepareNodeCaches() {
		DEBUG_MSG_FUNCTIONCALL("prepareNodeCaches()");

		for (TiXmlElement* currNode = __m_rootNode_graph->FirstChildElement(); currNode != 0; currNode = currNode->NextSiblingElement()) {

			// no check for element node necessary, as we are only traversing element nodes!
			string nodeType = currNode->ValueStr();
			__m_nodeCacheByType[nodeType].push_back(currNode);

			// attribute "name" has to be there for each node, so we don't need to check for existence!
			string nodeName = currNode->Attribute("name");
			__m_nodeCacheByName[nodeName] = currNode;

			DEBUG_MSG_RUNTIMEINFO("Cached node " << nodeName << " (type: " << nodeType << ")");
		}
	}


	void prepareApplicationParameterCaches() {
		DEBUG_MSG_FUNCTIONCALL("prepareApplicationParameterCaches()");

		if (!__m_rootNode_appParameters) {
			return;
		}

		for (TiXmlElement* currNode = __m_rootNode_appParameters->FirstChildElement(); currNode != 0; currNode = currNode->NextSiblingElement()) {

			string parameterName = currNode->ValueStr();
			__m_appParameterCacheByName[parameterName] = currNode;

			DEBUG_MSG_RUNTIMEINFO("Cached application parameter " << parameterName << "!");
		}
	}


	void prepareParameterGroupCaches() {
		DEBUG_MSG_FUNCTIONCALL("prepareParameterGroupCaches()");

		if (!__m_rootNode_parameterGroups) {
			return;
		}

		for (TiXmlElement* currNode = __m_rootNode_parameterGroups->FirstChildElement(); currNode != 0; currNode = currNode->NextSiblingElement()) {

			string elementName = currNode->ValueStr();

			std::cout << "element name: " << elementName << std::endl;
			if (elementName == "group") {
				string groupName = currNode->Attribute("name");
				std::cout << "groupName: " << groupName << std::endl;

				__m_parameterGroupCacheByName[groupName] = currNode;
			}

		}
	}


	/**
	 *  @brief normalize the graph, which means that for each predecessor, the respective successor information is added, as well as
	 *  the predecessor information for each successor. This is needed to avoid repeated iterations in the traversal generation.
	 */
	void normalizeGraph() {
		DEBUG_MSG_FUNCTIONCALL("checkForNormalForm()");

		DEBUG_MSG_RUNTIMEINFO("Traversing all processing nodes for normal form check...");

		// iterate over all nodes
		for (map<string, TiXmlElement*>::iterator itAll = __m_nodeCacheByName.begin(); itAll != __m_nodeCacheByName.end(); itAll++) {
			string currName = itAll->first;
			list<string> pred = getPredecessorNames(currName);

			DEBUG_MSG_RUNTIMEINFO("checking all predecessors of " << currName << "...");

			//check all predecessors of the current node for successor information
			for (list<string>::iterator itPred = pred.begin(); itPred != pred.end(); itPred++) {
				list<string> succForPred = getSuccessorNames(*itPred);

				DEBUG_MSG_RUNTIMEINFO("checking if " << currName << " is successor element of " << (*itPred) << "...");

				// successor missing? add!
				if (find(succForPred.begin(), succForPred.end(), currName) == succForPred.end()) {
					TiXmlElement* currPred = getNodeByName(*itPred);
					TiXmlElement* theNode = new TiXmlElement("successor");
					TiXmlText *text = new TiXmlText(currName);
					theNode->LinkEndChild(text);
					currPred->InsertBeforeChild(currPred->FirstChildElement(), *theNode);

					DEBUG_MSG_RUNTIMEINFO("GRAPH DOES NOT HAVE NORMAL FORM! ADDED SUCCESSOR INFORMATION!");
					//cerr << "ERROR! GRAPH DOES NOT HAVE NORMAL FORM!" << endl;
					//exit(1);
				} else {
					DEBUG_MSG_RUNTIMEINFO("OK!");
				}
			}

			list<string> succ = getSuccessorNames(itAll->first);

			DEBUG_MSG_RUNTIMEINFO("checking all successors of " << currName << "...");

			// check all successors of the current node for predecessor information
			for (list<string>::iterator itSucc = succ.begin(); itSucc != succ.end(); itSucc++) {
				list<string> predForSucc = getPredecessorNames(*itSucc);

				DEBUG_MSG_RUNTIMEINFO("checking if " << currName << " is predecessor element of " << (*itSucc) << "...");

				// predecessor missing? add!
				if (find(predForSucc.begin(), predForSucc.end(), currName) == predForSucc.end()) {
					TiXmlElement* currSucc = getNodeByName(*itSucc);
					TiXmlElement* theNode = new TiXmlElement("predecessor");
					TiXmlText *text = new TiXmlText(currName);
					theNode->LinkEndChild(text);
					currSucc->InsertBeforeChild(currSucc->FirstChildElement(), *theNode);

					DEBUG_MSG_RUNTIMEINFO("GRAPH DID NOT HAVE NORMAL FORM -> ADDED PREDECESSOR INFORMATION!");
				} else {
					DEBUG_MSG_RUNTIMEINFO("OK!");
				}
			}
		}

		DEBUG_MSG_RUNTIMEINFO("FINISHED! GRAPH HAS NORMAL FORM!");
	}

	/** @brief get node with the respective name from node cache */
	TiXmlElement* getNodeFromCache(const string& p_name) {
		DEBUG_MSG_FUNCTIONCALL("getNodeFromCache()");

		if (__m_nodeCacheByName.find(p_name) == __m_nodeCacheByName.end()) {
			DEBUG_MSG_WARNING("Node " << p_name << " does not exist or is not cached!");
		}

		return __m_nodeCacheByName[p_name];
	}

	/** @brief get child element of some node acting as root. p_nodeName can also be a path with the levels being separated by / */
	TiXmlElement* getChildNodeByName(TiXmlElement* p_rootNode, const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getChildNodeByName(" << p_rootNode->ValueStr() << ", " << p_nodeName << ")");

		vector<string> pathToNode;
		StringUtils::explode(p_nodeName, "/", pathToNode);

		TiXmlElement* currParentNode = p_rootNode;

		for (vector<string>::iterator currWantedNode = pathToNode.begin(); currWantedNode != pathToNode.end(); currWantedNode++) {
			for (TiXmlElement* currNode = currParentNode->FirstChildElement(); currNode != 0; currNode = currNode->NextSiblingElement()) {

				if (*currWantedNode == currNode->ValueStr()) {
					currParentNode = currNode;

					break;
				}

			}

		}

		return currParentNode != p_rootNode ? currParentNode : 0;
	}


	/** @brief same as getNodeFromCache(const string&) */
	TiXmlElement* getNodeByName(const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getNodeByName(" << p_nodeName << ")");

		return getNodeFromCache(p_nodeName);
	}


	/** @brief this makes it possible to get a node with a path given from the root node of type DOMDocument */
	TiXmlElement* getNodeByPath(const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getNodeByName(" << p_nodeName << ")");

		vector<string> pathToNode;
		StringUtils::explode(p_nodeName, "/", pathToNode);

		if (pathToNode.size() < 1) {
			return 0;
		}

		return getChildNodeByName(__m_rootNode_graph, p_nodeName);
	}


public:

	/** @brief create uninitialized GraphWrapper */
	CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("GraphWrapper()");

		__m_initialized = false;
	}

	/** @brief Create GraphWrapper and initialize with a path (XML file) */
	CLASSNAME(const string& p_path) {
		DEBUG_MSG_FUNCTIONCALL("GraphWrapper(" << p_path << ")");

		__m_initialized = false;

		init(p_path);
	}

	/** @brief standard destructor, does nothing */
	~CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("~GraphWrapper()");
	}

	/**
	 * @brief Initialize this instance with the given graph description:
	 * - load
	 * - prepare caches
	 * - normalize
	 */
	void init(const string& p_path) {
		DEBUG_MSG_FUNCTIONCALL("init(" << p_path << ")");

		if (!__m_initialized) {
			__m_pathToFile = p_path;

			load();

			prepareNodeCaches();
			prepareApplicationParameterCaches();
			prepareParameterGroupCaches();

			normalizeGraph();

			__m_initialized = true;
		}
	}


	/** @brief get list of elements with a specific name nested within an element */
	list<string> getNestedElementsList(const string& p_nodeName, const string& p_elementName) {
		DEBUG_MSG_FUNCTIONCALL("getNestedElementsList(" << p_nodeName << ", " << p_elementName << ")");

		list<string> theList;

		for (TiXmlElement* theNode = getNodeFromCache(p_nodeName)->FirstChildElement(); theNode != 0; theNode = theNode->NextSiblingElement()) {
			string nodeName = theNode->ValueStr();
			if (nodeName == p_elementName) {
				string nodeContent = theNode->GetText();
				theList.push_back(nodeContent);
			}
		}

		return theList;
	}


	/** @brief get node names of predecessors for a node given by p_nodeName */
	list<string> getPredecessorNames(const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getPredecessorNames(" << p_nodeName << ")");

		return getNestedElementsList(p_nodeName, "predecessor");
	}


	/** @brief get node names of successors for a node given by p_nodeName */
	list<string> getSuccessorNames(const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getSuccessorNames(" << p_nodeName << ")");

		return getNestedElementsList(p_nodeName, "successor");
	}


	/** @brief get node names of loop successors for a node given by p_nodeName */
	list<string> getLoopSuccessorNames(string p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getSuccessorNames(" << p_nodeName << ")");

		return getNestedElementsList(p_nodeName, "loop_successor");
	}


	/** @brief return all node names in the graph (each one is unique) */
	list<string> getAllNodeNames() {
		DEBUG_MSG_FUNCTIONCALL("getAllNodeNames()");

		list<string> theList;

		for (map<string, TiXmlElement*>::iterator it = __m_nodeCacheByName.begin(); it != __m_nodeCacheByName.end(); ++it) {
			theList.push_back(it->first);
		}

		return theList;
	}


	list<string> getAllParameterGroupNames() {
		DEBUG_MSG_FUNCTIONCALL("getAllParameterGroupNames()");

		list<string> parameterGroupNames;

		for(map<string, TiXmlElement*>::iterator it = __m_parameterGroupCacheByName.begin(); it != __m_parameterGroupCacheByName.end(); ++it) {
			parameterGroupNames.push_back(it->first);
		}

		return parameterGroupNames;
	}


	list<string> getAllParameterGroupElements(const string& p_groupName) {
		DEBUG_MSG_FUNCTIONCALL("getAllParameterGroupElements()");

		list<string> groupElements;

		TiXmlElement* groupNode = __m_parameterGroupCacheByName[p_groupName];
		for (TiXmlElement* currNode = groupNode->FirstChildElement(); currNode != 0; currNode = currNode->NextSiblingElement()) {
			if (currNode->ValueStr() == "element") {
				groupElements.push_back(currNode->GetText());
			}
		}

		return groupElements;
	}


	/** @brief return all application parameter names */
	list<string> getAllApplicationParameterNames() {
		DEBUG_MSG_FUNCTIONCALL("getAllApplicationParameterNames()");

		list<string> theList;

		for(map<string, TiXmlElement*>::iterator it = __m_appParameterCacheByName.begin(); it != __m_appParameterCacheByName.end(); ++it) {
			theList.push_back(it->first);
		}

		return theList;
	}


	/** @brief return all node types occuring in the graph */
	list<string> getAllNodeTypes() {
		DEBUG_MSG_FUNCTIONCALL("getAllNodeTypes()");

		list<string> theList;

		for (map<string, list<TiXmlElement*> >::iterator it = __m_nodeCacheByType.begin(); it != __m_nodeCacheByType.end(); it++) {
			theList.push_back(it->first);
		}

		return theList;
	}


	/** @brief get a list of all dynamic attributes for a specific node */
	list<string> getAllDynamicAttributesForNode(const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getAllDynamicAttributesForNode(" << p_nodeName << ")");

		list<string> theList;

		for (TiXmlElement* currElem = getNodeByName(p_nodeName)->FirstChildElement(); currElem != 0; currElem = currElem->NextSiblingElement()) {

			if (isAttributeDynamic(p_nodeName, currElem->ValueStr())) {
				theList.push_back(currElem->ValueStr());
			}

		}

		return theList;
	}


	/** @brief get a list of all static attributes for a specific node */
	list<string> getAllStaticAttributesForNode(const string& p_nodeName) {
		DEBUG_MSG_FUNCTIONCALL("getAllStaticAttributesForNode(" << p_nodeName << ")");

		list<string> theList;

		for (TiXmlElement* currElem = getNodeByName(p_nodeName)->FirstChildElement(); currElem != 0; currElem = currElem->NextSiblingElement()) {

			if (isAttributeStatic(p_nodeName, currElem->ValueStr())) {
				theList.push_back(currElem->ValueStr());
			}

		}

		return theList;
	}


	/** @brief return the number of nodes in the graph */
	size_t getNumberOfNodes() {
		DEBUG_MSG_FUNCTIONCALL("getNumberOfNodes()");

		return __m_nodeCacheByName.size();
	}


	/** @brief if an attribute is dynamic, it can be determined at runtime :) */
	bool isAttributeDynamic(const string& p_nodeName, const string& p_parameter) {
		DEBUG_MSG_FUNCTIONCALL("isAttributeDynamic(" << p_nodeName << ", " << p_parameter << ")");

		return getAttributeValue(p_nodeName, p_parameter, "mutability") == "dynamic";
	}


	/** @brief if an attribute is static, it is determined by reading it from the graph description */
	bool isAttributeStatic(const string& p_nodeName, const string& p_parameter) {
		DEBUG_MSG_FUNCTIONCALL("isAttributeStatic(" << p_nodeName << ", " << p_parameter << ")");

		return getAttributeValue(p_nodeName, p_parameter, "mutability") == "static";
	}


	/** @brief get attribute from a node, where the node is described by a path */
	string getAttributeValue(const string& p_nodeName, const string& p_parameter, const string& p_attribute) {
		DEBUG_MSG_FUNCTIONCALL("getAttributeValue(" << p_nodeName << ", " << p_parameter << ", " << p_attribute << ")");

		
		const string* attVal = 0;
		
		if (p_parameter != "") {
			attVal = (getChildNodeByName(getNodeByName(p_nodeName), p_parameter)->Attribute(p_attribute));
		} else {
			attVal = getNodeByName(p_nodeName)->Attribute(p_attribute);
		}

		return (attVal != 0) ? *attVal : "";
	}
	

	/** @brief get attribute from an application parameter, simply given by a name */
	string getApplicationParameterAttribute(const string& p_parameterName, const string& p_attribute) {
		DEBUG_MSG_FUNCTIONCALL("getApplicationParameterAttribute(" << p_parameterName << ", " << p_attribute << ")");

		const string* attVal = __m_appParameterCacheByName[p_parameterName]->Attribute(p_attribute);

		return (attVal != 0) ? *attVal : "";
	}


	/** @brief get node type by node name */
	void getNodeTypeByName(const string& p_nodeName, string& p_data) {
		DEBUG_MSG_FUNCTIONCALL("getNodeTypeByName(" << p_nodeName << ")");

		p_data = getNodeFromCache(p_nodeName)->ValueStr();
	}


	/** @brief name of the data type for a given i/o node can be requested with this method */
	void getDataTypeName(const string& p_nodeName, const string& p_elementName, string& p_data) {
		DEBUG_MSG_FUNCTIONCALL("getDataTypeName(" << p_nodeName << ", " << p_elementName << ", ...)");

		TiXmlElement* theNode;
		if ((theNode = getChildNodeByName(getNodeFromCache(p_nodeName), p_elementName)) != 0) {
			p_data = theNode->FirstChild()->ValueStr();
		}
	}

	/**
	 * @brief different data element types from the graph can directly be used as the specific formats
	 * Ensure that you only call this method on nodes which directly contain data!
	 */
	void getDataFromNode(const string& p_nodeName, const string& p_elementName, float& p_data) {
		DEBUG_MSG_FUNCTIONCALL("getDataFromNode(" << p_nodeName << ", " << p_elementName << ", ...)");

		TiXmlElement* theNode;

		if ((theNode = getChildNodeByName(getNodeFromCache(p_nodeName), p_elementName)) != 0) {
			string nodeValue = theNode->GetText();
			p_data = (float) atof(nodeValue.c_str());
		}
	}

	/**
	 * @brief different data element types from the graph can directly be used as the specific formats
	 * Ensure that you only call this method on nodes which directly contain data!
	 */
	void getDataFromNode(const string& p_nodeName, const string& p_elementName, int32_t& p_data) {
		DEBUG_MSG_FUNCTIONCALL("getDataFromNode(" << p_nodeName << ", " << p_elementName << ", ...)");

		TiXmlElement* theNode;
		if ((theNode = getChildNodeByName(getNodeFromCache(p_nodeName), p_elementName)) != 0) {
			string nodeValue = theNode->GetText();
			p_data = atoi(nodeValue.c_str());
		}
	}

	/**
	 * @brief different data element types from the graph can directly be used as the specific formats
	 * Ensure that you only call this method on nodes which directly contain data!
	 */
	void getDataFromNode(const string& p_nodeName, const string& p_elementName, uint32_t& p_data) {
		DEBUG_MSG_FUNCTIONCALL("getDataFromNode(" << p_nodeName << ", " << p_elementName << ", ...)");

		TiXmlElement* theNode;
		if ((theNode = getChildNodeByName(getNodeFromCache(p_nodeName), p_elementName)) != 0) {
			string nodeValue = theNode->GetText();
			p_data = atoi(nodeValue.c_str());
		}
	}

	/**
	 * @brief different data element types from the graph can directly be used as the specific formats
	 * Ensure that you only call this method on nodes which directly contain data!
	 */
	void getDataFromNode(const string& p_nodeName, const string& p_elementName, string& p_data) {
		DEBUG_MSG_FUNCTIONCALL("getDataFromNode(" << p_nodeName << ", " << p_elementName << ", ...)");

		TiXmlElement* theNode;
		if ((theNode = getChildNodeByName(getNodeFromCache(p_nodeName), p_elementName)) != 0) {
			p_data = theNode->GetText();
		}
	}

	/** @brief check if wrapper is initialized */
	bool isInitialized() {
		return __m_initialized;
	}

	/** @brief this is just an auxiliary method printing all necessary input parameters. Works only if the node name conventions are respected! */
	void printNecessaryInputParameters(vector<string>& p_orderedNames) {
		DEBUG_MSG_FUNCTIONCALL("printInputParameters()");

		list<string> nip;

		for (uint32_t i = 0; i < p_orderedNames.size(); ++i) {
			TiXmlElement* currentNode = getNodeByName(p_orderedNames[i]);

			//traverse current node
			for (TiXmlElement* innerNode = currentNode->FirstChildElement(); innerNode != 0; innerNode = innerNode->NextSiblingElement()) {
				if (innerNode->ValueStr().substr(0, 5) == "input") {

					string inputElementName = innerNode->ValueStr();

					string dataTypeName;
					getDataTypeName(p_orderedNames[i], inputElementName, dataTypeName);

					string identifier = getChildNodeByName(innerNode, dataTypeName)->GetText();

					string fullQualifier = dataTypeName + ", " + identifier;
					string fullInputInfo = p_orderedNames[i] + ": " + inputElementName + ", " + fullQualifier;

					bool outputBefore = false;

					// check if this qualifier is provided by an earlier processed node
					for (int j = i; j >= 0; --j) {
						TiXmlElement* reverseCurrentNode = getNodeByName(p_orderedNames[j]);

						// also generate qualifier here, so it can be compared
						for (TiXmlElement* reverseInnerNode = reverseCurrentNode->FirstChildElement(); reverseInnerNode != 0; reverseInnerNode = reverseInnerNode->NextSiblingElement()) {

							if (reverseInnerNode->ValueStr().substr(0, 6) == "output") {

								string reverseInputElementName = reverseInnerNode->ValueStr();
								string reverseDataTypeName;
								getDataTypeName(p_orderedNames[j], reverseInputElementName, reverseDataTypeName);

								string reverseIdentifier = getChildNodeByName(reverseInnerNode, reverseDataTypeName)->GetText();

								string reverseFullQualifier = reverseDataTypeName + ", " + reverseIdentifier;

								if (reverseFullQualifier == fullQualifier) {
									//cout << "REMOVING INPUT: " << fullInputInfo << endl;
									outputBefore = true;
									goto endOfLoop;
								}
							}
						}
					}

					endOfLoop:

					if (!outputBefore) {
						nip.push_back(fullInputInfo);
					}
				}
			}
		}

		cout << "*****NECESSARY INPUT: " << endl;

		for (list<string>::iterator it = nip.begin(); it != nip.end(); it++) {
			cout << *it << endl;
		}

	}

};

}

#endif /* GRAPHWRAPPER_H_ */
