/*
 * NodeHelper.h
 *
 *  Created on: 10.11.2010
 *      Author: troth
 *
 * changelog:
 * 	2011-12-23:
 * 		- Changed GRIP_GET_STATIC_OR_DYNAMIC_ATTRIBUTE to be a function, using the same backwards compatibility mechanism as in CudaHelper.h
 *
 */

#ifndef NODEHELPER_H_
#define NODEHELPER_H_

#include <Utils/Common.h>
#include <Data/UniversalContainer.h>
#include <string>

// the following macros rely on standardized code...see what they generate for further information

/**
 * this is used to get a filter attribute, which can be static or dynamic and has to be fetched
 * from the GraphWrapper or the UniversalContainer accordingly
 */
#define GRIP_GET_STATIC_OR_DYNAMIC_ATTRIBUTE(attName, attTarget, attType) \
	grip::gripGetStaticOrDynamicAttribute<attType>(p_gw, p_nodeName, attName, attTarget, p_uniCon);


/**
 * get image from universal container and set parameters for CUDA processing
 */
#define GRIP_GET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(datatype, deviceType, varName, uniconName, parmName) \
	datatype varName = p_uniCon.getObject<datatype>(uniconName); \
	parmName##_h = varName->getDataPointer().get(); \
	parmName##_d = p_uniCon.getDevicePtr<deviceType>(uniconName); \
	parmName##_dirty = p_uniCon.isDirty(uniconName);

/**
 * create a newly allocated target image and set the according parameters for CUDA processing
 */
#define GRIP_CREATE_TARGET_IMAGE_AND_SET_PARAMETERS_FOR_CUDA(type_basename, deviceType, resx, resy, varName, uniconName, parmName) \
	type_basename##_sptr varName; \
	if (!p_uniCon.objectExists(uniconName)) { \
		varName = type_basename##_sptr(new type_basename##_t(resx, resy)); \
	} else { \
		varName = p_uniCon.getObject<type_basename##_sptr>(uniconName); \
	} \
	parmName##_h = varName->getDataPointer().get(); \
	parmName##_d = p_uniCon.getDevicePtr<deviceType>(uniconName); \
	parmName##_dirty = p_uniCon.isDirty(uniconName);

/**
 * update dirty flag and device pointer in universal container
 */
#define GRIP_UPDATE_DIRTY_FLAG_AND_DEVICE_PTR(name, parmName, deviceType) \
	p_uniCon.setDirty(name, parmName##_dirty); \
	p_uniCon.setDevicePtr<deviceType>(name, parmName##_d);

namespace grip {

template<class FTYPE> INLINE void gripGetStaticOrDynamicAttribute(grip::GraphWrapper& p_gw, const string& p_nodeName, const string& p_attName, FTYPE& p_attTarget, grip::UniversalContainer& p_uniCon) {
	DEBUG_MSG_FUNCTIONCALL("gripGetStaticOrDynamicAttribute(" << &p_gw << ", " << p_nodeName << ", " << p_attName << ", " << &p_attTarget << ", " << &p_uniCon << ")");

	if (p_gw.isAttributeDynamic(p_nodeName, p_attName)) {
		string parameterName;
		p_gw.getDataFromNode(p_nodeName, p_attName, parameterName);
		p_attTarget = p_uniCon.getObject<FTYPE> (parameterName);
	} else {
		p_gw.getDataFromNode(p_nodeName, p_attName, p_attTarget);
	}

}

#undef CLASSNAME
#define CLASSNAME NodeHelper

/**
 * @class NodeHelper
 * @brief provides some auxiliary methods for handling data in processing node implementations
 */
class CLASSNAME {
private:
	string __m_nodeName;
	GraphWrapper __m_graphWrapper;
	UniversalContainer __m_universalContainer;

public:

	/** @brief create a new NodeHelper for a given nodeName, using the respective GraphWrapper and UniversalContainer instances */
	CLASSNAME(const string& p_nodeName, GraphWrapper& p_gw, UniversalContainer& p_uniCon) {
		DEBUG_MSG_FUNCTIONCALL("NodeHelper(" << p_nodeName << ", ...)");

		__m_nodeName = p_nodeName;
		__m_graphWrapper = p_gw;
		__m_universalContainer = p_uniCon;

	}

	/** @brief standard destructor */
	~CLASSNAME() {
		DEBUG_MSG_FUNCTIONCALL("~NodeHelper()");
	}

	/** @brief get I/O type (e.g. image_float_4) and the according identifier (e.g. input_image) */
	void getIOTypeAndIdentifier(const string& p_inputNodeName, string& p_type, string& p_identifier) {
		DEBUG_MSG_FUNCTIONCALL("getIOTypeAndIdentifier(" << p_inputNodeName << ", " << p_type << ", " << p_identifier << ")");

		__m_graphWrapper.getDataTypeName(__m_nodeName, p_inputNodeName, p_type);
		__m_graphWrapper.getDataFromNode(__m_nodeName, p_inputNodeName + "/" + p_type, p_identifier);
	}

	/** @brief call this if the passed types are inequal */
	void errorInequalTypes() {
		DEBUG_MSG_WARNING(__m_nodeName << ": ERROR, Inequal types!");
	}

	/** @brief call this if an unsupported format has been passed */
	void errorUnsupportedFormat() {
		DEBUG_MSG_WARNING(__m_nodeName << ": ERROR, Unsupported format!");
	}

};

}

#endif /* NODEHELPER_H_ */
