/*
 * NodeLoader.h
 *
 *  Created on: 06.11.2010
 *      Author: troth
 */

#ifndef NODELOADER_H_
#define NODELOADER_H_

/**
 * This file provides means to access processing nodes by name, which are implemented
 * as external shared objects.
 *
 * - e.g.: implementation of node "node_convolution" is available in libnode_convolution.so
 * - main method of the processing node implementation is extern "C" run_node(UniversalContainer& p_uniCon)
 * - the UniversalContainer is used to pass data between nodes
 * - template parameterization is done inside nodes
 */

#include <iostream>
#include <map>

#include <Data/UniversalContainer.h>
#include <Graph/FunctionTypes.h>

#include <dlfcn.h>

	namespace grip {

	using namespace std;

	#undef CLASSNAME
	#define CLASSNAME NodeLoader



	/**
	 * @class NodeLoader
	 * @brief This class takes care of dynamic implementation loading for processing nodes.
	 *
	 * Handles for dynamically loaded objects are handeled (lol!) internally with some kind of "handle cache",
	 * so the user of this class does not need to call dlclose() each time. This is done when the NodeLoader
	 * object is destroyed (see method closeHandles()).
	 */
	class CLASSNAME {
	private:
		map<string, void*> __m_handles;
		std::string        __m_path_prefix;
		std::string		   __m_path_directory;
		std::string		   __m_lib_prefix;
		std::string		   __m_lib_suffix;

		/** @brief close all handles which have been registered */
		void closeHandles() {
			DEBUG_MSG_FUNCTIONCALL("closeHandles()");

			for (map<string, void*>::iterator it = __m_handles.begin(); it != __m_handles.end(); it++) {
				dlclose(it->second);
			}
		}

		/** @brief get main interface handle for a specific node name */
		void getHandle(const std::string& p_nodeName, void*& p_handle) {
			DEBUG_MSG_FUNCTIONCALL("getHandle(" << p_nodeName << ", " << p_handle << ")");

			if (__m_handles.find(p_nodeName) == __m_handles.end()) { // handle has to be loaded first
				string pathToNode = "/home/troth/cg-lab/lib/";
				pathToNode += __m_path_prefix;
				pathToNode += __m_path_prefix != "" ? "/" : "";
				pathToNode += __m_path_directory;
				pathToNode += __m_path_directory != "" ? "/" : "";
				pathToNode += __m_lib_prefix + p_nodeName + __m_lib_suffix;

//				pathToNode = __m_path_prefix + "/" + __m_path_directory + "/" + __m_lib_prefix + p_nodeName + __m_lib_suffix;
				void* handle = dlopen(pathToNode.c_str(), RTLD_LAZY);

				if (!handle) {
					cerr << "Failed opening: " << dlerror() << endl;
					exit(1);
				}

				__m_handles[p_nodeName] = handle;
			}

			p_handle = __m_handles[p_nodeName];
		}

	public:
		/** @brief default constructor */
		CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("NodeLoader()");

		}

		/** @brief default destructor */
		~CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("~NodeLoader()");

			closeHandles();
		}

		void setPathAndLibraryInformation(const std::string& p_pathPrefix, const std::string& p_subDir, const std::string& p_libPrefix, const std::string& p_libSuffix) {
			DEBUG_MSG_FUNCTIONCALL("setPathAndLibraryInformation(" << p_pathPrefix << ", " << p_subDir << "," << p_libPrefix << ", " << p_libSuffix << ")");

			std::cout << "NodeLoader setting: " << p_pathPrefix << ", " << p_subDir << ", " << p_libPrefix << ", " << p_libSuffix << std::endl;

			__m_path_prefix = p_pathPrefix;
			__m_path_directory = p_subDir;
			__m_lib_prefix = p_libPrefix;
			__m_lib_suffix = p_libSuffix;
		}


		/** @brief by utilizing templates, this can be used to load arbitrary symbols from the object */
		template<class FTYPE> FTYPE loadNodeFunction(const std::string& p_nodeName, const std::string& p_funcName) {
			DEBUG_MSG_FUNCTIONCALL("loadNodeRunFunction(" << p_nodeName << ", " << p_funcName << ")");

			//TODO paths are not final
			void* handle = 0;
			getHandle(p_nodeName, handle);

			// load symbol
			// reset errors
			dlerror();

			FTYPE function = (FTYPE) dlsym(handle, p_funcName.c_str());

			const char* dlsym_error = dlerror();

			if (dlsym_error) {
				cerr << "Cannot load symbol: " << dlsym_error << endl;
				//exit(1);
				return 0;
			}

			return function;

		}
	};


}

#endif /* NODELOADER_H_ */
