/*
 * UniversalImage.h
 *
 *  Created on: 30.06.2010
 *      Author: troth
 */


#ifndef UNIVERSALIMAGE_H_
#define UNIVERSALIMAGE_H_

#include <Utils/NoDeleteDeleter.h>
#include <Data/DataTypes_primitive.h>
#include <Image/UniversalImageBase.h>
#include <tr1/memory>
#include <stdint.h>
#include <string>
#include <string.h>
#include <map>
#include <cuda_runtime.h>




#undef 	CLASSNAME
#define	CLASSNAME	UniversalImage

using namespace std;

/**
 * @class UniversalImage
 * @brief provides a universal image class, which can be used with arbitrary (supported) types
 *
 * This class provides data access methods. The passed template parameter has to be the base type, not the pointer type.
 */

namespace grip {

	template<class CTYPE, uint32_t CNUMCOMPONENTS> class CLASSNAME : public UniversalImageBase {
	private:
		tr1::shared_ptr<CTYPE> __m_data;

	public:
		typedef UniversalImageBase super;

		/** @brief default constructor, setting data to NULL and default orientation to left to right, top to bottom. No memory is allocated */
		INLINE CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("UniversalImage()");

			__m_data = tr1::shared_ptr<CTYPE>();

			// orientation is set in UniversalImageBase's constructor
		}

		/** @brief constructor taking a shared pointer to some data, as well as width and height */
		INLINE CLASSNAME(tr1::shared_ptr<CTYPE> p_data, uint32_t p_resx, uint32_t p_resy) : __m_data(p_data) {
			DEBUG_MSG_FUNCTIONCALL("UniversalImage(" << p_data << ", " << p_resx << ", " << p_resy << ")");

			setRes(p_resx, p_resy);
		}

		/** @brief constructor taking only width and height, no memory is allocated */
		INLINE CLASSNAME(uint32_t p_resx, uint32_t p_resy) {
			DEBUG_MSG_FUNCTIONCALL("UniversalImage(" << p_resx << ", " << p_resy << ")");

			setRes(p_resx, p_resy);
		}


		/** @brief default destructor */
		INLINE virtual ~CLASSNAME() {
			DEBUG_MSG_FUNCTIONCALL("~UniversalImage()");

		}


		/** @brief returns shared pointer to data */
		INLINE tr1::shared_ptr<CTYPE> getDataPointer() {
			DEBUG_MSG_FUNCTIONCALL("getDataPointer()");

			return __m_data;
		}


		/** @brief returns raw pointer to passed x/y position */
		INLINE CTYPE* getPointerToCoordinates(uint32_t p_x, uint32_t p_y) {
			DEBUG_MSG_FUNCTIONCALL("getPointerToCoordinates(" << p_x << ", " << p_y << ")");

			return &(__m_data.get()[getResX() * p_y * CNUMCOMPONENTS + p_x * CNUMCOMPONENTS]);
		}


		/**	@brief returns raw pointer to corresponding 1d-array position */
		INLINE CTYPE* getPointerToArrayPosition(uint32_t p_pos) {
			DEBUG_MSG_FUNCTIONCALL("getPointerToArrayPosition(" << p_pos << ")");

			return &(__m_data.get()[p_pos * CNUMCOMPONENTS]);
		}


		/** @brief set whole image data to 0 */
		INLINE void clearData() {
			DEBUG_MSG_FUNCTIONCALL("clearData()");

			if (__m_data == NULL) {
				
				allocateMemory();
			}
			
			memset(__m_data.get(), 0, sizeof(CTYPE) * CNUMCOMPONENTS * getResX() * getResY() );
		}


		/** @brief Allocate memory for the image. Depends on template parameters and the resolution */
		INLINE void allocateMemory() {
			DEBUG_MSG_FUNCTIONCALL("allocateMemory()");

			if (__m_data != NULL) {
				DEBUG_MSG_RUNTIMEINFO("WARNING: Data already allocated -> Overwriting!");
			}
			
			CTYPE* thePtr = 0;

			cudaMalloc(&thePtr, CNUMCOMPONENTS * getResX() * getResY() * sizeof(CTYPE));

			__m_data = tr1::shared_ptr<CTYPE>(thePtr);

		}


		/** @brief returns the total size in bytes */
		size_t getTotalSizeInBytes() {
			DEBUG_MSG_FUNCTIONCALL("getTotalSizeInBytes()");

			return sizeof(CTYPE) * CNUMCOMPONENTS * getResX() * getResY();
		}


		/* @brief data will be copied from the shared pointer p_data to __m_data, __m_data will be allocated to the needed size if necessary
		 * There is no check if the allocated memory has the right size if __m_data has already been initialized! */
		INLINE void setData(tr1::shared_ptr<CTYPE> p_data, size_t p_size) {
			DEBUG_MSG_FUNCTIONCALL("setData(" << p_data << ", " << p_size << ")");

			if (p_data == NULL) {
				DEBUG_MSG_WARNING("setData(), WARNING: p_data is NULL!");
			}

			if (__m_data != NULL) {
				DEBUG_MSG_WARNING("setData(), WARNING: __m_data is not NULL, will be overwritten, size is not checked!");
			}
			CTYPE* thePtr = 0;

			cudaMalloc(&thePtr, CNUMCOMPONENTS * p_size * sizeof(CTYPE));

			__m_data = tr1::shared_ptr<CTYPE>(thePtr);

			DEBUG_MSG_RUNTIMEINFO("COPYING " << (p_size * sizeof(CTYPE) * CNUMCOMPONENTS) << " Bytes");

			memcpy(__m_data.get(), p_data.get(), p_size * sizeof(CTYPE) * CNUMCOMPONENTS);
		}


		/** @brief data will be copied from the raw pointer p_data to __m_data. Again, no sanity taking size into account. */
		INLINE void setData(CTYPE* p_data, size_t p_size) {
			DEBUG_MSG_FUNCTIONCALL("setData(" << p_data << ", " << p_size << ")");

			if (p_data == NULL) {
				DEBUG_MSG_WARNING("setData(), WARNING: p_data is NULL!");
			}

			if (__m_data != NULL) {
				DEBUG_MSG_WARNING("setData(), WARNING: __m_data is not NULL, will be overwritten, size is not checked!");
			}

			CTYPE* thePtr = 0;

			cudaMalloc(&thePtr, CNUMCOMPONENTS * p_size * sizeof(CTYPE));


			__m_data = tr1::shared_ptr<CTYPE>(thePtr);

			DEBUG_MSG_RUNTIMEINFO("COPYING " << (p_size * sizeof(CTYPE) * CNUMCOMPONENTS) << " Bytes");

			memcpy(__m_data.get(), p_data, p_size * sizeof(CTYPE) * CNUMCOMPONENTS);

		}


		/**
		 * @brief set data to raw pointer, no copying is performed!
		 * WARNING: note that the passed pointer is handeled by the internal shared pointer after it has been bound!
		 */
		INLINE void bindData(CTYPE* p_data) {
			DEBUG_MSG_FUNCTIONCALL("bindData(" << p_data << ")");

			if (p_data == NULL) {
				DEBUG_MSG_WARNING("bindData(), WARNING: p_data is NULL!");
			}

			if (__m_data != NULL) {
				DEBUG_MSG_WARNING("bindData(), WARNING: __m_data is not NULL, will be overwritten!");
			}

			__m_data = tr1::shared_ptr<CTYPE>(p_data);
		}


		/**
		 * @brief set data to raw pointer, no copying is performed, nothing is done when reaching zero references!
		 * WARNING: note that the passed pointer is _not_ freed by the internal shared pointer after it has been bound!
		 */
		INLINE void bindDataNoDelete(CTYPE* p_data) {
			DEBUG_MSG_FUNCTIONCALL("bindDataNoDelete(" << p_data << ")");

			if (p_data == NULL) {
				DEBUG_MSG_WARNING("bindData(), WARNING: p_data is NULL!");
			}

			if (__m_data != NULL) {
				DEBUG_MSG_WARNING("bindData(), WARNING: __m_data is not NULL, will be overwritten!");
			}

			__m_data = tr1::shared_ptr<CTYPE>(p_data, NoDeleteDeleter<CTYPE>());
		}



		/** @brief set data to given shared pointer */
		INLINE void bindData(tr1::shared_ptr<CTYPE> p_data) {
			DEBUG_MSG_FUNCTIONCALL("bindData(" << p_data << ")");

			if (p_data == NULL) {
				DEBUG_MSG_WARNING("bindData(), WARNING: p_data is NULL!");
			}

			if (__m_data != NULL) {
				DEBUG_MSG_WARNING("bindData(), WARNING: __m_data is not NULL, will be overwritten!");
			}

			__m_data = p_data;
		}



		/** @brief set data at x/y coordinates to that in p_data. p_data should at least contain sizeof(CTYPE)*CNUMCOMPONENTS bytes. */
		INLINE void setDataAtCoordinates(uint32_t p_x, uint32_t p_y, CTYPE* p_data) {
			DEBUG_MSG_FUNCTIONCALL("setDataAtCoordinates(" << p_x << ", " << p_y << ", " << p_data << ")");

			uint32_t startIndex = __m_resx * p_y * CNUMCOMPONENTS + p_x * CNUMCOMPONENTS;
			memcpy(&(__m_data[startIndex]), p_data, sizeof(CTYPE) * CNUMCOMPONENTS);
		}


		/** @brief set data at array position to that in p_data. p_data should at least contain sizeof(CTYPE)*CNUMCOMPONENTS bytes. */
		INLINE void setDataAtArrayPosition(uint32_t p_pos, CTYPE* p_data) {
			DEBUG_MSG_FUNCTIONCALL("setDataAtPosition(" << p_pos << ", " << p_data << ")");

			uint32_t startIndex = p_pos * CNUMCOMPONENTS;
			memcpy(&(__m_data.get()[startIndex]), p_data, sizeof(CTYPE) * CNUMCOMPONENTS);
		}


		/** @brief changes orientation from l(eft) t(o) r(ight)-t(op) t(o) b(ottom) to l(eft) t(o) r(ight)-b(ottom) t(o) t(op);
		    remember that only these two orientations are supported */
		INLINE void changeOrientation(uint32_t p_newOrientation) {
			DEBUG_MSG_FUNCTIONCALL("changeOrientation(" << p_newOrientation << ")");

			UniversalImage<CTYPE, CNUMCOMPONENTS> tempImage;
			tempImage.setRes(getResX(), getResY());
			tempImage.setData(__m_data, getResX() * getResY());

			// copy it back in the right order for the new orientation
			uint32_t srcPos = getResX() * getResY() - getResX();
			for (uint32_t i = 0; i < getResX() * getResY(); ++i) {
				this->setDataAtArrayPosition(i, tempImage.getPointerToArrayPosition(srcPos));
				srcPos -= ((i + 1) % (getResX()) == 0) ? 2 * getResX() - 1 : -1;
			}

			setOrientation(p_newOrientation);
		}


		/** @brief output function for image contents */
		INLINE void print() {
			DEBUG_MSG_FUNCTIONCALL("print()");

			for (uint32_t i = 0; i < getResX() * getResY() * CNUMCOMPONENTS; ++i) {
				for (uint32_t j = 0; j < CNUMCOMPONENTS; ++j) {
					cout << "\t" << getPointerToArrayPosition(i)[j];
				}
				if ((i + 1) % (getResX() * CNUMCOMPONENTS) == 0) {
					cout << endl;
				}
			}
		}

	};

}

#endif /* UNIVERSALIMAGE_H_ */
