#include <cuda_runtime.h>
#include <iostream>
//#include <cutil.h>
//#include <cutil_inline.h>
#include <cmath>

#include <Utils/CudaHelper.h>
#include <Utils/TimerMacros.h>
#include <helper_cuda.h>
#include <helper_math.h>

#include "parameters.h"

#define BLOCK_DIM 16
#define CLAMP_TO_EDGE 0

typedef struct __PARAMETERS_FLOAT {
	float intensity;
	int resx;
	int resy;
} parm_float;

#undef PARM
#define PARM parm_float_c
__device__ __constant__ parm_float PARM;

int iDivUp(int a, int b) {
	return (a % b != 0) ? (a / b + 1) : (a / b);
}

// Transpose kernel (see transpose SDK sample for details)
__global__ void d_transpose(float4 *odata, float4 *idata, int width, int height) {
	__shared__ float4 block[BLOCK_DIM][BLOCK_DIM + 1];

	// read the matrix tile into shared memory
	unsigned int xIndex = blockIdx.x * BLOCK_DIM + threadIdx.x;
	unsigned int yIndex = blockIdx.y * BLOCK_DIM + threadIdx.y;

	if ((xIndex < width) && (yIndex < height)) {
		unsigned int index_in = yIndex * width + xIndex;
		block[threadIdx.y][threadIdx.x] = idata[index_in];
	}

	__syncthreads();

	// write the transposed matrix tile to global memory
	xIndex = blockIdx.y * BLOCK_DIM + threadIdx.x;
	yIndex = blockIdx.x * BLOCK_DIM + threadIdx.y;

	if ((xIndex < height) && (yIndex < width)) {
		unsigned int index_out = yIndex * height + xIndex;
		odata[index_out] = block[threadIdx.x][threadIdx.y];
	}
}



// Transpose kernel (see transpose SDK sample for details)
__global__ void d_transpose(float *odata, float *idata, int width, int height) {
	__shared__ float block[BLOCK_DIM][BLOCK_DIM + 1];
	
	// read the matrix tile into shared memory
	unsigned int xIndex = blockIdx.x * BLOCK_DIM + threadIdx.x;
	unsigned int yIndex = blockIdx.y * BLOCK_DIM + threadIdx.y;
	
	if ((xIndex < width) && (yIndex < height)) {
		unsigned int index_in = yIndex * width + xIndex;
		block[threadIdx.y][threadIdx.x] = idata[index_in];
	}
	
	__syncthreads();
	
	// write the transposed matrix tile to global memory
	xIndex = blockIdx.y * BLOCK_DIM + threadIdx.x;
	yIndex = blockIdx.x * BLOCK_DIM + threadIdx.y;
	
	if ((xIndex < height) && (yIndex < width)) {
		unsigned int index_out = yIndex * height + xIndex;
		odata[index_out] = block[threadIdx.x][threadIdx.y];
	}
}


void transpose(float4 *d_src, float4 *d_dest, uint width, int height) {
	dim3 grid(iDivUp(width, BLOCK_DIM), iDivUp(height, BLOCK_DIM), 1);
	dim3 threads(BLOCK_DIM, BLOCK_DIM, 1);
	d_transpose<<< grid, threads >>>(d_dest, d_src, width, height);
	getLastCudaError("Kernel execution failed");
}



void transpose(float *d_src, float *d_dest, uint width, int height) {
	dim3 grid(iDivUp(width, BLOCK_DIM), iDivUp(height, BLOCK_DIM), 1);
	dim3 threads(BLOCK_DIM, BLOCK_DIM, 1);
	d_transpose<<< grid, threads >>>(d_dest, d_src, width, height);
	getLastCudaError("Kernel execution failed");
}

/*
 recursive Gaussian filter

 parameters:
 id - pointer to input data (RGBA image packed into 32-bit integers)
 od - pointer to output data
 w  - image width
 h  - image height
 a0-a3, b1, b2, coefp, coefn - filter parameters
 */


__global__ void d_recursiveGaussian_monochrome(float *id, float *od, int w, int h, float a0, float a1, float a2, float a3, float b1, float b2, float coefp, float coefn) {
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
	
	if (x >= w)
		return;
	
	id += x; // advance pointers to correct column
	od += x;
	
	// forward pass
	float xp = 0.f;
	float yp = 0.f;
	float yb = 0.f;
	#if CLAMP_TO_EDGE
	xp = *id;
	yb = coefp * xp;
	yp = yb;
	#endif
	
	for (int y = 0; y < h; y++) {
		float xc = *id;
		float yc = a0 * xc + a1 * xp - b1 * yp - b2 * yb;
		*od = yc;
		id += w;
		od += w; // move to next row
		xp = xc;
		yb = yp;
		yp = yc;
	}
	
	// reset pointers to point to last element in column
	id -= w;
	od -= w;
	
	// reverse pass
	// ensures response is symmetrical
	float xn = 0.f;//make_float4(0.0f);
	float xa = 0.f;//make_float4(0.0f);
	float yn = 0.f;//make_float4(0.0f);
	float ya = 0.f;//make_float4(0.0f);
	#if CLAMP_TO_EDGE
	xn = xa = *id;
	yn = coefn * xn;
	ya = yn;
	#endif
	
	for (int y = h - 1; y >= 0; y--) {
		float xc = *id;
		float yc = a2 * xn + a3 * xa - b1 * yn - b2 * ya;
		xa = xn;
		xn = xc;
		ya = yn;
		yn = yc;
		*od = *od + yc;
		id -= w;
		od -= w; // move to previous row
	}
}


__global__ void d_recursiveGaussian_rgba(float4 *id, float4 *od, int w, int h, float a0, float a1, float a2, float a3, float b1, float b2, float coefp, float coefn) {
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;

	if (x >= w)
		return;

	id += x; // advance pointers to correct column
	od += x;

	// forward pass
	float4 xp = make_float4(0.0f); // previous input
	float4 yp = make_float4(0.0f); // previous output
	float4 yb = make_float4(0.0f); // previous output by 2
#if CLAMP_TO_EDGE
	xp = *id;
	yb = coefp * xp;
	yp = yb;
#endif

	for (int y = 0; y < h; y++) {
		float4 xc = *id;
		float4 yc = a0 * xc + a1 * xp - b1 * yp - b2 * yb;
		*od = yc;
		id += w;
		od += w; // move to next row
		xp = xc;
		yb = yp;
		yp = yc;
	}

	// reset pointers to point to last element in column
	id -= w;
	od -= w;

	// reverse pass
	// ensures response is symmetrical
	float4 xn = make_float4(0.0f);
	float4 xa = make_float4(0.0f);
	float4 yn = make_float4(0.0f);
	float4 ya = make_float4(0.0f);
#if CLAMP_TO_EDGE
	xn = xa = *id;
	yn = coefn * xn;
	ya = yn;
#endif

	for (int y = h - 1; y >= 0; y--) {
		float4 xc = *id;
		float4 yc = a2 * xn + a3 * xa - b1 * yn - b2 * ya;
		xa = xn;
		xn = xc;
		ya = yn;
		yn = yc;
		*od = *od + yc;
		id -= w;
		od -= w; // move to previous row
	}
}


__global__ void d_intensityAdd_monochrome(float* id, float* od, int h, int w) {
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
	
	if (x >= w)
		return;
	
	id += x; // advance pointers to correct column
	od += x;
	
	for (int y = 0; y < h; y++) {
		*od = (PARM.intensity * (*od) +  (*id)) / (1.f + PARM.intensity);
		
		
		id += w;
		od += w;
		
	}
}


__global__ void d_intensityAdd(float4* id, float4* od, int h, int w) {
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;

	if (x >= w)
		return;

	id += x; // advance pointers to correct column
	od += x;

	for (int y = 0; y < h; y++) {
		*od = (PARM.intensity * (*od) +  (*id)) / (1.f + PARM.intensity);


		id += w;
		od += w;

	}
}


// 8-bit RGBA version
extern "C" void gaussianFilterMonochrome(float *d_src, float *d_dest, float *d_temp, int width, int height, float sigma, int order, int nthreads) {
	// compute filter coefficients
	const float nsigma = sigma < 0.1f ? 0.1f : sigma, alpha = 1.695f / nsigma, ema = (float) std::exp(-alpha), ema2 = (float) std::exp(-2 * alpha), b1 = -2 * ema, b2 = ema2;
	
	float a0 = 0, a1 = 0, a2 = 0, a3 = 0, coefp = 0, coefn = 0;
	
	switch (order) {
		case 0: {
			const float k = (1 - ema) * (1 - ema) / (1 + 2 * alpha * ema - ema2);
			a0 = k;
			a1 = k * (alpha - 1) * ema;
			a2 = k * (alpha + 1) * ema;
			a3 = -k * ema2;
		}
		break;
		
		case 1: {
			const float k = (1 - ema) * (1 - ema) / ema;
			a0 = k * ema;
			a1 = a3 = 0;
			a2 = -a0;
		}
		break;
		
		case 2: {
			const float ea = (float) std::exp(-alpha), k = -(ema2 - 1) / (2 * alpha * ema), kn = (-2 * (-1 + 3 * ea - 3 * ea * ea + ea * ea * ea) / (3 * ea + 1 + 3 * ea * ea + ea * ea * ea));
			a0 = kn;
			a1 = -kn * (1 + k * alpha) * ema;
			a2 = kn * (1 - k * alpha) * ema;
			a3 = -kn * ema2;
		}
		break;
		
		default:
			fprintf(stderr, "gaussianFilter: invalid order parameter!\n");
			return;
	}
	
	coefp = (a0 + a1) / (1 + b1 + b2);
	coefn = (a2 + a3) / (1 + b1 + b2);
	
	// process columns
	
	d_recursiveGaussian_monochrome<<< iDivUp(width, nthreads), nthreads >>>(d_src, d_temp, width, height, a0, a1, a2, a3, b1, b2, coefp, coefn);
	
	getLastCudaError("Kernel execution failed");
	
	transpose(d_temp, d_dest, width, height);
	getLastCudaError("transpose: Kernel execution failed");
	
	// process rows
	
	d_recursiveGaussian_monochrome<<< iDivUp(height, nthreads), nthreads >>>(d_dest, d_temp, height, width, a0, a1, a2, a3, b1, b2, coefp, coefn);
	
	getLastCudaError("Kernel execution failed");
	
	transpose(d_temp, d_dest, height, width);
	
	d_intensityAdd_monochrome<<< iDivUp(width, nthreads), nthreads >>>(d_src, d_dest, height, width);
}




// 8-bit RGBA version
extern "C" void gaussianFilterRGBA(float4 *d_src, float4 *d_dest, float4 *d_temp, int width, int height, float sigma, int order, int nthreads) {
	// compute filter coefficients
	const float nsigma = sigma < 0.1f ? 0.1f : sigma, alpha = 1.695f / nsigma, ema = (float) std::exp(-alpha), ema2 = (float) std::exp(-2 * alpha), b1 = -2 * ema, b2 = ema2;

	float a0 = 0, a1 = 0, a2 = 0, a3 = 0, coefp = 0, coefn = 0;

	switch (order) {
	case 0: {
		const float k = (1 - ema) * (1 - ema) / (1 + 2 * alpha * ema - ema2);
		a0 = k;
		a1 = k * (alpha - 1) * ema;
		a2 = k * (alpha + 1) * ema;
		a3 = -k * ema2;
	}
		break;

	case 1: {
		const float k = (1 - ema) * (1 - ema) / ema;
		a0 = k * ema;
		a1 = a3 = 0;
		a2 = -a0;
	}
		break;

	case 2: {
		const float ea = (float) std::exp(-alpha), k = -(ema2 - 1) / (2 * alpha * ema), kn = (-2 * (-1 + 3 * ea - 3 * ea * ea + ea * ea * ea) / (3 * ea + 1 + 3 * ea * ea + ea * ea * ea));
		a0 = kn;
		a1 = -kn * (1 + k * alpha) * ema;
		a2 = kn * (1 - k * alpha) * ema;
		a3 = -kn * ema2;
	}
		break;

	default:
		fprintf(stderr, "gaussianFilter: invalid order parameter!\n");
		return;
	}

	coefp = (a0 + a1) / (1 + b1 + b2);
	coefn = (a2 + a3) / (1 + b1 + b2);

	// process columns

	d_recursiveGaussian_rgba<<< iDivUp(width, nthreads), nthreads >>>(d_src, d_temp, width, height, a0, a1, a2, a3, b1, b2, coefp, coefn);

	getLastCudaError("Kernel execution failed");

	transpose(d_temp, d_dest, width, height);
	getLastCudaError("transpose: Kernel execution failed");

	// process rows

	d_recursiveGaussian_rgba<<< iDivUp(height, nthreads), nthreads >>>(d_dest, d_temp, height, width, a0, a1, a2, a3, b1, b2, coefp, coefn);

	getLastCudaError("Kernel execution failed");

	transpose(d_temp, d_dest, height, width);

d_intensityAdd<<< iDivUp(width, nthreads), nthreads >>>(d_src, d_dest, height, width);
}

void run_cuda(Parameters<float*>& p_parm) {
	float* tempTargetImage_d = 0; // x-filter

	//	float* theKernel_h = calc1DGaussKernel(kernelSize, p_parm.sigma);
	/* calculate image size */
	int n = (p_parm.resx * p_parm.resy); // number of pixels
	size_t size = n * sizeof(float) * p_parm.numComponents; //RGBA float image


	GRIP_CUDA_MALLOC_AND_COPY_IF_NOD(p_parm.sourceImage_d, p_parm.sourceImage_h, size, p_parm.sourceImage_dirty);
	GRIP_CUDA_MALLOC_IF_NULL(p_parm.targetImage_d, size);

	/* allocate temp destination image memory on device, used  */
	cudaMalloc((void**) &tempTargetImage_d, size);

	parm_float parm_host;
	//	memcpy(parm_host.kernel, theKernel_h, sizeof(float) * kernelSize);
	parm_host.intensity = p_parm.intensity;
	parm_host.resx = p_parm.resx;
	parm_host.resy = p_parm.resy;
	cudaMemcpyToSymbol(PARM, &parm_host, sizeof(parm_host));

	___GRIP_INITIALIZE_TIMER(1);
	___GRIP_START_TIMER(1);

	unsigned int nthreads = 128;

	if (p_parm.numComponents == 1) {
		
		gaussianFilterMonochrome(p_parm.sourceImage_d, p_parm.targetImage_d, tempTargetImage_d, p_parm.resx, p_parm.resy, p_parm.sigma, 0, nthreads);
		
	} else {
		
		gaussianFilterRGBA((float4*) p_parm.sourceImage_d, (float4*) p_parm.targetImage_d, (float4*) tempTargetImage_d, p_parm.resx, p_parm.resy, p_parm.sigma, 0, nthreads);
		
	}

	cudaThreadSynchronize();

	___GRIP_END_TIMER_AND_PRINT(1, "node_gauss_cuda", "kernel execution time", "run_cuda");

	/* copy ready-calculated data back to host! TODO do this only if it is necessary/wanted */
	//cudaMemcpy(p_param.targetImage_h, p_param.targetImage_d, size, cudaMemcpyDeviceToHost);
	

	//	delete[] theKernel_h;

	// as this is only known locally, it has to be freed here
	cudaFree(tempTargetImage_d);
}
