#include <cuda.h>
#include <iostream>
//#include <cutil_inline.h>

#include <Utils/CudaHelper.h>
#include <helper_math.h>


#include <thrust/device_vector.h>
#include <thrust/reduce.h>


#include "parameters.h"

typedef struct __PARAMETERS_FLOAT {
	int resx;
	int resy;
	float alpha;
	float4* targetImage;
} parm_float;

#undef PARM
#define PARM parm_float_c
__device__ __constant__ parm_float PARM;




#define X(X) (  0.51413640f * X[pos].x + 0.26506800f * X[pos].y + 0.02411880f * X[pos].z)
#define Y(X) (  0.32387860f * X[pos].x + 0.67023428f * X[pos].y + 0.12281780f * X[pos].z);
#define Z(X) (  0.16036376f * X[pos].x + 0.06409157f * X[pos].y + 0.84442666f * X[pos].z);

#define XYZ_TO_R(X, Y, Z) ( 2.5651f * X - 1.0217f * Y + 0.0753f * Z )
#define XYZ_TO_G(X, Y, Z) (-1.1665f * X + 1.9777f * Y - 0.2543f * Z )
#define XYZ_TO_B(X, Y, Z) (-0.3986f * X + 0.0439f * Y + 1.1892f * Z )


__global__ void computeLuminance(float4* p_colorImage, float* p_targetImage) {

	for (int y = blockIdx.x; y < PARM.resy; y += gridDim.x) {
		for (int x = threadIdx.x; x < PARM.resx; x += blockDim.x) {
			int pos = y * PARM.resx + x ;
			
			float luminance = Y(p_colorImage);
			
			p_targetImage[pos] = luminance;
		}
	}
	
}

__global__ void computeLogLuminance(float* p_targetImage) {

	for (int y = blockIdx.x; y < PARM.resy; y += gridDim.x) {
		for (int x = threadIdx.x; x < PARM.resx; x += blockDim.x) {
			int pos = y * PARM.resx + x ;
			
			float logLuminance = fmaxf(0.f, logf(p_targetImage[pos] + 1e-4f));
			
			p_targetImage[pos] = logLuminance;
		}
	}
	
}
 

__global__ void reinhard(float4* p_colorImage, float p_worldLuminance, float p_maxLuminance, float p_worldLuminanceWeight) {
	/** for each pixel */
	for (int y = blockIdx.x; y < PARM.resy; y += gridDim.x) {
		for (int x = threadIdx.x; x < PARM.resx; x += blockDim.x) {
			int pos = y * PARM.resx + x ;
			
			if (p_maxLuminance < 1e-4f) {
				PARM.targetImage[pos] = make_float4(0.f);
			} else {
				
				float invMaxLuminance = 1.f / (p_worldLuminanceWeight * p_worldLuminance + (1.f - p_worldLuminanceWeight) * p_maxLuminance);
				float4 scaledInput = PARM.alpha * p_colorImage[pos] * invMaxLuminance;
				
				PARM.targetImage[pos] = (scaledInput) / (1.f + scaledInput);
				
				PARM.targetImage[pos].w = p_colorImage[pos].w;
			
			
				PARM.targetImage[pos] = fmaxf(make_float4(0.f), fminf(PARM.targetImage[pos], make_float4(1.f)));	
				
			}
			
			//			float m = fmaxf(PARM.targetImage[pos], fmaxf(PARM.targetImage[pos+1], PARM.targetImage[pos+2]));
//			if (m > 1.f) {
//				PARM.targetImage[pos] /= m;
//				PARM.targetImage[pos+1] /= m;
//				PARM.targetImage[pos+2] /= m;
//			}



		}
	}
}



void run_cuda(Parameters<float*>& p_parm) {
	int n = (p_parm.resx * p_parm.resy); // number of pixels
	size_t size = n * sizeof(float) * 4; //RGBA float image

	parm_float parm_host;
	parm_host.resx = p_parm.resx;
	parm_host.resy = p_parm.resy;
	parm_host.alpha = p_parm.alpha;
	parm_host.targetImage = (float4*)p_parm.targetImage_d;
	cudaMemcpyToSymbol(PARM, &parm_host, sizeof(parm_host));

	float* tempTargetImage_d = 0;
	cudaMalloc((void**) &tempTargetImage_d, size / 4);
	
	computeLuminance<<< 64, 256 >>>((float4*)p_parm.colorImage_d, tempTargetImage_d);
	
	thrust::device_ptr<float> devPtr((float*) (tempTargetImage_d));
	
	
	float maxLuminance = thrust::reduce(devPtr, devPtr + p_parm.resx * p_parm.resy, 0.f, thrust::maximum<float>());
	
// 	std::cout << "maxLuminance: " << maxLuminance << std::endl;
	
	computeLogLuminance<<< 64, 256 >>>(tempTargetImage_d);
	
	
	//	thrust::device_vector<float4> inputImage(p_parm.resx * p_parm.resy);
	
// 	float init = 0.f;//make_float4(0.f, 0.f, 0.f, 0.f);
	//	thrust::copy(devPtr, devPtr + p_parm.resx * p_parm.resy, inputImage.begin());
	float reductionResult = thrust::reduce(devPtr, devPtr + p_parm.resx * p_parm.resy, 0.f, thrust::placeholders::_1+thrust::placeholders::_2 );
	
// 	std::cout << "reductionResult: " << reductionResult << std::endl;
	
	reductionResult = expf(reductionResult / (p_parm.resx * p_parm.resy));
// 	std::cout << "Ywa: " << reductionResult << std::endl;
	
	reinhard<<< 64, 256 >>>((float4*)p_parm.colorImage_d, reductionResult, maxLuminance, p_parm.worldLuminanceWeight);
	checkCudaErrors(cudaGetLastError());
	cudaDeviceSynchronize();
	//cutilCheckMsg("Kernel execution failed!");

	cudaFree(tempTargetImage_d);
	
}
