/*
 * LightSampler.cu
 *
 *  Created on: Aug 15, 2011
 *      Author: troth22s
 */

#ifndef LIGHTSAMPLER__CU__
#define LIGHTSAMPLER__CU__

#include <helpers.h>
#include <constants.h>

//#include <Stratification.h>

#include <SphereLight.cu>
#include <PointLight.cu>

#include <optix_declaration_macros.h>

#ifndef __OPTIX_VARS_CAMERA__
#define __OPTIX_VARS_CAMERA__
__DECLARE_OPTIX_VARS_CAMERA__
#endif

#ifndef __OPTIX_VARS_FRAMEINFO__
#define __OPTIX_VARS_FRAMEINFO__
__DECLARE_OPTIX_VARS_FRAMEINFO__
#endif

#ifndef __OPTIX_VARS_SAMPLER_SETTINGS__
#define __OPTIX_VARS_SAMPLER_SETTINGS__
__DECLARE_OPTIX_VARS_SAMPLER_SETTINGS__
#endif

#ifndef __OPTIX_VARS_RAY_TYPES__
#define __OPTIX_VARS_RAY_TYPES__
__DECLARE_OPTIX_VARS_RAY_TYPES__
#endif


#ifndef __OPTIX_VARS_RAY_DATA__
#define __OPTIX_VARS_RAY_DATA__
__DECLARE_OPTIX_VARS_RAY_DATA__
#endif

#ifndef __OPTIX_VARS_ENVIRONMENT_INFO__
#define __OPTIX_VARS_ENVIRONMENT_INFO__
__DECLARE_OPTIX_VARS_ENVIRONMENT_INFO__
#endif

#ifndef __OPTIX_VARS_GEOMETRY_INFO__
#define __OPTIX_VARS_GEOMETRY_INFO__
__DECLARE_OPTIX_VARS_GEOMETRY_INFO__
#endif

#ifndef __OPTIX_VARS_LIGHT_BUFFERS__
#define __OPTIX_VARS_LIGHT_BUFFERS__
__DECLARE_OPTIX_VARS_LIGHT_BUFFERS__
#endif


#include <sobol.h>
#include <sobol_cuda.cu>

#include <Distribution2D_env_device.cu>

/** This file contains all light sampling functions
 * they are usually based on templates with the according BRDF type for the
 * current hit point as the sole template parameter
 */

/**
 * @brief method for sampling a spherical light source
 * @param p_lsIndex: spherelight index
 * @param p_lightProbability: probability for MIS to choose the light source's distribution for sampling
 * @param p_hitpoint: current hit point coordinates
 * @param p_wo: light's outgoing direction (i.e. the ray back towards the observer)
 * @param p_normal: current hit point's shading normal
 * @param p_normal_geometric: current hit point's geometric normal
 * @param p_brdf: current hit point's associated brdf
 */
template<class T> __device__ float3 sampleSphereLight ( unsigned int p_lsIndex, float p_lightProbability, float3 &p_hitpoint, float3 &p_wo, float3 &p_normal, float3 &p_normal_geometric, T &p_brdf ) {
    float prob1 = p_lightProbability;
    float prob2 = 1.f - prob1;
    float u1 = 0.f;
    float u2 = 0.f;
    u1 = rnd ( prd_pt.seed );
    u2 = rnd ( prd_pt.seed );
    optix::float3 lightEmission = ( ( sphereLights[p_lsIndex].getFluxTotal() * sphereLights[p_lsIndex].getColor() ) / ( sphereLights[p_lsIndex].getArea() * M_PIf ) );
    float brdfPdf;
    float lightPdf = 2.f / ( sphereLights[p_lsIndex].getArea() ); // INV_Tp_woPI as only the hemisphere is respected
    float brdfValue;
    PerRayData_shadow prd_shadow;
    PerRayData_occlusion prd_occlusion;
    optix::float3 lightDir;
    float maxLightDist = optix::length ( sphereLights[p_lsIndex].getPosition() - p_hitpoint ) + sphereLights[p_lsIndex].getSize();

    if ( rnd ( prd_pt.seed ) < prob1 ) { // sample from light source's distribution
        lightDir = sphereLights[p_lsIndex].getRandomSurfacePoint ( u1, u2 ) - p_hitpoint;
        lightDir = optix::normalize ( lightDir );
        brdfValue = p_brdf.f ( p_wo, lightDir, p_normal, &brdfPdf );

        if ( brdfPdf > 0.f && brdfValue > 0.f && optix::dot ( lightDir, p_normal ) > 0.f ) {
            prd_shadow.throughput = make_float3 ( 1.0f );
            prd_shadow.emittance = 0.0f;
            prd_shadow.queryIdentifier = sphereLights[p_lsIndex].getIdentifier();
            prd_shadow.hitDist = FLT_MAX;
            prd_shadow.occluded = false;
            prd_shadow.seed = prd_pt.seed;
            optix::Ray shadowRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), lightDir, shadow_ray_type, scene_epsilon, maxLightDist );
            rtTrace ( light_group, shadowRay, prd_shadow );
            ++prd_pt.numCastRays;

            rtPrintf("A: emittance %f\n", prd_shadow.emittance);

            prd_occlusion.occluded = false; 
            rtPrintf("A: occlusion before %d\n", prd_occlusion.occluded);
            optix::Ray occlusionRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ),  lightDir, occlusion_ray_type, scene_epsilon, maxLightDist );
            rtTrace ( top_object, occlusionRay, prd_occlusion );
            ++prd_pt.numCastRays;

            rtPrintf("A: occlusion after %d\n", prd_occlusion.occluded);

            if ( !prd_occlusion.occluded ) {
                float weight = prob2 > 0.f ? balanceHeuristic ( 1, lightPdf, 1, brdfPdf ) : 1.f;
                float geometryTerm = fminf ( ( prd_shadow.emittance * fmaxf ( 0.f, optix::dot ( lightDir, p_normal ) ) / ( prd_shadow.hitDist * prd_shadow.hitDist ) ), 1.f );
                optix::float3 E = ( lightEmission * geometryTerm * brdfValue * weight ) / lightPdf;
                E = fmaxf(make_float3(0.f), E);
                return ( E * prd_shadow.throughput );
            }
        }

    } else { // sample from brdf's distribution
        brdfValue = p_brdf.sample_f ( p_wo, &lightDir, p_normal, u1, u2, &brdfPdf );

        if ( brdfValue > 0.f && brdfPdf > 0.f ) {
            prd_shadow.throughput = make_float3 ( 1.0f );
            prd_shadow.emittance = 0.0f;
            prd_shadow.queryIdentifier = sphereLights[p_lsIndex].getIdentifier();
            prd_shadow.hitDist = FLT_MAX;
            prd_shadow.occluded = false;
            prd_shadow.seed = prd_pt.seed;
            optix::Ray shadowRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), lightDir, shadow_ray_type, scene_epsilon, maxLightDist );
            rtTrace ( light_group, shadowRay, prd_shadow );
            ++prd_pt.numCastRays;

            prd_occlusion.occluded = false;
            optix::Ray occlusionRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ),  lightDir, occlusion_ray_type, scene_epsilon, maxLightDist );
            rtTrace ( top_object, occlusionRay, prd_occlusion );
            ++prd_pt.numCastRays;


            if ( !prd_occlusion.occluded ) {
                float weight = prob1 > 0.f ? balanceHeuristic ( 1, brdfPdf, 1, lightPdf ) : 1.f;
                float geometryTerm = ( prd_shadow.emittance * fmaxf ( 0.f, optix::dot ( lightDir, p_normal ) ) );
                optix::float3 E = ( lightEmission * geometryTerm * brdfValue * weight ) / brdfPdf;
                E = fmaxf(make_float3(0.f), E);
                return ( E * prd_shadow.throughput );
            }
        }
    }

    return make_float3 ( 0 );
}


/**
 * @brief method for sampling a point light source
 * @param p_lsIndex: spherelight index
 * @param p_hitpoint: current hit point coordinates
 * @param p_wo: light's outgoing direction (i.e. the ray back towards the observer)
 * @param p_normal: current hit point's shading normal
 * @param p_normal_geometric: current hit point's geometric normal
 * @param p_brdf: current hit point's associated brdf
 */
template<class T> __device__ float3 samplePointLight ( unsigned int p_lsIndex, float3 &p_hitpoint, float3 &p_wo, float3 &p_normal, float3 &p_normal_geometric, T &p_brdf ) {
    optix::float3 lightEmission = ( ( pointLights[p_lsIndex].getFluxTotal() * pointLights[p_lsIndex].getColor() ) / ( FOURPI ) );
    float brdfPdf;
    float brdfValue;
    //
    optix::float3 lightDir = pointLights[p_lsIndex].getPosition() - p_hitpoint;
    float lightDist = optix::length ( lightDir );
    lightDir = optix::normalize ( lightDir );
    brdfValue = p_brdf.f ( p_wo, lightDir, p_normal, &brdfPdf );

    if ( brdfPdf > 0.f && brdfValue > 0.f && optix::dot ( lightDir, p_normal ) > 0.f ) {
        PerRayData_occlusion prd_shadow;
        prd_shadow.throughput = make_float3 ( 1.f );
        prd_shadow.occluded = false;
        prd_shadow.seed = prd_pt.seed;
        optix::Ray shadowRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), lightDir, occlusion_ray_type, 0.f, lightDist );
        rtTrace ( top_object, shadowRay, prd_shadow );
        ++prd_pt.numCastRays;

        if ( !prd_shadow.occluded ) {
            float geometryTerm = fminf ( ( fmaxf ( 0.f, optix::dot ( lightDir, p_normal ) ) / ( lightDist * lightDist ) ), 1.f );
            optix::float3 E = ( lightEmission * geometryTerm * brdfValue );
            return E * prd_shadow.throughput;
        }
    }

    return make_float3 ( 0 );
}


   // float3 envDir = ray.direction;

   //      possiblyFlipYZ(envDir, envFlipYZ);

   //      float phi = atan2f ( envDir.y, ray.direction.x ) + envRotPhi;
   //      //      float theta = M_PIf * 0.5f - acosf(ray.direction.y);
   //      float theta = acosf ( envDir.z );
   //      float u = ( phi + M_PIf ) * INV_TWOPI;

   //      float v = theta * INV_PI;
   //      float cosAngle = optix::dot ( ray.direction, directionalLight_direction );
   //      float cosSunScale = cosf ( sunScale );

   //      if ( cosAngle > cosSunScale ) {
   //          prd_pt.radiance = directionalLight_fluxRGB;
   //          prd_pt.texCol = make_float3 ( 1.f );

   //      } else {
   //          float3 bgCol = calculateSkyRGB ( directionalLight_direction, skyBrightness ) * make_float3 ( tex2D ( bg_map, u, v ) ); //fmaxf(MIN_SKY_RGB, MAX_SKY_RGB) * make_float3(tex2D(bg_map, u, v));// * (MAX_SKY_RGB * weightSun + MIN_SKY_RGB * (1.f - weightSun));

   //          prd_pt.radiance = bgCol;
   //          prd_pt.texCol = make_float3 ( 1.f ); //make_float3(w1) + w2 * skyBrightness * make_float3(tex2D(bg_map, u, v));
   //      }

/**
 * @brief method for sampling the environment
 * @param p_lightProbability: probability for MIS to choose the light source's distribution for sampling
 * @param p_hitpoint: current hit point coordinates
 * @param p_wo: light's outgoing direction (i.e. the ray back towards the observer)
 * @param p_normal: current hit point's shading normal
 * @param p_normal_geometric: current hit point's geometric normal
 * @param p_brdf: current hit point's associated brdf
 */
template<class T> __device__ float3 sampleEnvironment ( float p_lightProbability, float3 &p_hitpoint, float3 &p_wo, float3 &p_normal, float3 &p_normal_geometric, T &p_brdf ) {
    float3 envDir;
    float prob1 = p_lightProbability;
    float prob2 = 1.f - prob1;
    float3 retVal = make_float3 ( 0.f );
    unsigned int dimension = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_ENV_MIS_CHOOSE;

    if ( Sobol ( dimension, prd_pt.frame, prd_pt.seed ) < prob1 ) {
        float u1 = 0.f;
        float u2 = 0.f;

        if ( useLdSampling_environment ) {
            unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_ENV_U;
            unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_ENV_V;
            u1 = Sobol ( dimension_u, prd_pt.frame, prd_pt.seed );
            u2 = Sobol ( dimension_v, prd_pt.frame, prd_pt.seed );

        } else {
            u1 = rnd ( prd_pt.seed );
            u2 = rnd ( prd_pt.seed );
        }

        float uv[2] = { 0.f, 0.f };
        float mapPdf = 1.f;
        sampleContinuous ( u1, u2, uv, mapPdf );
        //  float phi = uv[0] * M_PIf * 2.f + M_PIf * 0.5f;
        float phi = uv[0] * M_PIf * 2.f - M_PIf - envRotPhi;
        phi = fmodf(phi, 2.f * M_PIf);
        float theta = M_PIf * uv[1];
        //  rtPrintf("%f, ", uv[1]);
        //  rtPrintf("%f, %f; ", uv[0], uv[1]);
        envDir.x = sinf ( theta ) * cosf ( phi );
        envDir.y = sinf ( theta ) * sinf ( phi );
        envDir.z = cosf ( theta );
        possiblyFlipYZ(envDir, envFlipYZ);



        envDir = optix::normalize ( envDir );
        float brdfPdf = 0.f;
        float brdfValue = p_brdf.f ( p_wo, envDir, p_normal, &brdfPdf );

        //  float brdfPdf = 0.f;
        //  float brdfValue = p_brdf.sample_f(p_wo, &envDir, p_normal, u1, u2, &brdfPdf);

        // do not sample directional light!
        if ( optix::dot ( envDir, directionalLight_direction ) < cosf ( sunScale ) && optix::dot ( envDir, p_normal ) > 0.f ) { // sun rays are sampled separately
            PerRayData_occlusion prd_occ_env;
            prd_occ_env.occluded = false;
            prd_occ_env.seed = prd_pt.seed;
            prd_occ_env.throughput = make_float3 ( 1.f );
            optix::Ray occ_envRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), envDir, occlusion_ray_type, 0.f, RT_DEFAULT_MAX );
            rtTrace ( top_object, occ_envRay, prd_occ_env );
            ++prd_pt.numCastRays;

            if ( !prd_occ_env.occluded ) {
                // a fixed factor is used here to brighten the environment contribution a bit
                if ( brdfPdf > 0.f ) {
                    //              float theta = atan2f(envDir.x, envDir.z);
                    //              float phi = M_PIf * 0.5f - acosf(envDir.y);
                    //              float u = (phi + M_PIf) * (0.5f * M_1_PIf);
                    //              float v = 0.5f * (1.0f + sinf(theta));
                    //float3 weightSun = calculateSunWeight(sunScale, directionalLight_fluxRGB, sunBrightness);
                    float3 bgCol = calculateSkyRGB ( directionalLight_direction, skyBrightness ) * make_float3 ( tex2D ( bg_map, uv[0], uv[1] ) ); //fmaxf(MIN_SKY_RGB, MAX_SKY_RGB) * make_float3(tex2D(bg_map, u, v));// * (MAX_SKY_RGB * weightSun + MIN_SKY_RGB * (1.f - weightSun));
                    float pdf = mapPdf / ( 2.f * M_PIf * M_PIf * sinf ( theta ) );
                    float weight = prob2 > 0.f ? balanceHeuristic ( 1, pdf, 1, brdfPdf ) : 1.f;
                    //retVal += (optix::dot(envDir, p_normal) * bgCol * brdfValue * prd_occ_env.throughput * weight) / pdf;
                    retVal += ( optix::dot ( envDir, p_normal ) * bgCol * brdfValue * prd_occ_env.throughput * weight ) / ( pdf * prob1 );
                }
            }
        }

    } else { // sample from BSDF
        //}{
        float u1 = 0.f;
        float u2 = 0.f;

        if ( useLdSampling_environment ) {
            unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_ENV_U;
            unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_ENV_V;
            u1 = Sobol ( dimension_u, prd_pt.frame, prd_pt.seed );
            u2 = Sobol ( dimension_v, prd_pt.frame, prd_pt.seed );

        } else {
            u1 = rnd ( prd_pt.seed );
            u2 = rnd ( prd_pt.seed );
        }

        float brdfPdf = 1.f;
        float brdfValue = p_brdf.sample_f ( p_wo, &envDir, p_normal, u1, u2, &brdfPdf );
        possiblyFlipYZ(envDir, envFlipYZ);
        
        // do not sample directional light!brdfPdf
        if ( optix::dot ( envDir, directionalLight_direction ) < cosf ( sunScale ) && optix::dot ( envDir, p_normal ) > 0.f ) { // sun rays are sampled separately
            PerRayData_occlusion prd_occ_env;
            prd_occ_env.occluded = false;
            prd_occ_env.seed = prd_pt.seed;
            prd_occ_env.throughput = make_float3 ( 1.f );

            optix::Ray occ_envRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), envDir, occlusion_ray_type, 0.f, RT_DEFAULT_MAX );
            rtTrace ( top_object, occ_envRay, prd_occ_env );
            ++prd_pt.numCastRays;

            if ( !prd_occ_env.occluded ) {
                // a fixed factor is used here to brighten the environment contribution a bit
                if ( brdfPdf > 0.f ) {
                    float phi = atan2f ( envDir.y, envDir.x ) + envRotPhi;
                    phi = fmodf(phi, 2.f * M_PIf);
                    //      float theta = M_PIf * 0.5f - acosf(ray.direction.y);
                    float theta = acosf ( envDir.z );

                    if ( theta < 0.f ) {
                        theta = M_PIf - theta;
                    }

                    //      float u = (phi + M_PIf) * (0.5f * M_1_PIf);
                    float u = fmodf(( phi + M_PIf ) * INV_TWOPI, 1.f);
                    //      float v = 0.5f * (1.0f + sinf(theta));
                    float v = theta * INV_PI;
                    //float3 weightSun = calculateSunWeight(sunScale, directionalLight_fluxRGB, sunBrightness);
                    float mapPdf = envPdf ( u, v ) / ( 2.f * M_PIf * M_PIf * sinf ( theta ) );
                    float weight = prob1 > 0.f ? balanceHeuristic ( 1, brdfPdf, 1, mapPdf ) : 1.f;
                    float3 bgCol = calculateSkyRGB ( directionalLight_direction, skyBrightness ) * make_float3 ( tex2D ( bg_map, u, v ) ); //fmaxf(MIN_SKY_RGB, MAX_SKY_RGB) * make_float3(tex2D(bg_map, u, v));// * (MAX_SKY_RGB * weightSun + MIN_SKY_RGB * (1.f - weightSun));
                    retVal += ( optix::dot ( envDir, p_normal ) * bgCol * brdfValue * prd_occ_env.throughput * weight ) / ( brdfPdf * prob2 );
                }
            }
        }
    }

    return retVal; // 2 samples have been taken
}

/**
 * @brief method for sampling the sun
 * @param p_lightProbability: probability for MIS to choose the light source's distribution for sampling
 * @param p_hitpoint: current hit point coordinates
 * @param p_wo: light's outgoing direction (i.e. the ray back towards the observer)
 * @param p_normal: current hit point's shading normal
 * @param p_normal_geometric: current hit point's geometric normal
 * @param p_brdf: current hit point's associated brdf
 */
template<class T> __device__ float3 sampleDirectionalLight ( float p_lightProbability, float3 &p_hitpoint, float3 &p_wo, float3 &p_normal, float3 &p_normal_geometric, T &p_brdf ) {
    float prob1 = p_lightProbability;
    float prob2 = 1.f - prob1;
    float cosScale = cosf ( sunScale );
    float conePdf = 1.f / ( 2.f * M_PIf * ( 1.f - cosScale ) );
    float3 retVal = make_float3 ( 0.f );
    unsigned int dimension = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_ENV_MIS_CHOOSE;

    if ( Sobol ( dimension, prd_pt.frame, prd_pt.seed ) < prob1 ) {
        float u1 = 0.f;
        float u2 = 0.f;

        if ( useLdSampling_sun ) {
            unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_SUN_U;
            unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_SUN_V;
            u1 = Sobol ( dimension_u, prd_pt.frame, prd_pt.seed );
            u2 = Sobol ( dimension_v, prd_pt.frame, prd_pt.seed );

        } else {
            u1 = rnd ( prd_pt.seed );
            u2 = rnd ( prd_pt.seed );
        }

        float3 envDir;
        optix::Onb onb ( directionalLight_direction );
        // uniform cone sampling!
        float cosTheta = ( 1.f - u1 ) + u1 * cosScale;
        float sinTheta = sqrtf ( 1.f - cosTheta * cosTheta );
        float phi = u2 * 2.f * M_PIf;
        envDir.x = cosf ( phi ) * sinTheta;
        envDir.y = sinf ( phi ) * sinTheta;
        envDir.z = cosTheta;
        onb.inverse_transform ( envDir );
        envDir = optix::normalize ( envDir );
        float theDot = optix::dot ( envDir, p_normal );

        if ( theDot > 0.f ) {
            PerRayData_occlusion prd_occ_env;
            prd_occ_env.occluded = false;
            prd_occ_env.seed = prd_pt.seed;
            prd_occ_env.throughput = make_float3 ( 1.f );
            float brdfPdf = 1.f;
            float brdfValue = p_brdf.f ( p_wo, envDir, p_normal, &brdfPdf );

            if ( brdfValue > 0.f && brdfPdf > 0.f ) {
                optix::Ray occ_envRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), envDir, occlusion_ray_type, 0.f, RT_DEFAULT_MAX );
                rtTrace ( top_object, occ_envRay, prd_occ_env );
                ++prd_pt.numCastRays;

                if ( !prd_occ_env.occluded ) {
                    float weight = prob2 > 0.f ? balanceHeuristic ( 1, conePdf, 1, brdfPdf ) : 1.f;
                    float cosAngle = optix::dot ( envDir, directionalLight_direction );
                    float cosSunScale = cosf ( sunScale );

                    if ( cosAngle > cosSunScale ) {
                        retVal += ( fmaxf ( 0.f, optix::dot ( envDir, p_normal ) ) * directionalLight_fluxRGB * brdfValue * weight * prd_occ_env.throughput ) / ( conePdf * prob1 );
                    }
                }
            }
        }

    } else { // sample from material brdf
        float u1 = 0.f;
        float u2 = 0.f;

        if ( useLdSampling_sun ) {
            unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_SUN_U;
            unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_SUN_V;
            u1 = Sobol ( dimension_u, prd_pt.frame, prd_pt.seed );
            u2 = Sobol ( dimension_v, prd_pt.frame, prd_pt.seed );

        } else {
            u1 = rnd ( prd_pt.seed );
            u2 = rnd ( prd_pt.seed );
        }

        float3 envDir;
        float brdfPdf = 1.f;
        float brdfValue = p_brdf.sample_f ( p_wo, &envDir, p_normal, u1, u2, &brdfPdf );

        if ( brdfValue > 0.f && brdfPdf > 0.f ) {
            PerRayData_occlusion prd_occ_env;
            prd_occ_env.occluded = false;
            prd_occ_env.seed = prd_pt.seed;
            prd_occ_env.throughput = make_float3 ( 1.f );
            optix::Ray occ_envRay = optix::make_Ray ( getAdjustedRayStart ( p_hitpoint, p_normal_geometric, scene_epsilon ), envDir, occlusion_ray_type, 0.f, RT_DEFAULT_MAX );
            rtTrace ( top_object, occ_envRay, prd_occ_env );
            ++prd_pt.numCastRays;

            if ( !prd_occ_env.occluded ) {
                float weight = prob1 > 0.f ? balanceHeuristic ( 1, brdfPdf, 1, conePdf ) : 1.f;
                float cosAngle = optix::dot ( envDir, directionalLight_direction );
                float cosSunScale = cosf ( sunScale );

                if ( cosAngle > cosSunScale ) {
                    retVal += ( fmaxf ( 0.f, optix::dot ( envDir, p_normal ) ) * directionalLight_fluxRGB * brdfValue * weight * prd_occ_env.throughput ) / ( brdfPdf * prob2 );
                }
            }
        }
    }

    return retVal; // 2 samples have been taken
}


/* sample all light sources */
template<class T> __device__ optix::float3 sampleAllLightSources ( optix::float3 &p_hitpoint, optix::float3 &p_normal, optix::float3 &p_normal_geometric, float p_probLight, T &p_brdf ) {
    optix::float3 lightEmittedRadiance = make_float3 ( 0.f );
    float3 wo = -ray.direction;

    for ( int lsIndex = 0; lsIndex < sphereLights.size(); ++lsIndex ) {
        lightEmittedRadiance += sampleSphereLight ( lsIndex, p_probLight, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf );
    }

    for ( int lsIndex = 0; lsIndex < pointLights.size(); ++lsIndex ) { // sample point lights
        lightEmittedRadiance += samplePointLight ( lsIndex, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf );
    }

    //  for (unsigned int i = 0; i < s; i++) {
    if ( skyBrightness > 0.f ) {
        lightEmittedRadiance += sampleEnvironment ( p_probLight, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf );
    }

    //  }

    if ( sunBrightness > 0.f && sunScale > 0.f ) {
        lightEmittedRadiance += sampleDirectionalLight ( p_probLight, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf );
    }

    return lightEmittedRadiance;
}


/* sample only one randomly chosen light source */
template<class T> __device__ optix::float3 sampleOneLightSource ( optix::float3 &p_hitpoint, optix::float3 &p_normal, optix::float3 &p_normal_geometric, float p_probLight, T &p_brdf ) {
    float samplingProbability = 1.f / ( sphereLights.size() + pointLights.size() + 2 );
    optix::float3 lightEmittedRadiance = make_float3 ( 0.f );
    float3 wo = -ray.direction;
    unsigned int dimension = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_LIGHT_SOURCE_CHOOSE;
    unsigned int lsIndex = Sobol ( dimension, prd_pt.frame, prd_pt.seed ) * ( sphereLights.size() + pointLights.size() + 2 );

    if ( lsIndex < sphereLights.size() ) {
        lightEmittedRadiance += sampleSphereLight ( lsIndex, p_probLight, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf ) / samplingProbability;

    } else if ( lsIndex >= sphereLights.size() && lsIndex < sphereLights.size() + pointLights.size() ) { // sample point lights
        lsIndex -= sphereLights.size();
        lightEmittedRadiance += samplePointLight ( lsIndex, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf ) / samplingProbability;

    } else if ( lsIndex == sphereLights.size() + pointLights.size() && ( skyBrightness > 0.f || sunBrightness > 0.f ) ) { // sample environment
        lightEmittedRadiance += sampleEnvironment ( p_probLight, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf ) / samplingProbability;

    } else if ( lsIndex == sphereLights.size() + pointLights.size() + 1 && sunBrightness > 0.f && sunScale > 0.f ) { // sample directional light
        lightEmittedRadiance += sampleDirectionalLight ( p_probLight, p_hitpoint, wo, p_normal, p_normal_geometric, p_brdf ) / samplingProbability;
    }

    return lightEmittedRadiance;
}

#endif
