/*
 * LightSource.cu
 *
 *  Created on: Jun 29, 2011
 *      Author: troth22s
 */

#ifndef LIGHTSOURCE_CU_
#define LIGHTSOURCE_CU_


#ifndef __CUDACC__
#include <optixu/optixpp_namespace.h>
#include <common_structs.h>
#endif


/**
 * @class LightSource
 * @brief This class defines most of the information a light source needs
 *
 * Concrete light source classes are derived from this class and should additionaly implement a cuda method
 * for sampling random surface points
 */
class LightSource {
  private:
    optix::float3 __m_position;
    float __m_area;
    float __m_size;
    optix::float3 __m_color;
    float __m_fluxTotal;
    unsigned int __m_identifier;
    static unsigned int __sm_currentIdentifier;

    virtual float calcArea() = 0;

  public:

    __host__ __device__ optix::float3 getPosition() {
        return __m_position;
    }

    __host__ __device__ float getSize() {
        return __m_size;
    }

    __host__ __device__ optix::float3 getColor() {
        return __m_color;
    }

    __host__ __device__ float getFluxTotal() {
        return __m_fluxTotal;
    }

    __host__ __device__ float getFluxPerUnitArea() {
        return __m_fluxTotal / __m_area;
    }

    __host__ __device__ float getArea() {
        return __m_area;
    }

    __host__ __device__ unsigned int getIdentifier() {
        return __m_identifier;
    }

// host only functions!
    #ifndef __CUDACC__
  private:
    static optix::Program __sm_lightShaderPt;
    static optix::Program __sm_lightShaderShadow;
    static optix::Program __sm_lightShaderOcc;
    static optix::Material __sm_lightMaterial;
    

  public:
    LightSource ( optix::Context p_context) {
        __m_identifier = ++__sm_currentIdentifier;

        if ( !__sm_lightShaderPt ) {
            __sm_lightShaderPt = p_context->createProgramFromPTXFile ( generatePtxPath ( "light_shader.cu" ), "light_shader_pt" );
        }

        if ( !__sm_lightShaderShadow ) {
            __sm_lightShaderShadow = p_context->createProgramFromPTXFile ( generatePtxPath ( "light_shader.cu" ), "light_shader_shadow" );
        }

        if ( !__sm_lightMaterial ) {
            __sm_lightMaterial = p_context->createMaterial();
            __sm_lightMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, __sm_lightShaderPt );
            __sm_lightMaterial->setClosestHitProgram ( ENUM_RAYTYPE_SHADOW, __sm_lightShaderShadow );

            __sm_lightShaderOcc = p_context->createProgramFromPTXFile ( generatePtxPath ( "occlusion_shader.cu" ), "occlusion_shader" );
            __sm_lightMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, __sm_lightShaderOcc );
        }
    }

    LightSource() {
        __m_identifier = ++__sm_currentIdentifier;
    }

    virtual optix::Geometry getUnitGeometry() = 0;

    void setPosition ( optix::float3 p_position ) {
        __m_position = p_position;
    }

    void setSize ( float p_size ) {
        __m_size = p_size;
        __m_area = calcArea();
    }

    void setColor ( optix::float3 p_color ) {
        __m_color = p_color;
    }

    void setFluxTotal ( float p_flux ) {
        __m_fluxTotal = p_flux;
    }

    void setFluxPerUnitArea ( float p_flux ) {
        __m_fluxTotal = p_flux * __m_area;
    }

    optix::Transform createGeometry ( optix::Context p_context ) {
        optix::GeometryInstance gi = p_context->createGeometryInstance();
        gi->setMaterialCount ( 1 );
        gi->setGeometry ( getUnitGeometry() );
        gi->setMaterial ( 0, __sm_lightMaterial );
        gi["color"]->setFloat ( __m_color );
        gi["flux"]->setFloat ( __m_fluxTotal );
        gi["area"]->setFloat ( __m_area );
        gi["identifier"]->setUint ( __m_identifier );
        optix::GeometryGroup giGroup = p_context->createGeometryGroup();
        giGroup->setChildCount ( 1 );
        giGroup->setChild ( 0, gi );
        giGroup->setAcceleration ( p_context->createAcceleration ( "NoAccel", "NoAccel" ) );
        optix::Transform transform = p_context->createTransform();
        optix::Matrix4x4 scaleMatrix = optix::Matrix4x4::scale ( optix::make_float3 ( __m_size ) );
        optix::Matrix4x4 translateMatrix = optix::Matrix4x4::translate ( __m_position );
        optix::Matrix4x4 rotateMatrix = optix::Matrix4x4::rotate ( 0.0f, optix::make_float3 ( 0.0f, 1.0f, 0.0f ) );
        transform->setMatrix ( 0, ( translateMatrix * rotateMatrix * scaleMatrix ).getData(), 0 );
        transform->setChild ( giGroup );
        return transform;
    }
    #endif
};

/** static variables */
unsigned int LightSource::__sm_currentIdentifier = 0;

/** host specific static variables */
#ifndef __CUDACC__
optix::Program LightSource::__sm_lightShaderPt = NULL;
optix::Program LightSource::__sm_lightShaderShadow = NULL;
optix::Program LightSource::__sm_lightShaderOcc = NULL;
optix::Material LightSource::__sm_lightMaterial = NULL;
#endif

#endif /* LIGHTSOURCE_CU_ */
