/*
 * helpers.h
 *
 *  Created on: May 13, 2011
 *      Author: troth22s
 */


#ifndef HELPERS_H_
#define HELPERS_H_

#include <constants.h>
#include <optixu/optixu_math_namespace.h>
#include <math.h>
#include <iostream>




/* return max component of a float3 */
__device__ __host__ __inline__ float max_component ( const optix::float3 &p_vec ) {
    return fmaxf ( p_vec.x, fmaxf ( p_vec.y, p_vec.z ) );
}


/* return max component of a float4 */
__device__ __host__ __inline__ float max_component ( const optix::float4 &p_vec ) {
    return fmaxf ( p_vec.x, fmaxf ( p_vec.y, fmaxf ( p_vec.z, p_vec.w ) ) );
}


__device__ __host__ __inline__ float fresnel ( float p_eta, float p_k, float p_cosThetaI ) {
    float tmp = ( p_eta * p_eta + p_k * p_k ) * p_cosThetaI * p_cosThetaI;
    float rparl2 = ( tmp - ( 2.f * p_eta * p_cosThetaI ) + 1.f ) / ( tmp + ( 2.f * p_eta * p_cosThetaI ) + 1.f );
    float tmp_f = p_eta * p_eta + p_k * p_k;
    float rperp2 = ( tmp_f - ( 2.f * p_eta * p_cosThetaI ) + p_cosThetaI * p_cosThetaI ) / ( tmp_f + ( 2.f * p_eta * p_cosThetaI ) + p_cosThetaI * p_cosThetaI );
    return ( rparl2 + rperp2 ) * 0.5f;
}


__device__ __host__ __inline__ float fresnelDiel ( float p_cosI, float p_cosT, float p_etaI, float p_etaT ) {
    float Rparl = ( ( p_etaT * p_cosI ) - ( p_etaI * p_cosT ) ) / ( ( p_etaT * p_cosI ) + ( p_etaI * p_cosT ) );
    float Rperp = ( ( p_etaI * p_cosI ) - ( p_etaT * p_cosT ) ) / ( ( p_etaI * p_cosI ) + ( p_etaT * p_cosT ) );
    return fminf ( ( Rparl * Rparl + Rperp * Rperp ) * 0.5f, 1.f );
}


__device__ __host__ __inline__ float powerHeuristic ( int p_nf, float p_fPdf, int p_ng, float p_gPdf ) {
    float f = p_nf * p_fPdf;
    float g = p_ng * p_gPdf;
    return ( f * f ) / ( f * f + g * g );
}


__device__ __host__ __inline__ float balanceHeuristic ( int p_nf, float p_fPdf, int p_ng, float p_gPdf ) {
    return ( p_nf * p_fPdf ) / ( p_nf * p_fPdf + p_ng * p_gPdf );
}


/* compute RGB sun intensity based on position and set brightness */
__device__ __host__ __inline__ optix::float3 calculateSunRGB ( optix::float3 &p_position, float p_sunBrightness ) {
    if ( p_position.z >= 0.f ) {
        optix::float3 weight = optix::make_float3 ( powf ( p_position.z, SUN_EXPONENT_R ), powf ( p_position.z, SUN_EXPONENT_G ), powf ( p_position.z, SUN_EXPONENT_B ) );
        return ( weight * MAX_SUN_RGB + ( 1.f - weight ) * MIN_SUN_RGB ) * p_sunBrightness;

    } else {
        return MIN_SUN_RGB * p_sunBrightness;
    }
}

__device__ __host__ __inline__ optix::float3 calculateSunRGB ( float *p_position_x, float p_sunBrightness ) {
    optix::float3 p_position = * ( optix::float3 * ) p_position_x;

    if ( p_position.z >= 0.f ) {
        optix::float3 weight = optix::make_float3 ( powf ( p_position.z, SUN_EXPONENT_R ), powf ( p_position.z, SUN_EXPONENT_G ), powf ( p_position.z, SUN_EXPONENT_B ) );
        return ( weight * MAX_SUN_RGB + ( 1.f - weight ) * MIN_SUN_RGB ) * p_sunBrightness;

    } else {
        return MIN_SUN_RGB * p_sunBrightness;
    }
}


__device__ __host__ __inline__ void possiblyFlipYZ(optix::float3& p_v, unsigned int p_doFlip) {
    if (p_doFlip) {
        float tmp = p_v.y;
        p_v.y = p_v.z;
        p_v.z = tmp;
    }
}

/* compute RGB sky intensity based on sun position and sky brightness */
__device__ __host__ __inline__ optix::float3 calculateSkyRGB ( optix::float3 &p_position, float p_skyBrightness ) {
    if ( p_position.z >= 0.f ) {
        optix::float3 weight = optix::make_float3 ( powf ( p_position.z, SUN_EXPONENT_R ), powf ( p_position.z, SUN_EXPONENT_G ), powf ( p_position.z, SUN_EXPONENT_B ) );
        return ( weight * MAX_SKY_RGB + ( 1.f - weight ) * MIN_SKY_RGB ) * p_skyBrightness;

    } else {
        return MIN_SKY_RGB * p_skyBrightness;
    }
}


/* compute absolute value of all of a float3's components */
__device__ __host__ __inline__ optix::float3 fabsf ( optix::float3 &p_vec ) {
    return optix::make_float3 ( fabsf ( p_vec.x ), fabsf ( p_vec.y ), fabsf ( p_vec.z ) );
}


/* adjust ray start by p_sceneEpsilon in the direction of p_normal */
__device__ __host__ __inline__ optix::float3 getAdjustedRayStart ( optix::float3 &p_hitpoint, optix::float3 &p_normal, float &p_sceneEpsilon ) {
    return p_hitpoint + fmaxf ( 1.f, ( log10f ( fmaxf ( fabsf ( p_hitpoint ) ) ) ) ) * p_normal * p_sceneEpsilon;
}


#ifndef __CUDACC__

#include <optixu/optixpp_namespace.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <sys/time.h>

std::string generatePtxPath ( const std::string &base );


/** @brief create an instance of p_geometry with the desired material, color and transformation */
optix::Transform createObject ( optix::Context p_context, optix::Geometry p_geometry, optix::Material p_material, float p_exponent, optix::float3 p_color, optix::float3 p_translate, optix::float3 p_scale, float p_rotate );

optix::Transform createObject ( optix::Context p_context, optix::Geometry p_geometry, optix::Material p_material, optix::float3 p_color, optix::float3 p_translate, optix::float3 p_scale, float p_rotate );




/** @brief add the elapsed timespan between p_start and p_end to p_elapsedSeconds */
void updateElapsedSeconds ( timespec &p_start, timespec &p_end, double &p_elapsedSeconds );

#endif

#endif /* HELPERS_H_ */
