#include <stdlib.h>
#include <math.h>
#include <iostream>

#include <glm/glm.hpp>
#include <unistd.h>

#include <vector>



void printData(int *data, int sizex, int sizey) {
	for (int y = 0; y < sizex; ++y) {
		for (int x = 0; x < sizey; ++x) {
			if (data[y * sizex + x] != -1) {
				std::cout << "\t" << data[y * sizex + x];
			} else {
				std::cout << "\t" << "-";
			}
		}
		std::cout << std::endl;
	}
}

/**
 *	U: 		spatial domain with d dimensions and u=\bar{u}^d positions (i.e., \bar{u} is the extent per dimension)
 *	S: 		sparse data S \subset U of n grid positions, where each grid position p \in S has an associated data record D(p)
 *			=> data density is the fraction \rho = n / u
 *	H:		d-dimensional hash table, array of size m = \bar{m}^d >= n, containing the data records D(p), d \in S.
 *			\bar{m} is the hash table size per dimension 
 *	h(p):	Perfect hash function h(p): U -> H, this is an injective map when restricted to S, mapping each position p \in S
 *			to a unique slot s = h(p) in the hash table.
 *
 *			h(p) = h_0(p) + \phi(h_1(p)) mod \bar{m}
 *			
 *				- \phi: offset table, d-dimensional array of size r = \bar{r}^d = \sigma * n containing d-dimensional offset vectors
 *				- h0: p -> M_0 * p mod \bar{m} is a simple linear transform with a d^2 matrix M_0 modulo the hash table size
 *				- h1: p -> M_1 * p mod \bar{r} is a simple linear transform with a d^2 matrix M_1 modulo the offset table size
 *				- Here, M_0 and M_1 are usually identity matrices
 */


typedef glm::ivec2 point;
typedef glm::mat2 matrix;

matrix M0(1, 0,
	      0, 1);
	      

matrix M1(1, 0,
          0, 1);




// point ComputeHash(point p, point* offsetMatrix, int mbar, int rbar) {
// 	point h0 = M0 * p;	
// 	point h1 = M1 * p;
// 	point offset = offsetMatrix[(int)h1.y * rbar + (int)h1.x] * oscale;
// 	return h0 + offset;
// 	// TODO offset
// 	// point offset = offset[h1] * oscale;
// 	// return h0 + offset
// 	//return h0 + offset;
// }



// construct hash and offset tables
// table sizes  \bar{m}, \bar{r}, hash coefficients M0, M1 and offset values in \phi are assigned such that:
// - h(p) is a perfect hash function
// - \phi and H are as compact as possible
// - accesses to \phi(h_1(p)) and H(h(p)) have good spatial coherence regarding p
// ;
// - first construct the hash table to be as compact as possible to contain the given data
// - then construct the offset table to be as small as possible while still allowing for a perfect hash function
// offset table is filled with a greedy algorithm, which is iteratively called with different offset table sizes.
// as success is guaranteed if the offset table is large enough, this method always yields a valid result.


int determineHashTableSize(int n, int d, float multiplier) {
	return ceilf(powf(multiplier * n, 1.f / d));
}



int determineOffsetTableSize_fast(int n, int d) {
	float sigma = 1.f / (2 * d);
	return ceilf(powf(n * sigma, 1.f / d));
}


int commonFactor(int a, int b) {
	std::cout << "checking " << a << ", " << b << " for common factor" << std::endl;
	for (int i = 2; i <= (int)fminf(a, b); ++i) {
		if (a % i == 0 && b % i == 0) {
			return i;
		}
	}
	return 1;
}


class DataElement {
public:
	bool valid = false;
	int data = -1;

	DataElement() {
		valid = false;
		data = -1;
	}

	DataElement(int p_data) {
		valid = true;
		data = p_data;
	}
};


int main() {
	time_t rndSeed = time(NULL);
	// time_t rndSeed = 1480590884;
	// time_t rndSeed = 1480592669;
	srand(rndSeed);

	//first try hashing some 2d data
	// int sizex = ((float)rand() / RAND_MAX) * 20 + 10;
	int sizex = 10;
	int sizey = sizex;
	int size = sizex * sizey;
	

	int validEntries = 0;

	float dataDensity = (float)rand() / RAND_MAX * 0.8f;

	int *data = new int[size];
	for (int i = 0; i < size; ++i) {
		if ((float) rand() / RAND_MAX < dataDensity) {
			data[i] = i;
			++validEntries;
		} else {
			data[i] = -1;
		}
	}

	printData(data, sizex, sizey);



	// first determine hash table size, which is m = \bar{m}^d >= n
	int n = validEntries;
	int d = 2;
	
	int mbar = fmaxf(2, determineHashTableSize(n, d, 1.01f));
	int m = powf(mbar, d);

	int rbar = fmaxf(2, determineOffsetTableSize_fast(n, d));
	int r = powf(rbar, d);

	bool fullRebuild = false;
	do {
		if (fullRebuild) {
			++mbar;
			std::cout << "mbar:" << mbar << std::endl;
			sleep(5);
		}

		fullRebuild = false;
		
		
		
		// skip rbar values where rbar has a common factor with mbar or if mbar mod rbar is in {1, rbar - 1}
		int cf = 1;

		while((( cf = commonFactor(mbar, rbar)) != 1) 
		      || mbar % rbar == 1 
		      || mbar % rbar == rbar - 1) {

			if (cf != 1) {
				// std::cout << "COMMON FACTOR: " << cf << std::endl;
			}  else {
				std::cout << "mbar % rbar: " << mbar % rbar << std::endl;
			}
			
			++rbar;

		}

		r = powf(rbar, d);




		std::cout << "validEntries: " << validEntries << std::endl;
		std::cout << "dimension: " << d << std::endl;
		std::cout << "m, mbar: " << m << ", " << mbar << std::endl;
		std::cout << "r, rbar: " << r << ", " << rbar << std::endl;


		// compute offset table

		// determine offset order
		int offsetSizes[rbar * rbar];
		for (int i = 0; i < rbar * rbar; ++i) {
			offsetSizes[i] = 0;
		}


		float avgOffsetSize = 0.f;

		std::vector< std::vector<point> > bucketOrgIndices(r);

		for (int y = 0; y < sizey; ++y) {
			for (int x = 0; x < sizex; ++x) {

				if (data[y * sizex + x] != -1) {
					// int randx = ((float)rand() / RAND_MAX) * (SIZEX - 1);
					// int randy = ((float)rand() / RAND_MAX) * (SIZEY - 1);

					// there is actual data to be dealt with
					point xy = point(x,y);
					point offsetIndex = M1 * xy;
					int offsetIndex1d = offsetIndex.y * rbar + offsetIndex.x;
					offsetIndex1d %= r;
				
					offsetSizes[offsetIndex1d] += 1;

					bucketOrgIndices[offsetIndex1d].push_back(xy);
					avgOffsetSize += 1.f / (rbar * rbar);

					
				}

			}
		}

			// for (int i = 0; i < bucketOrgIndices.size(); ++i) {
			// 			for (int j = 0; j < bucketOrgIndices[i].size(); ++j) {
			// 				point p = bucketOrgIndices[i][j];
			// 				std::cout << "(" << p.x << ", " << p.y << "); ";
			// 			}
			// 			std::cout << std::endl;
					// }
	// return 0;

		printData(offsetSizes, rbar, rbar);
		
		// std::cout << "avg number of data elements per offset entry: " << avgOffsetSize << std::endl;

		uint maxEntries = 0;
		uint maxIndex = 0;


		std::vector< std::vector<point> > hashOrder;
		int curHashOrderIndex = 0;


		do {
			// std::cout << "iterating" << std::endl;
			maxEntries = 0;
			maxIndex = 0;

			// find maximum
			for (uint i = 0; i < bucketOrgIndices.size(); ++i) {
				if (bucketOrgIndices[i].size() > maxEntries) {
					// std::cout << "new max at index " << i << std::endl;
					maxEntries = bucketOrgIndices[i].size();
					maxIndex = i;
					std::cout << "max entries at offset " << maxIndex << ": " << maxEntries << std::endl;

				}
			}

			// std::cout << "max entries at offset " << maxIndex << ": " << maxEntries << std::endl;

			
			std::vector<point> hashOrderForBucket;
			for (uint i = 0; i < bucketOrgIndices[maxIndex].size(); ++i) {
				point p = bucketOrgIndices[maxIndex][i];
				hashOrderForBucket.push_back(p);
				// std::cout << "pushing " << i;
			}
			if (hashOrderForBucket.size() > 0) {
				hashOrder.push_back(hashOrderForBucket);
			}


			// std::cout << std::endl;

			bucketOrgIndices.erase(bucketOrgIndices.begin() + maxIndex);
			// std::cout << "size: " << bucketOrgIndices.size() << std::endl;
			
			// std::cout << "maxEntries: " << maxEntries << std::endl;

			++curHashOrderIndex;

		} while (maxEntries > 0 && bucketOrgIndices.size() > 0);
		// std::cout << "exit loop" << std::endl;

		// hash order is now present
		std::cout << "Hash order: " << std::endl;
		for (uint i = 0; i < hashOrder.size(); ++i) {
			
			for (uint j = 0; j < hashOrder[i].size(); ++j) { 
				std::cout << "(" << hashOrder[i][j].x << ", " << hashOrder[i][j].y << "); ";
			}
			std::cout << std::endl;
		}

		std::cout << std::endl;


		//now determine offsets and stuff ;)
		//offsets originate from the positioning in the hash table!
		
		std::vector<DataElement> emptyHashTable(m);
		std::vector<int> emptyOffsetTable(r);

		//create intermediate hash table
		std::vector<DataElement> iHashTable(m);
		std::vector<int> iOffsetTable = emptyOffsetTable;
		for (int i = 0; i < r; ++i) {
			emptyOffsetTable[i] = 0;
		}

		int curOffset = 0;

		//each bucket of the hash order has a fixed offset, which has to be determined!
		uint i = 0;

		int attemptIndex = 0;

		while(i < hashOrder.size()) {
			// std::cout << "index " << i << " / " << (hashOrder.size() - 1) << std::endl;


			//find valid offset value
			int startOffset = ((float)rand() / RAND_MAX) * m;
			
			curOffset = startOffset;
			bool offsetValid = true;
			

			do {
				offsetValid = true;

				for (uint j = 0; j < hashOrder[i].size(); ++j) {

					// std::cout << "checking pt " << j << "(" << hashOrder[i][j].x << ", " << hashOrder[i][j].y << ")" << std::endl;
 
					point pos = hashOrder[i][j];
					point h0 = (M0 * pos);
					// h0.x %= mbar;
					// h0.y %= mbar;
					
					int h = (h0.y * mbar + h0.x + curOffset) % m;
					// h.x %= mbar;
					// h.y %= mbar;

					//check if position is free
					bool posAvailable = !(iHashTable[h].valid);

					if (!posAvailable) {
						// std::cout << "Position conflict, modifying offset" << std::endl;
						offsetValid = false;
						break;
					}

				}	

				if (!offsetValid) { // check if we have done everything we could :(
					// std::cout << "oldOffset: " << curOffset << std::endl;
					++curOffset;
					curOffset %= m;
					
					if (curOffset == startOffset) {
						std::cout << "Attempt " << attemptIndex++ << ": could not find valid offset; RESTARTING" << std::endl;
						if (attemptIndex > 10) {
							fullRebuild = true;
							break;
						}
						i = 0;
						iHashTable = emptyHashTable;
						iOffsetTable = emptyOffsetTable;
						break;
					}
					// std::cout << "newOffset: " << curOffset << std::endl;	
				} else {
					for (uint j = 0; j < hashOrder[i].size(); ++j) {
						point pos = hashOrder[i][j];
						point h0 = (M0 * pos);
						// h0.x %= mbar;
						// h0.y %= mbar;

						int h = (h0.y * mbar + h0.x + curOffset) % m;
						
						

						// std::cout << "hash pt: (" << pos.x << ", " << pos.y << ") -> (" << h0.x << ", " << h0.y << "), " << (h0.y * mbar + h0.x) << ", " << h << ": " << data[pos.y * sizex + pos.x] << ", offset " << curOffset << std::endl;

						iHashTable[h].valid = true;
						iHashTable[h].data = data[pos.y * sizex + pos.x];
					
					}
					// std::cout << "--- offset " << curOffset << " valid --- " << std::endl;

				}

				// we need to check if valid offset could be found! if not, restart the process!
			} while (!offsetValid);

			if (fullRebuild) {
				break;
			}

			if (offsetValid) {
				// for (uint j = 0; j < hashOrder[i].size(); ++j) {
					point h1 = (M1 * hashOrder[i][0]);
			
					// std::cout << "Writing offset to table: " << curOffset << std::endl;
					iOffsetTable[(h1.y * rbar + h1.x) % r] = curOffset;
				// }
				// std::cout << "\r" << i << "/" << hashOrder.size() << std::endl;
				++i;

			}
			// std::cout << "Found valid offset" << std::endl;


			//store offset value in offset table


			// for (int y = 0; y < mbar; ++y) {
			// 	for (int x = 0; x < mbar; ++x) {
			// 		std::cout << "\t" << iHashTable[y * mbar + x].data;
			// 	}
			// 	std::cout << std::endl;
			// }

			// std::cout << std::endl;

			// for (int y = 0; y < rbar; ++y) {
			// 	for (int x = 0; x < rbar; ++x) {
			// 		std::cout << "\t(" << iOffsetTable[y * rbar + x].x << ", " << iOffsetTable[y * rbar + x].y << ")";
			// 		if (iOffsetTable[y * rbar + x].x < 10 || iOffsetTable[y * rbar + x].y < 10) {
			// 			std::cout << "\t";
			// 		}
			// 	}
			// 	std::cout << std::endl;
			// }

			// std::cout << "Progress: " << i << " / " << hashOrder.size() << std::endl;

		}



		std::cout << "Out of loop" << std::endl;

		bool error = false;


		for (int y = 0; y < mbar; ++y) {
			for (int x = 0; x < mbar; ++x) {
									
				int value = iHashTable[y * mbar + x].data;

				std::cout << value << "\t";
			}
			std::cout << std::endl;
		}


		std::cout << std::endl << std::endl;


		int val1 = 0;
		int val2 = 0;


		// verify hashing by comparing values
		for (int y = 0; y < sizey; ++y) {
			for (int x = 0; x < sizex; ++x) {
				if (data[y * sizex + x] != -1) {
					point h0 = (M0 * point(x, y));
					point h1 = (M1 * point(x, y));
					// h0.x %= mbar;
					// h0.y %= mbar;
					// h1.x %= rbar;
					// h1.y %= rbar;
								
					int offset = iOffsetTable[(h1.y * rbar + h1.x) % r];
					int h = (h0.y * mbar + h0.x + offset) % m;

					int value = iHashTable[h].data;
					std::cout << "offset: " << offset << ", h: " << h << ", value: " << value << " coords: (" << x << ", " << y << ")" << std::endl;

					std::cout << value << "/" << data[y * sizex + x] << "\t ";

					if (value != data[y * sizex + x]) {
						val1 = value;
						val2 = data[y * sizex + x];

						float orgsize = (sizex * sizey) * 4;
						float hashsize = m * 4;
						float offsetsize = r * 8;
						float totalsize = hashsize + offsetsize;
						std::cout << std::endl << "###### ERROR, values not equal (" << val1 << " != " << val2 << ")" << std::endl;
						std::cout << "m=" << m << ", r=" << r << "; compression: " << orgsize << " -> " << totalsize << " Bytes (" << (totalsize * 100.f / orgsize) << "%)" << std::endl;
						std::cout << "Seed: " << rndSeed << std::endl;
						fullRebuild = true;
						break;
					}
				}
				else {
					std::cout << "-\t";
				}

			}
			std::cout << std::endl;
			if (fullRebuild) break;
		}
	} while (fullRebuild);
	
	float orgsize = (sizex * sizey) * 4;
	float hashsize = m * 4;
	float offsetsize = r * 8;
	float totalsize = hashsize + offsetsize;
	std::cout << "SUCCESS, " << "m=" << m << ", r=" << r << "; compression: " << orgsize << " -> " << totalsize << " Bytes (" << (totalsize * 100.f / orgsize) << "%)" << std::endl;


	return 0;
	


}