#include "AssimpLoader.h"
#include <ImageLoader.h>
#include <common_structs.h>




bool AssimpLoader::load_file(optix::Context context, const Path& path, optix::Group sceneGroup, const optix::Matrix4x4 &transform ) {

        _path = new Path(path);

        //check if file exists
        {
                std::ifstream fin(path);

                if(!fin.fail())
                {
                        fin.close();
                }
                else
                {
                        std::cerr << "Error: Couldn't open file: " << path.path() << " : " << importer.GetErrorString() << std::endl;
                        return false;
                }
        }
        _context = context;

        scene_ = importer.ReadFile( path, aiProcessPreset_TargetRealtime_Quality | aiProcess_PreTransformVertices | aiProcess_GenNormals | aiProcess_Triangulate | aiProcess_JoinIdenticalVertices | aiProcess_SortByPType);


        std::string _pathname = _filename.substr ( 0, _filename.find_last_of ( "/\\" ) + 1 );

        // If the import failed, report it
        if( !scene_)
        {
                std::cerr << "Error: " << importer.GetErrorString() << std::endl;
                return false;
        }

        createMaterial();


        //Compute Scene's bounding box
        
        optix::float3 extent = scene_bbox_.m_max - scene_bbox_.m_min;
        float scale = 1.f;//1.f/std::max(std::max(extent.x, extent.y), extent.z);

        scene_center_ = scene_bbox_.m_min + (scene_bbox_.m_max * 0.5f);

        // Now we can access the file's contents.
        std::cout << "Import of scene " << path.path() << " succeeded." << std::endl;

        bool retval = read_assimp(context, sceneGroup, _material, scale, transform);



        return retval;
}


bool AssimpLoader::read_assimp(optix::Context context, optix::Group& sceneGroup, optix::Material* materials, float scale, const optix::Matrix4x4 &transform ) {


        if ( !_intersect_program.get() ) {
            std::string path = std::string ( generatePtxPath ( "triangle_mesh.cu" ) );
            _intersect_program = _context->createProgramFromPTXFile ( path, "mesh_intersect" );
        }

        if ( !_bbox_program.get() ) {
            std::string path = std::string ( generatePtxPath ( "triangle_mesh.cu" ) );
            _bbox_program = _context->createProgramFromPTXFile ( path, "mesh_bounds" );
        }


        std::vector<GeometryInstance> instances;

        struct AssimpMaterial aMat;

        static unsigned int primID=0;

       
        //TODO maybe sort meshes by material to cluster them for better efficiency
        std::cout << scene_->mNumMeshes << std::endl;
      

        unsigned int globalFaceIndex = 0;

        for (unsigned int n = 0; n < scene_->mNumMeshes; ++n) {
            bool validMesh = true;
            std::cout << "Mesh number: " << n << std::endl;
            const aiMesh *mesh = scene_->mMeshes[n];

            unsigned int numTrisInMesh = 0;

            for (unsigned int t = 0; t < mesh->mNumFaces; ++t) {
                const aiFace *face = &mesh->mFaces[t];
                if (face->mNumIndices != 3) {
                    validMesh = false;
                    break;
                } else {
                    ++numTrisInMesh;
                }
            }
            if (!validMesh) continue;

            Buffer vindex_buffer = context->createBuffer( RT_BUFFER_INPUT, RT_FORMAT_INT3, numTrisInMesh );
            int3 *vindex_buffer_data = static_cast<int3 *> ( vindex_buffer->map() );

            /** findex_buffer stores global face indices needed for compatibility with preprocessing step */
            Buffer findex_buffer = context->createBuffer( RT_BUFFER_INPUT, RT_FORMAT_INT, numTrisInMesh );
            int *findex_buffer_data = static_cast<int*>( findex_buffer->map() );
           

            optix::Geometry optixMesh;

            unsigned int faceIndex = 0;
            int wrong_face_mode = 0;

            for (unsigned int t = 0; t < mesh->mNumFaces; ++t) {
              

                const aiFace *face = &mesh->mFaces[t];

                if (face->mNumIndices != 3) {

                    std::cerr << "Wrong face mode: " << face->mNumIndices << std::endl;
                    std::cerr << "Only triangles are supported" << std::endl;
                    exit(1);
                } else {
                    vindex_buffer_data[faceIndex] = make_int3(face->mIndices[0], face->mIndices[1], face->mIndices[2]);
                    findex_buffer_data[faceIndex] = globalFaceIndex;
                    ++faceIndex;
                    ++globalFaceIndex;
                }
                
            }

            vindex_buffer->unmap();
            findex_buffer->unmap();


            optix::Buffer vbuffer = context->createBuffer( RT_BUFFER_INPUT, RT_FORMAT_FLOAT3, mesh->mNumVertices);
            if (mesh->HasPositions()) {

                float3 *vbuffer_h = static_cast<float3 *> ( vbuffer->map() );
                std::cout << "mesh->mNumVertices: " << mesh->mNumVertices << std::endl;

                // imba_mesh.set_vertex_count(imba_mesh.vertex_count() + mesh->mNumVertices);
                for(int i=0; i < mesh->mNumVertices; i++) {

                    vbuffer_h[i] = make_float3((mesh->mVertices[i].x), (mesh->mVertices[i].y), (mesh->mVertices[i].z));
                    float4 v4 = make_float4(vbuffer_h[i], 1.f);
                    vbuffer_h[i] = make_float3(transform * v4);
                    scene_bbox_.include(vbuffer_h[i]);

                }

                vbuffer->unmap();

            }

            const optix::Matrix4x4 norm_transform = transform.inverse().transpose();

            optix::Buffer nbuffer = context->createBuffer( RT_BUFFER_INPUT, RT_FORMAT_FLOAT3, mesh->mNumVertices);
            
            if (mesh->HasNormals()) {

                float3 *nbuffer_h = static_cast<float3 *> ( nbuffer->map() );

                for (int i=0; i < mesh->mNumVertices; ++i) {
                    nbuffer_h[i] = make_float3(mesh->mNormals[i].x, mesh->mNormals[i].y, mesh->mNormals[i].z);
                    float4 v4 = make_float4(nbuffer_h[i], 0.f);
                    nbuffer_h[i] = make_float3(norm_transform * v4);
                }

                nbuffer->unmap();

            }


            optix::Buffer tbuffer = context->createBuffer( RT_BUFFER_INPUT, RT_FORMAT_FLOAT2, mesh->mNumVertices);
           // buffer for vertex texture coordinates
            if (mesh->HasTextureCoords(0)) {
                float2 *tbuffer_h = static_cast<float2 *> ( tbuffer->map() );

                //memcpy(uv_attrib_ptr,  mesh->mTextureCoords[0], sizeof(float)*2*mesh->mNumVertices);
                for(int i=0; i < mesh->mNumVertices; i++) {
                    tbuffer_h[i] = make_float2(mesh->mTextureCoords[0][i].x, mesh->mTextureCoords[0][i].y);
                }

                tbuffer->unmap();
            }


 //         imba::float4 c(0.8f, 0.8f, 0.8f, 1.0f);
        //         aiColor4D diffuse;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_DIFFUSE, &diffuse))
        //                 c = conv_aiColor_to_float4(diffuse);
        //         imba_mat->setDiffuse(c);

        //         c = imba::float4(0.2f, 0.2f, 0.2f, 1.0f);
        //         aiColor4D ambient;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_AMBIENT, &ambient))
        //                 c = conv_aiColor_to_float4(ambient);
        //         imba_mat->setAmbient(c);

        //         c = imba::float4(0.f, 0.f, 0.f, 1.0f);
        //         aiColor4D specular;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_SPECULAR, &specular))
        //                 c = conv_aiColor_to_float4(specular);
        //         imba_mat->setSpecular(c);

        //         c = imba::float4(0.f, 0.f, 0.f, 1.0f);
        //         aiColor4D emission;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_EMISSIVE, &emission))
        //                 c = conv_aiColor_to_float4(emission);
        //         imba_mat->setEmissive(c);

        //         float shininess = 1.0;
        //         unsigned int max;
        //         if(AI_SUCCESS == aiGetMaterialFloatArray(mtl, AI_MATKEY_SHININESS, &shininess, &max)) {
        //                 //std::cout << "shininess: " << shininess << " max: " << max << std::endl;
        //                 //TODO: DIVISON By FOUR ONLY VALID FOR OBJ?
        //                 imba_mat->setShininess(shininess / 4.0);
        //         }
            

           

                  
            
                // if (shininess > 1.f && specular.r + specular+g + specular.b > 0.f) {
                //     ++propertyCount;

                //     if ()
                // }
            // }

            // TODO add alpha
            // std::cout << "Material index: " << matIndex << std::endl;

            optix::Buffer mbuffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_INT, numTrisInMesh );
            unsigned int* mbuffer_h = static_cast<unsigned int *> (mbuffer->map());
            for (unsigned int t = 0; t < mesh->mNumFaces; ++t) {
                // TODO we need to know here which material type we deal with
                mbuffer_h[t] = 0;
            }
            mbuffer->unmap();
            
            std::cout << "Creating optix mesh" << std::endl;
            optixMesh = context->createGeometry();
            optixMesh->setPrimitiveCount(numTrisInMesh);
            optixMesh->setIntersectionProgram( _intersect_program );
            optixMesh->setBoundingBoxProgram( _bbox_program );
            optixMesh["vertex_buffer"]->setBuffer(vbuffer);
            optixMesh["vindex_buffer"]->setBuffer(vindex_buffer);
            optixMesh["normal_buffer"]->setBuffer(nbuffer);
            optixMesh["texcoord_buffer"]->setBuffer(tbuffer);
            optixMesh["material_buffer"]->setBuffer(mbuffer);
            optixMesh["findex_buffer"]->setBuffer(findex_buffer);



            // TODO material processing goes here

            unsigned int matIndex = ENUM_MAT_TYPE_UNIVERSAL;
         
           
         
  // aiString texPath; //contains filename of texture
        //         if(AI_SUCCESS == mtl->GetTexture(aiTextureType_DIFFUSE, 0, &texPath)) {
        //                 std::cout << "Diffuse Texture found: " << texPath.data << std::endl;
        //                 //imba_mat->setTexCount(1);
        //                 imba_mat->setTexPathDiffuse(texPath.data);
        //         } /*else {
        //              imba_mat->setTexCount(0);
        //              }   */

        //         if(AI_SUCCESS == mtl->GetTexture( aiTextureType_OPACITY, 0, &texPath)) {
        //                 std::cout << "Opacity Texture count" << texPath.data << std::endl;
        //                 //imba_mat->setTexCount(1);
        //                 imba_mat->setTexPathOpacity(texPath.data);
        //         } /*else {
        //              imba_mat->setTexCount(0);
        //              }       */

        //         // TR: We handle bump maps as normal maps, too, as normal maps are mostly given as map_bump in our obj files
        //         if (AI_SUCCESS == mtl->GetTexture(aiTextureType_NORMALS, 0, &texPath)) {
        //                 std::cout << "NORMAL/HEIGHT MAP FOUND: " << texPath.C_Str() << std::endl;
        //                 imba_mat->setTexPathNormal(texPath.data);
        //         } else if (AI_SUCCESS == mtl->GetTexture(aiTextureType_HEIGHT, 0, &texPath)) {
        //                 std::cout << "NORMAL/HEIGHT MAP FOUND: " << texPath.C_Str() << std::endl;
        //                 imba_mat->setTexPathNormal(texPath.data);
        //         }
            aiMaterial *mtl = scene_->mMaterials[mesh->mMaterialIndex]; 
            unsigned int propertyCount = 0;


            std::cout << "mtl: " << mtl << std::endl;

           
               std::cout << "Reading material" << std::endl;
    
               aiString matName;
               aiGetMaterialString(mtl, AI_MATKEY_NAME, &matName);
               std::cout << "material name: " << matName.data << std::endl;
               std::cout << mtl << std::endl;

                unsigned int max;

               
                aiColor4D ambient(0.f, 0.f, 0.f, 0.f);
                aiColor4D diffuse(1.f, 1.f, 1.f, 1.f);
                aiColor4D specular(0.f, 0.f, 0.f, 0.f);
                aiColor4D transparency(1.f, 1.f, 1.f, 1.f);
                float shininess = 32.f;
                float refraction = 1.f;
                float opacity = 1.f;



                bool has_shininess = aiGetMaterialFloat(mtl, AI_MATKEY_SHININESS, &shininess) == AI_SUCCESS;
                bool has_refraction = aiGetMaterialFloat(mtl, AI_MATKEY_REFRACTI, &refraction) == AI_SUCCESS;
                bool has_opacity = aiGetMaterialFloat(mtl, AI_MATKEY_OPACITY, &opacity) == AI_SUCCESS;
        // shininess /= 4.f;
                std::cout << "has_opacity: " << opacity << std::endl;
      
                // if (has_diffuse && (diffuse.r + diffuse.g + diffuse.b > 0.f)) {
                if (AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_DIFFUSE, &diffuse)) {
                    if ((diffuse.r + diffuse.g + diffuse.b > 0.f)) {
                        ++propertyCount;
                        std::cout << "has diffuse: " << diffuse.r << ", " << diffuse.g << ", " << diffuse.b << std::endl;
                        matIndex = ENUM_MAT_TYPE_DIFFUSE;
                    }
                }

                if (AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_SPECULAR, &specular)) {
                    if ((specular.r + specular.g + specular.b > 0.f)) {
                        ++propertyCount;
                        std::cout << "has specular: " << specular.r << ", " << specular.g << ", " << specular.b << "; shininess: " << shininess << std::endl;
                        if (shininess >= 10000.f) {
                            matIndex = ENUM_MAT_TYPE_MIRROR;
                        } else {
                            matIndex = ENUM_MAT_TYPE_GLOSSY;
                        }
                    }
                }


                if (AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_TRANSPARENT, &transparency)) {
                    if (transparency.r + transparency.g + transparency.b > 0.f && opacity < 1.f) {
                        ++propertyCount;
                        std::cout << "has transparency: " << transparency.r << ", " << transparency.g << ", " << transparency.b << std::endl;
                        matIndex = ENUM_MAT_TYPE_GLASS;    
                    }                    
                }


                if ( propertyCount > 1 ) {
                    std::cout << "propertyCount > 1!" << std::endl;
                    std::cerr << "MATERIALS WITH MORE THAN ONE PROPERTY (LAYERED MATERIALS) ARE CURRENTLY NOT SUPPORTED FOR CACHING!" << std::endl;
                    // exit(1);
                    matIndex = ENUM_MAT_TYPE_UNIVERSAL;
                } else if (propertyCount == 0) {
                    std::cout << "No properties found" << std::endl;
                }

                GeometryInstance instance = context->createGeometryInstance(optixMesh, &(_material[matIndex]), &(_material[matIndex])+1);

          





                aiString texPath; //contains filename of texture

                if(AI_SUCCESS == mtl->GetTexture(aiTextureType_DIFFUSE, 0, &texPath)) {
                    std::cout << "Diffuse Texture found: " << texPath.data << std::endl;
                    instance["diffuse_map"]->setTextureSampler ( loadTexture ( _context, _path->base_name() + std::string("/") + texPath.data, make_float3 ( 1.f) ) );
                } else {
                    instance["diffuse_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 1.f) ) );
                }
                


                
                if (AI_SUCCESS == mtl->GetTexture(aiTextureType_SPECULAR, 0, &texPath)) {
                    std::cout << "Specular Texture found: " << texPath.data << std::endl;
                    instance["specular_map"]->setTextureSampler ( loadTexture ( _context, _path->base_name() + std::string("/") + texPath.data, make_float3 ( 1.0f ) ) );
                } else {
                    instance["specular_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 1.0f ) ) );
                }
                

           

                if (AI_SUCCESS == mtl->GetTexture(aiTextureType_OPACITY, 0, &texPath)) {
                    std::cout << "Opacity Texture found: " << texPath.data << std::endl;
                     instance["dissolve_map"]->setTextureSampler ( loadTexture ( _context, _path->base_name() + std::string("/") + texPath.data, make_float3 ( 1.0f) ) );
                } else {
                     instance["dissolve_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 1.0f) ) );
                }
               
                
                bool foundNormals = false;
                if (AI_SUCCESS == mtl->GetTexture(aiTextureType_NORMALS, 0, &texPath)) {
                    std::cout << "Normal Texture found: " << texPath.data << std::endl;
                    instance["bump_map"]->setTextureSampler ( loadTexture ( _context, _path->base_name() + std::string("/") + texPath.data, make_float3 ( 0.f ) ) );
                } else if (AI_SUCCESS == mtl->GetTexture(aiTextureType_HEIGHT, 0, &texPath)) {
                    std::cout << "Normal Texture found: " << texPath.data << std::endl;
                    instance["bump_map"]->setTextureSampler ( loadTexture ( _context, _path->base_name() + std::string("/") + texPath.data, make_float3 ( 0.f ) ) );
                } else {
                    instance["bump_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 0.f ) ) );
                }
                


                instance["phong_exp"]->setFloat ( shininess );
                //    gi[ "reflectivity" ]->setFloat( 0.0f, 0.0f, 0.0f );
                //    gi[ "illum" ]->setInt( 2 );
                instance["kd"]->setFloat ( diffuse.r, diffuse.g, diffuse.b );
                instance["ks"]->setFloat ( specular.r, specular.g, specular.b );
                instance["tf"]->setFloat ( transparency.r, transparency.g, transparency.b );
                instance["alpha"]->setFloat ( opacity );
                instance["ni"]->setFloat ( refraction );


            std::cout << "Material index set to " << matIndex << std::endl;

            //    gi["ambient_map"]->setTextureSampler( loadTexture( _context, "", make_float3( 0.2f, 0.2f, 0.2f ) ) );
          
            

            optix::GeometryGroup instanceGroup = _context->createGeometryGroup();
            instanceGroup->setChildCount(1);
            instanceGroup->setChild(0, instance);
            optix::Acceleration acceleration = context->createAcceleration("Sbvh", "Bvh");
            acceleration->setProperty("vertex_buffer_name", "vertex_buffer");
            acceleration->setProperty("index_buffer_name", "vindex_buffer");
            instanceGroup->setAcceleration(acceleration);

            const unsigned int current_child_count = sceneGroup->getChildCount();
            sceneGroup->setChildCount(current_child_count + 1);
            sceneGroup->setChild(current_child_count, instanceGroup);

        }   
        // exit(1);


        return true;
    }
// {

        // // For each mesh
        // for (unsigned int n = 0; n < scene_->mNumMeshes; ++n) {
        //         const aiMesh* mesh = scene_->mMeshes[n];
        //         size_t meshVertexOffset =  imba_mesh.vertex_count();

        //         size_t faceIndex = imba_mesh.index_count();
        //         imba_mesh.set_index_count(imba_mesh.index_count() + mesh->mNumFaces * 3);

        //         int wrong_face_mode = 0;
        //         for (unsigned int t = 0; t < mesh->mNumFaces; ++t) {
        //                 const aiFace* face = &mesh->mFaces[t];

        //                 if(face->mNumIndices != 3) {
        //                         std::cerr << "Wrong face mode: " << face->mNumIndices << std::endl;
        //                         std::cerr << "Only Triangles are supported" << std::endl;
        //                         wrong_face_mode++;
        //                 } else {
        //                         //memcpy(&(imba_mesh.indices()[faceIndex]), face->mIndices, 3 * sizeof(uint32_t));
        //                         imba_mesh.indices()[faceIndex + 0] = meshVertexOffset + face->mIndices[0];
        //                         imba_mesh.indices()[faceIndex + 1] = meshVertexOffset + face->mIndices[1];
        //                         imba_mesh.indices()[faceIndex + 2] = meshVertexOffset + face->mIndices[2];
        //                         faceIndex += 3;
        //                 }
        //         }


        //         imba_mesh.set_index_count(imba_mesh.index_count() - wrong_face_mode * 3);

        //         size_t vertex_index = imba_mesh.vertex_count();

        //         // generate Vertex Array for mesh
        //         // buffer for vertex positions
        //         if (mesh->HasPositions()) {
        //                 std::cout << "mesh->mNumVertices: " << mesh->mNumVertices << std::endl;

        //                 imba_mesh.set_vertex_count(imba_mesh.vertex_count() + mesh->mNumVertices);
        //                 for(int i=0; i < mesh->mNumVertices; i++) {
        //                         imba_mesh.vertices()[vertex_index + i] = imba::float4((mesh->mVertices[i].x - scene_center_.x) * scale, (mesh->mVertices[i].y - scene_center_.y) * scale, (mesh->mVertices[i].z - scene_center_.z) * scale, 1.f);
        //                 }
        //         }

        //         // buffer for vertex normals
        //         if (mesh->HasNormals()) {
        //                 uint8_t* norm_attrib_ptr = &(imba_mesh.attribute(0)[vertex_index * imba_mesh.attribute_stride(0)]);
        //                 memcpy(norm_attrib_ptr, mesh->mNormals, sizeof(float)*3*mesh->mNumVertices);
        //                 // uint8_t* tan_attrib_ptr = &(imba_mesh.attribute(3)[vertex_index * imba_mesh.attribute_stride(3)]);
        //                 // memcpy(tan_attrib_ptr, mesh->mTangents, sizeof(float)*3*mesh->mNumVertices);
        //                 // uint8_t* bitan_attrib_ptr = &(imba_mesh.attribute(4)[vertex_index * imba_mesh.attribute_stride(4)]);
        //                 // memcpy(bitan_attrib_ptr, mesh->mBitangents, sizeof(float)*3*mesh->mNumVertices);
        //         }

        //         // buffer for vertex texture coordinates
        //         if (mesh->HasTextureCoords(0)) {
        //                 imba::float2* uv_attrib_ptr = (imba::float2*) &(imba_mesh.attribute(1)[vertex_index * imba_mesh.attribute_stride(1)]);
        //                 //memcpy(uv_attrib_ptr,  mesh->mTextureCoords[0], sizeof(float)*2*mesh->mNumVertices);
        //                 for(int i=0; i < mesh->mNumVertices; i++) {
        //                         uv_attrib_ptr[i] = imba::float2(mesh->mTextureCoords[0][i].x, mesh->mTextureCoords[0][i].y);
        //                 }
        //         }


        //         // buffer for material indices
        //         uint32_t* mat_attrib_ptr = (uint32_t*) &(imba_mesh.attribute(2)[vertex_index * imba_mesh.attribute_stride(2)]);
        //         for(int i = 0; i < mesh->mNumVertices; i++) {
        //                 mat_attrib_ptr[i] = mesh->mMaterialIndex;
        //         }

        //         std::cout << "Processing Vertex Colors!" << std::endl;
        //         imba::float4* vcol_attrib_ptr = (imba::float4*) &(imba_mesh.attribute(3)[vertex_index * imba_mesh.attribute_stride(3)]);
        //         if (mesh->HasVertexColors(0)) {
        //                 // memcpy(vcol_attrib_ptr, mesh->mColors[0].data(), sizeof(float)*4*mesh->mNumVertices);

        //                 for (int i = 0; i < mesh->mNumVertices; ++i) {
        //                         vcol_attrib_ptr[i].x = mesh->mColors[0][i].r;
        //                         vcol_attrib_ptr[i].y = mesh->mColors[0][i].g;
        //                         vcol_attrib_ptr[i].z = mesh->mColors[0][i].b;
        //                         vcol_attrib_ptr[i].w = mesh->mColors[0][i].a;
        //                         //std::cout << "vertex color (mesh " << n << ", vertex " << i << "): " << mesh->mColors[0][i].r << ", " << mesh->mColors[0][i].g << ", " << mesh->mColors[0][i].b << std::endl;
        //                 }
        //                 // exit(1);
        //         } else {
        //                 for (int i = 0; i < mesh->mNumVertices; ++i) {
        //                         vcol_attrib_ptr[i].x = 1.f;
        //                         vcol_attrib_ptr[i].y = 1.f;
        //                         vcol_attrib_ptr[i].z = 1.f;
        //                         vcol_attrib_ptr[i].w = 1.f;
        //                         //std::cout << "Setting vertex color " << i << " to default" << std::endl;
        //                         // std::cout << "vertex color (mesh " << n << ", vertex " << i << "): " << mesh->mColors[0][i].r << ", " << mesh->mColors[0][i].g << ", " << mesh->mColors[0][i].b << std::endl;
        //                 }
        //         }


        // }
 // }

        //Lambda function for conversion
        // auto conv_aiColor_to_float4 = [] (const aiColor4D &c)->imba::float4 {
        //         return imba::float4(c.r, c.g, c.b, c.a);
        // };

        // ////
        // //Read materials from assimp

        // for(int i = 0; i < scene_->mNumMaterials; i++) {
        //         // create material uniform buffer
        //         aiMaterial *mtl = scene_->mMaterials[i];

        //         imba::AssimpMaterial * imba_mat = new imba::AssimpMaterial();

        //         aiString texPath; //contains filename of texture
        //         if(AI_SUCCESS == mtl->GetTexture(aiTextureType_DIFFUSE, 0, &texPath)) {
        //                 std::cout << "Diffuse Texture found: " << texPath.data << std::endl;
        //                 //imba_mat->setTexCount(1);
        //                 imba_mat->setTexPathDiffuse(texPath.data);
        //         } /*else {
        //              imba_mat->setTexCount(0);
        //              }	*/

        //         if(AI_SUCCESS == mtl->GetTexture( aiTextureType_OPACITY, 0, &texPath)) {
        //                 std::cout << "Opacity Texture count" << texPath.data << std::endl;
        //                 //imba_mat->setTexCount(1);
        //                 imba_mat->setTexPathOpacity(texPath.data);
        //         } /*else {
        //              imba_mat->setTexCount(0);
        //              }		*/

        //         // TR: We handle bump maps as normal maps, too, as normal maps are mostly given as map_bump in our obj files
        //         if (AI_SUCCESS == mtl->GetTexture(aiTextureType_NORMALS, 0, &texPath)) {
        //                 std::cout << "NORMAL/HEIGHT MAP FOUND: " << texPath.C_Str() << std::endl;
        //                 imba_mat->setTexPathNormal(texPath.data);
        //         } else if (AI_SUCCESS == mtl->GetTexture(aiTextureType_HEIGHT, 0, &texPath)) {
        //                 std::cout << "NORMAL/HEIGHT MAP FOUND: " << texPath.C_Str() << std::endl;
        //                 imba_mat->setTexPathNormal(texPath.data);
        //         }

        //         imba::float4 c(0.8f, 0.8f, 0.8f, 1.0f);
        //         aiColor4D diffuse;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_DIFFUSE, &diffuse))
        //                 c = conv_aiColor_to_float4(diffuse);
        //         imba_mat->setDiffuse(c);

        //         c = imba::float4(0.2f, 0.2f, 0.2f, 1.0f);
        //         aiColor4D ambient;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_AMBIENT, &ambient))
        //                 c = conv_aiColor_to_float4(ambient);
        //         imba_mat->setAmbient(c);

        //         c = imba::float4(0.f, 0.f, 0.f, 1.0f);
        //         aiColor4D specular;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_SPECULAR, &specular))
        //                 c = conv_aiColor_to_float4(specular);
        //         imba_mat->setSpecular(c);

        //         c = imba::float4(0.f, 0.f, 0.f, 1.0f);
        //         aiColor4D emission;
        //         if(AI_SUCCESS == aiGetMaterialColor(mtl, AI_MATKEY_COLOR_EMISSIVE, &emission))
        //                 c = conv_aiColor_to_float4(emission);
        //         imba_mat->setEmissive(c);

        //         float shininess = 1.0;
        //         unsigned int max;
        //         if(AI_SUCCESS == aiGetMaterialFloatArray(mtl, AI_MATKEY_SHININESS, &shininess, &max)) {
        //                 //std::cout << "shininess: " << shininess << " max: " << max << std::endl;
        //                 //TODO: DIVISON By FOUR ONLY VALID FOR OBJ?
        //                 imba_mat->setShininess(shininess / 4.0);
        //         }

        //         imba_materials.push_back(std::unique_ptr<imba::AssimpMaterial> (imba_mat));
        // }

        // return true;
// }



void AssimpLoader::createMaterial() {
     std::cout << "- alpha_occlusion_teser" << std::endl;
    optix::Program alpha_occlusion_tester = _context->createProgramFromPTXFile ( generatePtxPath ( "occlusion_shader.cu" ), "alpha_occlusion_tester" );
    std::cout << "- alpha_tester" << std::endl;
    optix::Program alpha_tester = _context->createProgramFromPTXFile ( generatePtxPath ( "alpha_tester.cu" ), "alpha_tester" );
    std::cout << "- shadow_shader" << std::endl;
    optix::Program shadow_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "light_shader.cu" ), "shadow_shader" );
    std::cout << "Creating universal_shader:" << std::endl;
    optix::Material universalMaterial = _context->createMaterial();
    std::cout << "- universal_shader" << std::endl;
    optix::Program universal_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "universal_shader.cu" ), "universal_shader" );
    std::cout << "- setting hit programs" << std::endl;
    universalMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, universal_shader );
    universalMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    universalMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    universalMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating diffuse_shader:" << std::endl;
    optix::Material diffuseMaterial = _context->createMaterial();
    std::cout << "- diffuse_shader" << std::endl;
    optix::Program diffuse_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "diffuse_shader.cu" ), "diffuse_shader" );
    diffuseMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, diffuse_shader );
    diffuseMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    diffuseMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    diffuseMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating mirror_shader:" << std::endl;
    optix::Material mirrorMaterial = _context->createMaterial();
    std::cout << "- mirror_shader" << std::endl;
    optix::Program mirror_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "mirror_shader.cu" ), "mirror_shader" );
    mirrorMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, mirror_shader );
    mirrorMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    mirrorMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    mirrorMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating glossy_shader:" << std::endl;
    optix::Material glossyMaterial = _context->createMaterial();
    std::cout << "- glossy_shader" << std::endl;
    optix::Program glossy_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "glossy_shader.cu" ), "glossy_shader" );
    glossyMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, glossy_shader );
    glossyMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    glossyMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    glossyMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating glass_shader:" << std::endl;
    optix::Material glassMaterial = _context->createMaterial();
    std::cout << "- glass_shader" << std::endl;
    optix::Program glass_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "glass_shader.cu" ), "glass_shader" );
    glassMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, glass_shader );
    glassMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    glassMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    glassMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    _material[ENUM_MAT_TYPE_UNIVERSAL] = universalMaterial;
    _material[ENUM_MAT_TYPE_DIFFUSE]    = diffuseMaterial;
    _material[ENUM_MAT_TYPE_MIRROR]     = mirrorMaterial;
    _material[ENUM_MAT_TYPE_GLOSSY]     = glossyMaterial;
    _material[ENUM_MAT_TYPE_GLASS]      = glassMaterial;
}