#include <ContextWrapper.h>

#include <vector>
#include <iostream>

#include <optixu/optixpp_namespace.h>
#include <common_structs.h>

#include <Camera.h>

#include <helpers.h>
#include <assert.h>


ContextWrapper::ContextWrapper() {
    __m_isInitialized = false;
}

ContextWrapper::ContextWrapper ( unsigned int p_renderWidth, unsigned int p_renderHeight ) {
    initialize ( p_renderWidth, p_renderHeight );
}


ContextWrapper::ContextWrapper ( optix::Context p_context, unsigned int p_renderWidth, unsigned int p_renderHeight ) {
    initialize ( p_context, p_renderWidth, p_renderHeight );
}


void ContextWrapper::initialize ( unsigned int p_renderWidth, unsigned int p_renderHeight ) {
    assert ( !__m_isInitialized );
    __m_context = optix::Context::create();
    __m_maxDepth = 3;
    __m_renderWidth = p_renderWidth;
    __m_renderHeight = p_renderHeight;
    __m_isInitialized = true;
}


void ContextWrapper::initialize ( optix::Context p_context, unsigned int p_renderWidth, unsigned int p_renderHeight ) {
    assert ( !__m_isInitialized );
    __m_context = p_context;
    __m_maxDepth = 3;
    __m_renderWidth = p_renderWidth;
    __m_renderHeight = p_renderHeight;
    __m_isInitialized = true;
}


ContextWrapper::~ContextWrapper() {
    __m_context->destroy();
}


void ContextWrapper::initializeContext() {
    __m_context->setPrintEnabled ( false );
    __m_context->setExceptionEnabled ( RT_EXCEPTION_ALL, false );
    __m_context->setStackSize ( 4000 );
    // initialize used values so __m_context is not recompiled
    __m_context["frame"]->setUint ( 0 );
    __m_context["maxDepth"]->setUint ( 0 );
    __m_context["scene_epsilon"]->setFloat ( 1.e-3f );
    __m_context["sunScale"]->setFloat ( 0 );
    __m_context["sunBrightness"]->setFloat ( 0 );
    __m_context["skyBrightness"]->setFloat ( 0 );
    __m_context["lightProb_diffuse"]->setFloat ( 0.5f );
    __m_context["lightProb_specular"]->setFloat ( 0.5f );
    __m_context["scene_extent"]->setFloat ( 0.f );
    __m_context["eye"]->setFloat ( optix::make_float3 ( 0.f ) );
    __m_context["U"]->setFloat ( optix::make_float3 ( 0.f ) );
    __m_context["V"]->setFloat ( optix::make_float3 ( 0.f ) );
    __m_context["W"]->setFloat ( optix::make_float3 ( 0.f ) );
    __m_context["directionalLight_direction"]->setFloat ( optix::make_float3 ( 0.f ) );
    __m_context["directionalLight_fluxRGB"]->setFloat ( optix::make_float3 ( 0.f ) );
    __m_context["numPreviousFrames"]->setUint ( 0 );
    __m_context["envFlipYZ"]->setUint(1);
    __m_context["focalDist"]->setFloat ( 1.f );
    __m_context["apertureRadius"]->setFloat ( 0.f );
    __m_context["envRotPhi"]->setFloat ( 0.f );
    __m_context["realResolutionX"]->setUint ( __m_renderWidth );
    __m_context["realResolutionY"]->setUint ( __m_renderHeight );
    __m_context["addRays"]->setUint ( 0 );
    __m_context["useLdSampling_image"]->setUint ( 1 );
    __m_context["useLdSampling_lens"]->setUint ( 1 );
    __m_context["useLdSampling_sun"]->setUint ( 1 );
    __m_context["useLdSampling_environment"]->setUint ( 1 );
    __m_context["useLdSampling_integrator"]->setUint ( 1 );
    __m_context["projectionType"]->setUint ( 0 );
    __m_context["samplesPerLaunch"]->setUint ( 1 );
    __m_context["depthSampleX"]->setFloat ( __m_renderWidth / 2.f );
    __m_context["depthSampleY"]->setFloat ( __m_renderHeight / 2.f );
    __m_context["frame"]->setUint ( 0 );
    __m_context["numPreviousFrames"]->setUint ( 0 );
    /* set ray types:
    	* - pt_ray_type, used for normal path tracing
    	* - occlusion_ray_type, used for sampling light sources where
    	*   the intensity is only dependent on light source distance and
    	*   incoming direction (e.g. environment, sun, normal point light)
    	* - shadow_ray_type, used for all other light source types, e.g.
    	*   geometrically defined lights
    	*/
    std::cout << "Setting Ray Type Count to " << ENUM_RAYTYPE_SIZE << std::endl;
    __m_context->setRayTypeCount ( ENUM_RAYTYPE_SIZE );
    __m_context["pt_ray_type"]->setUint ( ENUM_RAYTYPE_PT );
    __m_context["occlusion_ray_type"]->setUint ( ENUM_RAYTYPE_OCCLUSION );
    __m_context["shadow_ray_type"]->setUint ( ENUM_RAYTYPE_SHADOW );
    std::cout << "pt_ray_type: " << ENUM_RAYTYPE_PT << std::endl;
    std::cout << "occlusion_ray_type: " << ENUM_RAYTYPE_OCCLUSION << std::endl;
    std::cout << "shadow_ray_type: " << ENUM_RAYTYPE_SHADOW << std::endl;
}


void ContextWrapper::createOptixPrograms() {
    std::cout << "Setting entry point count" << std::endl;
    /* two entry point types:
    	* - path tracing camera
    	* - refocusing camera
    	*/
    __m_context->setEntryPointCount ( ENUM_NUM_ENTRYPOINTS );
    /* create programs */
    std::cout << "Creating pt_camera program for generating lightfield data" << std::endl;
    optix::Program pt_camera = __m_context->createProgramFromPTXFile ( generatePtxPath ( "pt_camera_separate_buffers.cu" ), "pt_camera" );
    __m_context->setRayGenerationProgram ( ENUM_ENTRYPOINT_PT_LF, pt_camera );
	std::cout << "Creating refocus_camera program" << std::endl;
    optix::Program refocus_camera = __m_context->createProgramFromPTXFile ( generatePtxPath ( "refocus_camera.cu" ), "refocus_camera" );
    __m_context->setRayGenerationProgram ( ENUM_ENTRYPOINT_REFOCUS, refocus_camera );
	std::cout << "Creating pt_camera program for generating lightfield data" << std::endl;
    optix::Program pt_camera_nolf = __m_context->createProgramFromPTXFile ( generatePtxPath ( "pt_camera_separate_buffers_nolf.cu" ), "pt_camera" );
    __m_context->setRayGenerationProgram ( ENUM_ENTRYPOINT_PT, pt_camera_nolf );
    std::cout << "Setting background shader" << std::endl;
    optix::Program background_shader = __m_context->createProgramFromPTXFile ( generatePtxPath ( "background_shader.cu" ), "background_shader" );
    __m_context->setMissProgram ( ENUM_RAYTYPE_PT, background_shader );
}


void ContextWrapper::setMaxDepth ( unsigned int p_maxDepth ) {
    __m_maxDepth = p_maxDepth;
}


optix::Context &ContextWrapper::getContext() {
    return __m_context;
}


void ContextWrapper::setDevices_consecutive ( int p_firstDev, int p_numDevs ) {
    std::vector<int> devices = __m_context->getEnabledDevices();
    __m_context->setDevices ( devices.begin() + p_firstDev, devices.begin() + p_firstDev + p_numDevs );
    devices = __m_context->getEnabledDevices();
    std::cout << "Enabled Devices: " << __m_context->getEnabledDeviceCount() << std::endl;

    for ( std::vector<int>::iterator it = devices.begin(); it != devices.end(); ++it ) {
        std::cout << *it << std::endl;
    }
}


void ContextWrapper::setBuffer ( const std::string &p_name, optix::Buffer p_buffer ) {
    __m_context[p_name]->set ( p_buffer );
}



