/*
 * Copyright (c) 2008 - 2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and proprietary
 * rights in and to this software, related documentation and any modifications thereto.
 * Any use, reproduction, disclosure or distribution of this software and related
 * documentation without an express license agreement from NVIDIA Corporation is strictly
 * prohibited.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS*
 * AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS OR IMPLIED,
 * INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES WHATSOEVER (INCLUDING, WITHOUT
 * LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF
 * BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGES
 */

#include <common_structs.h>
#include <ObjLoader.h>
#include <optixu/optixu.h>
#include <optixu/optixu_math_namespace.h>
#include <ImageLoader.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <cassert>
#include <string.h>
#include <map>
#include <stdint.h>


using namespace optix;

//------------------------------------------------------------------------------
//
//  Helper functions
//
//------------------------------------------------------------------------------

namespace {
    std::string getExtension ( const std::string &filename ) {
        // Get the filename extension
        std::string::size_type extension_index = filename.find_last_of ( "." );
        return extension_index != std::string::npos ? filename.substr ( extension_index + 1 ) : std::string();
    }
}

//------------------------------------------------------------------------------
//
//  ObjLoader class definition
//
//------------------------------------------------------------------------------

/*ObjLoader::ObjLoader(const char* filename, Context context, GeometryGroup geometrygroup, Material material, bool force_load_material_params, const char* ASBuilder, const char* ASTraverser, bool large_geom) :
 *	_filename(filename), _context(context), _geometrygroup(geometrygroup), _vbuffer(0), _nbuffer(0), _tbuffer(0), _material(material), _intersect_program(0), _have_default_material(true), _force_load_material_params(force_load_material_params), _ASBuilder(ASBuilder), _ASTraverser(ASTraverser), _large_geom(large_geom), _aabb() {
 *	_pathname = _filename.substr(0, _filename.find_last_of("/\\") + 1);
 * }*/

ObjLoader::ObjLoader ( const char *filename, Context context, GeometryGroup geometrygroup, const char *ASBuilder, const char *ASTraverser, bool large_geom ) :
    _filename ( filename ), _context ( context ), _geometrygroup ( geometrygroup ), _vbuffer ( 0 ), _nbuffer ( 0 ), _tbuffer ( 0 ), _intersect_program ( 0 ), _bbox_program ( 0 ), _have_default_material ( false ), _force_load_material_params ( false ), _ASBuilder ( ASBuilder ), _ASTraverser ( ASTraverser ), _large_geom ( large_geom ), _aabb() {
    _pathname = _filename.substr ( 0, _filename.find_last_of ( "/\\" ) + 1 );
    _material = new optix::Material[ENUM_NUM_MAT_TYPES];
}

void ObjLoader::load() {
    load ( optix::Matrix4x4::identity() );
}

void ObjLoader::setIntersectProgram ( Program intersect_program ) {
    _intersect_program = intersect_program;
}

void ObjLoader::setBboxProgram ( Program bbox_program ) {
    _bbox_program = bbox_program;
}

void ObjLoader::load ( const optix::Matrix4x4 &transform ) {
    // parse the OBJ file
    GLMmodel *model = glmReadOBJ ( _filename.c_str() );

    if ( !model ) {
        std::stringstream ss;
        ss << "ObjLoader::loadImpl - glmReadOBJ( '" << _filename << "' ) failed" << std::endl;
        throw Exception ( ss.str() );
    }

    // Create a single material to be shared by all GeometryInstances
    createMaterial();
    // Create vertex data buffers to be shared by all Geometries
    loadVertexData ( model, transform );
    // Create a GeometryInstance and Geometry for each obj group
    createMaterialParams ( model );
    createGeometryInstances ( model );
    // Create a data for sampling light sources
    createLightBuffer ( model );
    glmDelete ( model );
}

void ObjLoader::createMaterial() {
    if ( _have_default_material )
    { return; }

    std::cout << "- alpha_occlusion_teser" << std::endl;
    optix::Program alpha_occlusion_tester = _context->createProgramFromPTXFile ( generatePtxPath ( "occlusion_shader.cu" ), "alpha_occlusion_tester" );
    std::cout << "- alpha_tester" << std::endl;
    optix::Program alpha_tester = _context->createProgramFromPTXFile ( generatePtxPath ( "alpha_tester.cu" ), "alpha_tester" );
    std::cout << "- shadow_shader" << std::endl;
    optix::Program shadow_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "light_shader.cu" ), "shadow_shader" );
    std::cout << "Creating universal_shader:" << std::endl;
    optix::Material universalMaterial = _context->createMaterial();
    std::cout << "- universal_shader" << std::endl;
    optix::Program universal_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "universal_shader.cu" ), "universal_shader" );
    std::cout << "- setting hit programs" << std::endl;
    universalMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, universal_shader );
    universalMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    universalMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    universalMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating diffuse_shader:" << std::endl;
    optix::Material diffuseMaterial = _context->createMaterial();
    std::cout << "- diffuse_shader" << std::endl;
    optix::Program diffuse_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "diffuse_shader.cu" ), "diffuse_shader" );
    diffuseMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, diffuse_shader );
    diffuseMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    diffuseMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    diffuseMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating mirror_shader:" << std::endl;
    optix::Material mirrorMaterial = _context->createMaterial();
    std::cout << "- mirror_shader" << std::endl;
    optix::Program mirror_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "mirror_shader.cu" ), "mirror_shader" );
    mirrorMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, mirror_shader );
    mirrorMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    mirrorMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    mirrorMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating glossy_shader:" << std::endl;
    optix::Material glossyMaterial = _context->createMaterial();
    std::cout << "- glossy_shader" << std::endl;
    optix::Program glossy_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "glossy_shader.cu" ), "glossy_shader" );
    glossyMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, glossy_shader );
    glossyMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    glossyMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    glossyMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    std::cout << "Creating glass_shader:" << std::endl;
    optix::Material glassMaterial = _context->createMaterial();
    std::cout << "- glass_shader" << std::endl;
    optix::Program glass_shader = _context->createProgramFromPTXFile ( generatePtxPath ( "glass_shader.cu" ), "glass_shader" );
    glassMaterial->setClosestHitProgram ( ENUM_RAYTYPE_PT, glass_shader );
    glassMaterial->setAnyHitProgram ( ENUM_RAYTYPE_PT, alpha_tester );
    glassMaterial->setAnyHitProgram ( ENUM_RAYTYPE_OCCLUSION, alpha_occlusion_tester );
    glassMaterial->setAnyHitProgram ( ENUM_RAYTYPE_SHADOW, shadow_shader );
    _material[ENUM_MAT_TYPE_UNIVERSAL] = universalMaterial;
    _material[ENUM_MAT_TYPE_DIFFUSE] 	= diffuseMaterial;
    _material[ENUM_MAT_TYPE_MIRROR] 	= mirrorMaterial;
    _material[ENUM_MAT_TYPE_GLOSSY] 	= glossyMaterial;
    _material[ENUM_MAT_TYPE_GLASS] 		= glassMaterial;
    /*std::string path = std::string(sutilSamplesPtxDir()) + "/cuda_compile_ptx_generated_obj_material.cu.ptx";
     *
     *	Program closest_hit = _context->createProgramFromPTXFile(path, "closest_hit_radiance");
     *	Program any_hit = _context->createProgramFromPTXFile(path, "any_hit_shadow");
     *	_material = _context->createMaterial();
     *	_material->setClosestHitProgram(0u, closest_hit);
     *	_material->setAnyHitProgram(1u, any_hit);*/
}

void ObjLoader::loadVertexData ( GLMmodel *model, const optix::Matrix4x4 &transform ) {
    unsigned int num_vertices = model->numvertices;
    unsigned int num_texcoords = model->numtexcoords;
    unsigned int num_normals = model->numnormals;
    // Create vertex buffer
    _vbuffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT3, num_vertices );
    float3 *vbuffer_data = static_cast<float3 *> ( _vbuffer->map() );
    std::cout << "number of normals: " << model->numnormals << std::endl;
    // Create normal buffer
    _nbuffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT3, num_normals );
    float3 *nbuffer_data = static_cast<float3 *> ( _nbuffer->map() );
    // Create texcoord buffer
    _tbuffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_FLOAT2, num_texcoords );
    float2 *tbuffer_data = static_cast<float2 *> ( _tbuffer->map() );

    // Transform and copy vertices.
    for ( unsigned int i = 0; i < num_vertices; ++i ) {
        const float3 v3 = * ( ( float3 * ) &model->vertices[ ( i + 1 ) * 3] );
        float4 v4 = make_float4 ( v3, 1.0f );
        vbuffer_data[i] = make_float3 ( transform * v4 );
    }

    // Transform and copy normals.
    const optix::Matrix4x4 norm_transform = transform.inverse().transpose();

    for ( unsigned int i = 0; i < num_normals; ++i ) {
        const float3 v3 = * ( ( float3 * ) &model->normals[ ( i + 1 ) * 3] );
        float4 v4 = make_float4 ( v3, 0.0f );
        nbuffer_data[i] = make_float3 ( norm_transform * v4 );
    }

    // Copy texture coordinates.
    memcpy ( static_cast<void *> ( tbuffer_data ), static_cast<void *> ( & ( model->texcoords[2] ) ), sizeof ( float ) * num_texcoords * 2 );

    // Calculate bbox of model
    for ( unsigned int i = 0; i < num_vertices; ++i )
    { _aabb.include ( vbuffer_data[i] ); }

    // Unmap buffers.
    _vbuffer->unmap();
    _nbuffer->unmap();
    _tbuffer->unmap();
}

void ObjLoader::createGeometryInstances ( GLMmodel *model ) {
    // Load triangle_mesh programs
    if ( !_intersect_program.get() ) {
        std::string path = std::string ( generatePtxPath ( "triangle_mesh.cu" ) );
        _intersect_program = _context->createProgramFromPTXFile ( path, "mesh_intersect" );
    }

    if ( !_bbox_program.get() ) {
        std::string path = std::string ( generatePtxPath ( "triangle_mesh.cu" ) );
        _bbox_program = _context->createProgramFromPTXFile ( path, "mesh_bounds" );
    }

    std::vector<GeometryInstance> instances;
    std::map< uint32_t, std::vector< GLMgroup * > > glmGroupsByMaterial;
    std::map< uint32_t, uint32_t > numTriByMaterial;
    // Loop over all groups -- grab the triangles and material props from each group
    unsigned int triangle_count = 0u;
    unsigned int group_count = 0u;
    std::cout << "Sorting meshes by material..." << std::endl;
    unsigned int index = 0;

    for ( GLMgroup *obj_group = model->groups; obj_group != 0; obj_group = obj_group->next, group_count++ ) {
        uint32_t materialId = obj_group->material;
        unsigned int num_triangles = obj_group->numtriangles;

        if ( numTriByMaterial.find ( materialId ) == numTriByMaterial.end() ) {
            numTriByMaterial[materialId] = num_triangles;

        } else {
            numTriByMaterial[materialId] += num_triangles;
            std::cout << "current number of triangles for material " << materialId << ": " << numTriByMaterial[materialId] << std::endl;
        }

        if ( num_triangles == 0 )
        { continue; }

        glmGroupsByMaterial[obj_group->material].push_back ( obj_group );
    }

    for ( std::map< uint32_t, std::vector< GLMgroup * > >::iterator it = glmGroupsByMaterial.begin(); it != glmGroupsByMaterial.end(); ++it ) {
        uint32_t totalNumTrisForMat = numTriByMaterial[it->first];
        // Create vertex index buffers
        Buffer vindex_buffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_INT3, totalNumTrisForMat );
        int3 *vindex_buffer_data = static_cast<int3 *> ( vindex_buffer->map() );
        Buffer tindex_buffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_INT3, totalNumTrisForMat );
        int3 *tindex_buffer_data = static_cast<int3 *> ( tindex_buffer->map() );
        Buffer nindex_buffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_INT3, totalNumTrisForMat );
        int3 *nindex_buffer_data = static_cast<int3 *> ( nindex_buffer->map() );
        Buffer mbuffer = _context->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_INT, totalNumTrisForMat );
        unsigned int *mbuffer_data = static_cast<unsigned int *> ( mbuffer->map() );
        optix::Geometry mesh;
        unsigned int offset = 0;

        for ( std::vector< GLMgroup * >::iterator it2 = it->second.begin(); it2 != it->second.end(); ++it2 ) {
            for ( unsigned int i = 0; i < ( *it2 )->numtriangles; ++i, ++triangle_count ) {
                unsigned int tindex = ( *it2 )->triangles[i];
                int3 vindices;
                vindices.x = model->triangles[tindex].vindices[0] - 1;
                vindices.y = model->triangles[tindex].vindices[1] - 1;
                vindices.z = model->triangles[tindex].vindices[2] - 1;
                assert ( vindices.x <= static_cast<int> ( model->numvertices ) );
                assert ( vindices.y <= static_cast<int> ( model->numvertices ) );
                assert ( vindices.z <= static_cast<int> ( model->numvertices ) );
                vindex_buffer_data[offset + i] = vindices;
                int3 nindices;
                nindices.x = model->triangles[tindex].nindices[0] - 1;
                nindices.y = model->triangles[tindex].nindices[1] - 1;
                nindices.z = model->triangles[tindex].nindices[2] - 1;
                assert ( nindices.x <= static_cast<int> ( model->numnormals ) );
                assert ( nindices.y <= static_cast<int> ( model->numnormals ) );
                assert ( nindices.z <= static_cast<int> ( model->numnormals ) );
                int3 tindices;
                tindices.x = model->triangles[tindex].tindices[0] - 1;
                tindices.y = model->triangles[tindex].tindices[1] - 1;
                tindices.z = model->triangles[tindex].tindices[2] - 1;
                assert ( tindices.x <= static_cast<int> ( model->numtexcoords ) );
                assert ( tindices.y <= static_cast<int> ( model->numtexcoords ) );
                assert ( tindices.z <= static_cast<int> ( model->numtexcoords ) );
                nindex_buffer_data[offset + i] = nindices;
                tindex_buffer_data[offset + i] = tindices;
                mbuffer_data[offset + i] = 0; // See above TODO
            }

            offset += ( *it2 )->numtriangles;
        }

        std::cout << "Creating unified mesh" << std::endl;
        mesh = _context->createGeometry();
        mesh->setPrimitiveCount ( totalNumTrisForMat );
        mesh->setIntersectionProgram ( _intersect_program );
        mesh->setBoundingBoxProgram ( _bbox_program );
        mesh["vertex_buffer"]->setBuffer ( _vbuffer );
        mesh["vindex_buffer"]->setBuffer ( vindex_buffer );
        mesh["normal_buffer"]->setBuffer ( _nbuffer );
        mesh["texcoord_buffer"]->setBuffer ( _tbuffer );
        mesh["tindex_buffer"]->setBuffer ( tindex_buffer );
        mesh["nindex_buffer"]->setBuffer ( nindex_buffer );
        mesh["material_buffer"]->setBuffer ( mbuffer );
        vindex_buffer->unmap();
        tindex_buffer->unmap();
        nindex_buffer->unmap();
        mbuffer->unmap();
        ++index;
        // check material parameters and set according type
        unsigned int matIndex = ENUM_MAT_TYPE_DIFFUSE;
        unsigned int propertyCount = 0;

        if ( _material_params.size() > 0 ) {
            MatParams &mp = _material_params[ ( it->second ) [0]->material];

            if ( fmaxf ( mp.kd ) > 0.f ) {
                ++propertyCount;
                matIndex = ENUM_MAT_TYPE_DIFFUSE;
                std::cout << "mp.kd: " << mp.kd.x << ", " << mp.kd.y << ", " << mp.kd.z << std::endl;
            }

            if ( mp.phong_exp > 1.f && fmaxf ( mp.ks ) > 0.f ) {
                ++propertyCount;

                if ( mp.phong_exp >= 10000.f ) {
                    matIndex = ENUM_MAT_TYPE_MIRROR;

                } else {
                    matIndex = ENUM_MAT_TYPE_GLOSSY;
                }

                std::cout << "mp.ks: " << mp.ks.x << ", " << mp.ks.y << ", " << mp.ks.z << std::endl;
            }

            if ( mp.alpha < 1.f && mp.ni != 1.f ) {
                ++propertyCount;
                matIndex = ENUM_MAT_TYPE_GLASS;
                std::cout << "mp.alpha: " << mp.alpha << std::endl;
            }

            std::cout << "propertyCount: " << propertyCount << std::endl;

            if ( propertyCount > 1 ) {
                std::cout << "propertyCount > 1!" << std::endl;
                matIndex = ENUM_MAT_TYPE_UNIVERSAL;
            }

            std::cout << "matIndex: " << matIndex << std::endl;
            // Create the geom instance to hold mesh and material params
            GeometryInstance instance = _context->createGeometryInstance ( mesh, & ( _material[matIndex] ), & ( _material[matIndex] ) + 1 );
            loadMaterialParams ( instance, ( it->second ) [0]->material );
            instances.push_back ( instance );

        } else {
            GeometryInstance instance = _context->createGeometryInstance ( mesh, & ( _material[matIndex] ), & ( _material[matIndex] ) + 1 );
            loadMaterialParams ( instance, 0 );
            instances.push_back ( instance );
        }
    }

    std::cout << "meshes: " <<  index << std::endl;
    assert ( triangle_count == model->numtriangles );
    // Set up group
    const unsigned current_child_count = _geometrygroup->getChildCount();
    _geometrygroup->setChildCount ( current_child_count + static_cast<unsigned int> ( instances.size() ) );
    optix::Acceleration acceleration = _context->createAcceleration ( _ASBuilder, _ASTraverser );

    if ( _large_geom ) {
        acceleration->setProperty ( "leaf_size", "1" );

    } else {
        if ( _ASBuilder == std::string ( "Sbvh" ) || _ASBuilder == std::string ( "TriangleKdTree" ) || _ASTraverser == std::string ( "KdTree" ) ) {
            acceleration->setProperty ( "vertex_buffer_name", "vertex_buffer" );
            acceleration->setProperty ( "index_buffer_name", "vindex_buffer" );
        }
    }

    _geometrygroup->setAcceleration ( acceleration );
    acceleration->markDirty();

    for ( unsigned int i = 0; i < instances.size(); ++i )
    { _geometrygroup->setChild ( current_child_count + i, instances[i] ); }

    std::cout << "GeometryGroup children: " << _geometrygroup->getChildCount() << std::endl;

    if ( _large_geom ) {
        rtBufferDestroy ( _vbuffer->get() );
    }
}

bool ObjLoader::isMyFile ( const char *filename ) {
    return getExtension ( filename ) == "obj";
}

/**
 * 46 typedef struct _GLMmaterial
 * 47 {
 * 48   char* name;        	  // name of material
 * 49   float diffuse[4];       // Kd diffuse component
 * 50   float ambient[4];       // Ka ambient component
 * 51   float specular[4];      // Ks specular component
 * 52   float emissive[4];      // emissive component
 * 53   float shininess;        // Ns specular exponent
 * 54   float refraction;       // Tr
 * 55   float alpha;            // d
 * 56   float reflectivity;     // reflection
 * 57   int   shader;           // illum
 * 58
 * 59   // Texture maps, zero length if not specified.
 * 60   char  ambient_map [MaxStringLength]; // map_Ka
 * 61   char  diffuse_map [MaxStringLength]; // map_Kd
 * 62   char  specular_map[MaxStringLength]; // map_Ks
 * 63   char  dissolve_map[MaxStringLength]; // map_D
 * 64   char  bump_map[MaxStringLength];
 * 65
 * 66   // Scaling for texture maps (initialized to 0 for not set)
 * 67   float ambient_map_scaling[2];
 * 68   float diffuse_map_scaling[2];
 * 69   float specular_map_scaling[2];
 * 70   float dissolve_map_scaling[2];
 * 71   float bump_map_scaling[2];
 * 72
 * 73 } GLMmaterial;
 */

void ObjLoader::loadMaterialParams ( GeometryInstance gi, unsigned int index ) {
    // We dont need any material params if we have default material
    if ( _have_default_material && !_force_load_material_params ) {
        return;
    }

    // If no materials were given in model use reasonable defaults
    if ( _material_params.empty() ) {
        //    gi[ "emissive" ]->setFloat( 0.0f, 0.0f, 0.0f );
        gi["phong_exp"]->setFloat ( 32.0f );
        //    gi[ "reflectivity" ]->setFloat( 0.0f, 0.0f, 0.0f );
        //    gi[ "illum" ]->setInt( 2 );
        gi["kd"]->setFloat ( 1.0f, 1.0f, 1.0f );
        gi["ks"]->setFloat ( 0.0f, 0.0f, 0.0f );
        gi["tf"]->setFloat ( 1.0f, 1.0f, 1.0f );
        gi["alpha"]->setFloat ( 1.f );
        gi["ni"]->setFloat ( 1.f );
        //    gi["ambient_map"]->setTextureSampler( loadTexture( _context, "", make_float3( 0.2f, 0.2f, 0.2f ) ) );
        gi["diffuse_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 1.f, 1.f, 1.f ) ) );
        gi["specular_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 1.0f, 1.0f, 1.0f ) ) );
        gi["dissolve_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 1.0f, 1.0f, 1.0f ) ) );
        gi["bump_map"]->setTextureSampler ( loadTexture ( _context, "", make_float3 ( 0.f, 0.f, 0.f ) ) );
        return;
    }

    // Load params from this material into the GI
    if ( index < _material_params.size() ) {
        MatParams &mp = _material_params[index];
        //    gi[ "emissive"     ]->setFloat( mp.emissive );
        //    gi[ "reflectivity" ]->setFloat( mp.reflectivity );
        gi["phong_exp"]->setFloat ( mp.phong_exp );
        //    gi[ "illum"        ]->setInt( mp.illum );
        gi["alpha"]->setFloat ( mp.alpha );
        gi["kd"]->setFloat ( mp.kd );
        gi["ks"]->setFloat ( mp.ks );
        gi["tf"]->setFloat ( mp.tf );
        gi["ni"]->setFloat ( mp.ni );
        //    gi[ "ambient_map"  ]->setTextureSampler( mp.ambient_map );
        gi["diffuse_map"]->setTextureSampler ( mp.diffuse_map );
        gi["specular_map"]->setTextureSampler ( mp.specular_map );
        gi["dissolve_map"]->setTextureSampler ( mp.dissolve_map );
        gi["bump_map"]->setTextureSampler ( mp.bump_map );
        /*std::cout << "Material number: " << index << std::endl;
         *		std::cout << "phong exp: " << mp.phong_exp << std::endl;
         *		std::cout << "kd: " << mp.kd.x << ", " << mp.kd.y << ", " << mp.kd.z << std::endl;
         *		std::cout << "ks: " << mp.ks.x << ", " << mp.ks.y << ", " << mp.ks.z << std::endl;
         *		std::cout << "tf: " << mp.tf.x << ", " << mp.tf.y << ", " << mp.tf.z << std::endl;
         *		std::cout << "alpha: " << mp.alpha << std::endl;
         *		std::cout << "ni: " << mp.ni << std::endl;*/
        return;
    }

    // Should never reach this point
    std::cerr << "WARNING -- ObjLoader::loadMaterialParams given index out of range: " << index << std::endl;
}

void ObjLoader::createMaterialParams ( GLMmodel *model ) {
    std::cout << "numMaterials: " << model->nummaterials << std::endl;
    _material_params.resize ( model->nummaterials );

    for ( unsigned int i = 0; i < model->nummaterials; ++i ) {
        GLMmaterial &mat = model->materials[i];
        MatParams &params = _material_params[i];
        //    params.emissive     = make_float3( mat.emissive[0], mat.emissive[1], mat.emissive[2] );
        //    params.reflectivity = make_float3( mat.specular[0], mat.specular[1], mat.specular[2] );
        params.phong_exp = mat.shininess;
        params.kd = fminf ( make_float3 ( 1.f ), make_float3 ( mat.diffuse[0], mat.diffuse[1], mat.diffuse[2] ) ); // + make_float3(mat.ambient[0], mat.ambient[1], mat.ambient[2]));
        params.ks = fminf ( make_float3 ( 1.f ), make_float3 ( mat.specular[0], mat.specular[1], mat.specular[2] ) );
        //    optix::float3 weightTotal = params.kd + params.ks;
        //    params.kd = params.kd / weightTotal;
        //    params.ks = params.ks / weightTotal;
        //    params.illum        = ( (mat.shader > 3) ? 2 : mat.shader ); // use 2 as default if out-of-range
        params.alpha = fminf ( 1.f, mat.alpha * mat.diffuse[3] );
        params.ni = mat.refraction;
        params.tf = fminf ( make_float3 ( 1.f ), make_float3 ( mat.transmission[0], mat.transmission[1], mat.transmission[2] ) );
        //	std::cout << "alpha channel: " << mat.diffuse[3] << std::endl;
        // load textures relatively to OBJ main file
        //    std::string ambient_map  = strlen(mat.ambient_map)  ? _pathname + mat.ambient_map  : "";
        std::string diffuse_map = strlen ( mat.diffuse_map ) ? _pathname + mat.diffuse_map : "";
        std::string specular_map = strlen ( mat.specular_map ) ? _pathname + mat.specular_map : "";
        std::string dissolve_map = strlen ( mat.dissolve_map ) ? _pathname + mat.dissolve_map : "";
        std::string bump_map = strlen ( mat.bump_map ) ? _pathname + mat.bump_map : "";
        //    params.ambient_map = loadTexture( _context, ambient_map, Ka );
        params.diffuse_map = loadTexture ( _context, diffuse_map, make_float3 ( 1.f ) );
        params.specular_map = loadTexture ( _context, specular_map, make_float3 ( 1.f ) );
        params.dissolve_map = loadTexture ( _context, dissolve_map, make_float3 ( 1.f ) );
        params.bump_map = loadTexture ( _context, bump_map, make_float3 ( 0.f ) );
    }
}

void ObjLoader::createLightBuffer ( GLMmodel *model ) {
    // create a buffer for the next-event estimation
    _light_buffer = _context->createBuffer ( RT_BUFFER_INPUT );
    _light_buffer->setFormat ( RT_FORMAT_USER );
    _light_buffer->setElementSize ( sizeof ( TriangleLight ) );
    // light sources
    std::vector<TriangleLight> lights;
    unsigned int num_light = 0u;
    unsigned int group_count = 0u;

    if ( model->nummaterials > 0 ) {
        for ( GLMgroup *obj_group = model->groups; obj_group != 0; obj_group = obj_group->next, group_count++ ) {
            unsigned int num_triangles = obj_group->numtriangles;

            if ( num_triangles == 0 )
            { continue; }

            GLMmaterial &mat = model->materials[obj_group->material];

            if ( ( mat.emissive[0] + mat.emissive[1] + mat.emissive[2] ) > 0.0f ) {
                // extract necessary data
                for ( unsigned int i = 0; i < obj_group->numtriangles; ++i ) {
                    // indices for vertex data
                    unsigned int tindex = obj_group->triangles[i];
                    int3 vindices;
                    vindices.x = model->triangles[tindex].vindices[0];
                    vindices.y = model->triangles[tindex].vindices[1];
                    vindices.z = model->triangles[tindex].vindices[2];
                    TriangleLight light;
                    light.v1 = * ( ( float3 * ) &model->vertices[vindices.x * 3] );
                    light.v2 = * ( ( float3 * ) &model->vertices[vindices.y * 3] );
                    light.v3 = * ( ( float3 * ) &model->vertices[vindices.z * 3] );
                    // normal vector
                    light.normal = normalize ( cross ( light.v2 - light.v3, light.v1 - light.v3 ) );
                    light.emission = make_float3 ( mat.emissive[0], mat.emissive[1], mat.emissive[2] );
                    lights.push_back ( light );
                    num_light++;
                }
            }
        }
    }

    // write to the buffer
    _light_buffer->setSize ( 0 );

    if ( num_light != 0 ) {
        _light_buffer->setSize ( num_light );
        memcpy ( _light_buffer->map(), &lights[0], num_light * sizeof ( TriangleLight ) );
        _light_buffer->unmap();
    }
}
