
#include <RenderProcess.h>
#include <sobol.h>
#include <sutil.h>

#include <assert.h>


void RenderProcess::preComputeSobolVectors() {
    //compute sobol stuff for maximum dimension with ray depth 32
    int dimensions = ENUM_DIMENSION_BASE_NUM + __m_renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_MAX_SOBOL_DEPTH ) * ENUM_DIMENSION_BOUNCE_NUM;
    dimensions = min ( dimensions, SOBOL_MAX_DIMENSIONS );
    optix::Buffer sobol_directions = __m_contextWrapper.getContext()->createBuffer ( RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_INT, SOBOL_BITS * dimensions );
    uint *directions = ( uint * ) sobol_directions->map();
    sobol_generate_direction_vectors ( ( uint ( * ) [SOBOL_BITS] ) directions, dimensions );
    sobol_directions->unmap();
    __m_contextWrapper.setBuffer ( "sobol_directions", sobol_directions );
}




void RenderProcess::createOptixBuffer ( enum_BufferTypes p_bufferType ) {
    BufferInfo bi = ( *__m_bufInfo ) [p_bufferType];

    switch ( bi.m_dimension ) {
        case ENUM_BUFFERDIMENSION_1D : {
                __m_outputBuffers->m_optixBuffers[p_bufferType] = __m_contextWrapper.getContext()->createBuffer (
                            RT_BUFFER_OUTPUT,
                            bi.m_format,
                            bi.m_width
                        );
				if (bi.m_elementSize) {
					__m_outputBuffers->m_optixBuffers[p_bufferType]->setElementSize(bi.m_elementSize);					
				}
                break;
            }

        case ENUM_BUFFERDIMENSION_2D : {
                __m_outputBuffers->m_optixBuffers[p_bufferType] = __m_contextWrapper.getContext()->createBuffer (
                            RT_BUFFER_INPUT_OUTPUT,
                            bi.m_format,
                            bi.m_width,
                            bi.m_height
                        );
				if (bi.m_elementSize) {
					__m_outputBuffers->m_optixBuffers[p_bufferType]->setElementSize(bi.m_elementSize);					
				}
                break;
            }

        default :
            break;
    }
}


void RenderProcess::setDevicePointerForOptixBuffer ( enum_BufferTypes p_bufferType ) {
    __m_outputBuffers->m_optixBuffers[p_bufferType]->getDevicePointer ( 0, (void**) & ( __m_outputBuffers->m_devicePointers[p_bufferType] ) );
}


void RenderProcess::setOptixBufferInContext ( enum_BufferTypes p_bufferType ) {
    std::cout << ( std::string ( "output_buffer_" ) + ( *__m_bufInfo ) [p_bufferType].m_name ) << std::endl;
    std::string bufferName = std::string ( "output_buffer_" ) + ( *__m_bufInfo ) [p_bufferType].m_name;
    __m_contextWrapper.getContext() [bufferName]->set ( __m_outputBuffers->m_optixBuffers[p_bufferType] );
}





void RenderProcess::createBuffers() {
    for ( int i = 0; i < ENUM_NUM_BUFFERTYPES; ++i ) {
        std::cout << "Creating Buffer " << i << std::endl;
        enum_BufferTypes bufferType = ( enum_BufferTypes ) i;
        std::cout << "Optix Buffer..." << std::endl;
        createOptixBuffer ( bufferType );
        std::cout << "Device Pointer..." << std::endl;
        setDevicePointerForOptixBuffer ( bufferType );
        std::cout << "Set Buffer in Context..." << std::endl;
        setOptixBufferInContext ( bufferType );
    }
}




RenderProcess::RenderProcess() {
    __m_isInitialized = false;
    __m_renderParameters = 0;
    __m_outputBuffers = 0;
}


RenderProcess::RenderProcess ( RenderParameters *p_renderParameters, OutputBuffers *p_outputBuffers, std::vector<BufferInfo> *p_bufInfo ) {
    initialize ( p_renderParameters, p_outputBuffers, p_bufInfo );
}


RenderProcess::~RenderProcess() {
}


void RenderProcess::initialize ( RenderParameters *p_renderParameters, OutputBuffers *p_outputBuffers, std::vector<BufferInfo> *p_bufInfo ) {
    assert ( !__m_isInitialized );
    __m_renderParameters = p_renderParameters;
    __m_bufInfo = p_bufInfo;
    __m_outputBuffers = p_outputBuffers;
    __m_contextWrapper.initialize (
        p_renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_RES_X ),
        p_renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_RES_Y )
    );
    __m_contextWrapper.initializeContext();
    __m_contextWrapper.setDevices_consecutive (
        p_renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_FIRST_GPU ),
        p_renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_NUM_GPUS )
    );
    __m_contextWrapper.createOptixPrograms();
    __m_renderParameters->setParameterInt ( ENUM_RENDERPARAMETER_MAX_SOBOL_DEPTH, 32 );
    preComputeSobolVectors();
    createBuffers();
    updateRenderParameters();
    __m_isInitialized = true;
}


void RenderProcess::updateRenderParameters() {
    optix::Context &context = __m_contextWrapper.getContext();

    for ( int i = 0; i < ENUM_NUM_RENDERPARAMETERS_INT; ++i ) {
        enum_parameterNames_int parameter = ( enum_parameterNames_int ) i;
        std::string name = __m_renderParameters->getParameterNameOptix ( parameter );

        if ( name == "" ) { continue; };

        context[name]->setUint ( __m_renderParameters->getParameterInt ( parameter ) );
    }

    for ( int i = 0; i < ENUM_NUM_RENDERPARAMETERS_FLOAT; ++i ) {
        enum_parameterNames_float parameter = ( enum_parameterNames_float ) i;
        std::string name = __m_renderParameters->getParameterNameOptix ( parameter );

        if ( name == "" ) { continue; };

        context[name]->setFloat ( __m_renderParameters->getParameterFloat ( parameter ) );
    }

    //"directionalLight_direction""directionalLight_fluxRGB
    __m_scene.setSun (
        context,
        __m_renderParameters->getParameterNameOptix ( ENUM_RENDERPARAMETER_SUN_BRIGHTNESS ),
        __m_renderParameters->getParameterNameOptix ( ENUM_RENDERPARAMETER_SUN_SCALE ),
        __m_renderParameters->getParameterFloat ( ENUM_RENDERPARAMETER_SUN_AXIS_PHI ),
        __m_renderParameters->getParameterFloat ( ENUM_RENDERPARAMETER_SUN_AXIS_THETA ),
        __m_renderParameters->getParameterFloat ( ENUM_RENDERPARAMETER_SUN_BRIGHTNESS ),
        __m_renderParameters->getParameterFloat ( ENUM_RENDERPARAMETER_SUN_SCALE )
    );
}


void RenderProcess::loadScene ( TrialConfig p_tc, const std::string &p_envPath, const std::string &p_upVector ) {
    __m_scene.initialize ( __m_contextWrapper.getContext() );
    __m_scene.loadModel ( p_tc.sceneFile, p_upVector );
    __m_scene.createLights(p_tc);
    __m_scene.setGroups();
    __m_scene.setEnvironmentMap ( p_envPath, __m_renderParameters->getParameterFloat ( ENUM_RENDERPARAMETER_ENV_GAMMA ) );
}


void RenderProcess::setCamera ( Camera &p_camera ) {
    optix::Context &context = __m_contextWrapper.getContext();
    p_camera.setHfov ( __m_renderParameters->getParameterFloat ( ENUM_RENDERPARAMETER_HFOV ) );
    optix::float3 eye = p_camera.getEye();
    optix::float3 look_at = p_camera.getLookAt();
    optix::float3 up = p_camera.getUp();
    float hfov = p_camera.getHfov();
    float aspect_ratio = p_camera.getAspectRatio();
    optix::float3 camera_u, camera_v, camera_w;
    sutil::calculateCameraVariables ( eye, look_at, up, hfov, aspect_ratio, camera_u, camera_v, camera_w );
    context["eye"]->setFloat ( eye );
    context["U"]->setFloat ( camera_u );
    context["V"]->setFloat ( camera_v );
    context["W"]->setFloat ( camera_w );
}


ContextWrapper &RenderProcess::getContextWrapper() {
    return __m_contextWrapper;
}


Scene &RenderProcess::getScene() {
    return __m_scene;
}
