/*
 * NormalShader.cu
 *
 *  Created on: May 11, 2011
 *      Author: troth22s
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>

#include <common_structs.h>
#include <random.h>
#include <tr1/float.h>

//#include <lightsources/CubeLight.cu>
#include <SphereLight.cu>
#include <LightSampler.cu>

#include <sobol_cuda.cu>
//#include <intersectionVars.cu>

#include <constants.h>

#include <ConfigurationMacros.h>


#include <optix_declaration_macros.h>

#ifndef __OPTIX_VARS_SHADING_INFO__
#define __OPTIX_VARS_SHADING_INFO__
__DECLARE_OPTIX_VARS_SHADING_INFO__
#endif

#ifndef __OPTIX_VARS_TEXTURES_SPECULAR__
#define __OPTIX_VARS_TEXTURES_SPECULAR__
__DECLARE_OPTIX_VARS_TEXTURES_SPECULAR__
#endif

#ifndef __OPTIX_VARS_TEXTURES_DIFFUSE__
#define __OPTIX_VARS_TEXTURES_DIFFUSE__
__DECLARE_OPTIX_VARS_TEXTURES_DIFFUSE__
#endif

#ifndef __OPTIX_VARS_SAMPLER_SETTINGS__
#define __OPTIX_VARS_SAMPLER_SETTINGS__
__DECLARE_OPTIX_VARS_SAMPLER_SETTINGS__
#endif



class TorranceSparrowBRDF {
  private:
    float __m_exponent;
    float __m_eta;
    float __m_k;


    /* microfacet distribution */
    __device__ float d ( const float p_cosThetaH, float p_exponent ) {
        return ( p_exponent + 2.f ) * INV_TWOPI * powf ( p_cosThetaH, p_exponent );
    }


    /* pdf for microfacet distribution */
    __device__ float pdf ( float p_cosThetaNH, float p_cosThetaOH, float p_exponent ) {
        return ( ( p_exponent + 1.f ) * powf ( p_cosThetaNH, p_exponent ) ) * INV_EIGHTPI / p_cosThetaOH;
    }


    /* sample microfacet distribution */
    __device__ void sample_d ( const float3 &p_wo, float3 *p_wi, const float3 &p_normal, float p_u1, float p_u2, float *p_pdf, float p_exponent ) {
        float costheta = powf ( p_u1, 1.f / ( p_exponent + 1.f ) );
        float sintheta = sqrtf ( fmaxf ( 0.f, 1.f - costheta * costheta ) );
        float phi = p_u2 * TWOPI;
        optix::Onb onb ( p_normal );
        optix::float3 wh = optix::make_float3 ( sintheta * cosf ( phi ), sintheta * sinf ( phi ), costheta );
        //  optix::cosine_sample_hemisphere(u1, u2, wh);
        onb.inverse_transform ( wh );
        wh = optix::normalize ( wh );
        *p_wi = optix::normalize ( optix::reflect ( ray.direction, wh ) );
        *p_pdf = pdf ( optix::dot ( p_normal, wh ), optix::dot ( p_wo, wh ), p_exponent );
    }

  public:


    /* get pdf */
    __device__ float pdf ( const optix::float3 &p_wo, const optix::float3 &p_wi, const optix::float3 &p_normal ) {
        optix::float3 wh = optix::normalize ( p_wo + p_wi );
        float cosThetaNH = optix::dot ( p_normal, wh );
        float cosThetaOH = optix::dot ( p_wo, wh );
        return pdf ( cosThetaNH, cosThetaOH, __m_exponent );
    }


    __device__ TorranceSparrowBRDF ( float p_exponent, float p_eta, float p_k = 1.f ) {
        __m_exponent = fminf ( p_exponent, 1000000000.f );
        __m_eta = p_eta;
        __m_k = p_k;
    }


    __device__ float geometricTerm ( const float p_ndotwh, const float p_ndotwo, const float p_ndotwi, const float p_wodotwh ) {
        return fminf ( 1.f, fminf ( ( 2.f * p_ndotwh * p_ndotwo / p_wodotwh ), ( 2.f * p_ndotwh * p_ndotwi / p_wodotwh ) ) );
    }


    /* sample brdf */
    __device__ float sample_f ( const float3 &p_wo, float3 *p_wi, const float3 &p_normal, float p_u1, float p_u2, float *p_pdf ) {
        sample_d ( p_wo, p_wi, p_normal, p_u1, p_u2, p_pdf, __m_exponent );
        return f ( p_wo, *p_wi, p_normal, p_pdf );
    }


    /* get brdf value and pdf for incoming and outgoing directions */
    __device__ float f ( const float3 &p_wo, const float3 &p_wi, const float3 &p_normal, float *p_pdf ) {
        float cosThetaO = optix::dot ( p_wo, p_normal );
        float cosThetaI = optix::dot ( p_wi, p_normal );
        float3 wh = optix::normalize ( p_wi + p_wo );
        float cosThetaNH = optix::dot ( p_normal, wh );
        float cosThetaIH = optix::dot ( p_wi, wh );
        float cosThetaOH = optix::dot ( p_wo, wh );

        if ( cosThetaO > 0.f && cosThetaI > 0.f && cosThetaNH > 0.f && cosThetaIH > 0.f && cosThetaOH > 0.f ) {
            // should be wavelength dependent
            float fresnelTerm = fresnel ( __m_eta, __m_k, cosThetaO );
            float factor = ( d ( cosThetaNH, __m_exponent ) * geometricTerm ( cosThetaNH, cosThetaO, cosThetaI, cosThetaOH ) * fresnelTerm ) / ( 4.f * cosThetaI * cosThetaO );
            *p_pdf = pdf ( p_wo, p_wi, p_normal );
            return factor;

        } else {
            *p_pdf = 0.f;
            return 0.f;
        }
    }

};




class DiffuseBRDF {

  public:

    __device__ float pdf ( const optix::float3 &p_wo, const optix::float3 &p_wi, const optix::float3 &p_normal ) {
        return optix::dot ( p_wi, p_normal ) * INV_PI;
    }

    __device__ float sample_f ( const float3 &p_wo, float3 *p_wi, float3 &p_normal, float p_u1, float p_u2, float *p_pdf ) {
        // orthonormal basis for choosing the new sampling direction
        optix::Onb onb ( p_normal );
        optix::cosine_sample_hemisphere ( p_u1, p_u2, *p_wi );
        onb.inverse_transform ( *p_wi );
        *p_wi = optix::normalize ( *p_wi );
        return f ( p_wo, *p_wi, p_normal, p_pdf );
    }

    __device__ float f ( const float3 &p_wo, const float3 &p_wi, const float3 &p_normal, float *p_pdf ) {
        *p_pdf = pdf ( p_wo, p_wi, p_normal );
        return INV_PI;
    }

};




/* @brief closest hit program for path tracing of glossy materials (torrance/sparrow) */
RT_PROGRAM void glossy_shader() {
    // if (prd_pt.depth >= 1) {

    //     float3 hit_point = ray.origin + t_hit * ray.direction;
    //     float3 diffTex = make_float3 ( tex2D ( diffuse_map, texcoord.x, texcoord.y ) ) * ks;
    //     float3 geometricNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, geometric_normal ) );
    //     float3 geometricNormal_world_ff = optix::faceforward ( geometricNormal_world, -ray.direction, geometricNormal_world );
    //     float3 shadingNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, shading_normal ) );
    //     /* front facing normals are necessary if we look at the primitive from behind */
    //     float3 shadingNormal_world_ff = optix::faceforward ( shadingNormal_world, -ray.direction, geometricNormal_world );

    //     if (prd_pt.cachingSearch) prd_pt.done = true;

    //     if (prd_pt.done) {
    //         prd_pt.normal_geometric = geometricNormal_world_ff;
    //         prd_pt.normal = shadingNormal_world_ff;
    //         prd_pt.origin = hit_point;
    //         prd_pt.texCol = diffTex;
    //         prd_pt.hitDist = t_hit;
    //         prd_pt.kd = 1.f;
    //         return;
    //     }

        
       
    //     float u1 = 0.f;
    //     float u2 = 0.f;
        

        
    //     prd_pt.shininess = 0.f;
        
    //     if ( useLdSampling_integrator ) {
    //         unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_U;
    //         unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_V;
    //         u1 = Sobol ( dimension_u, prd_pt.frame, prd_pt.seed );
    //         u2 = Sobol ( dimension_v, prd_pt.frame, prd_pt.seed );

    //     } else {
    //         u1 = rnd ( prd_pt.seed );
    //         u2 = rnd ( prd_pt.seed );
    //     }

    //     DiffuseBRDF dbrdf;
    //     prd_pt.texCol = diffTex;
    //     prd_pt.f = make_float3 ( dbrdf.sample_f ( -ray.direction, & ( prd_pt.direction ), shadingNormal_world_ff, u1, u2, & ( prd_pt.pdf ) ) );
    //     prd_pt.dotProduct = fmaxf ( 0.f, optix::dot ( prd_pt.direction, shadingNormal_world_ff ) );

    //     if ( optix::dot ( prd_pt.direction, geometricNormal_world_ff ) < 0.f ) {
    //         prd_pt.done = true;
    //     }

    //     if (!prd_pt.cachingSearch) {
    //         if ( prd_pt.depth < 2 ) {
    //             prd_pt.radiance = sampleAllLightSources<DiffuseBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, lightProb_diffuse, dbrdf );

    //         } else {
    //             prd_pt.radiance = sampleOneLightSource<DiffuseBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, lightProb_diffuse, dbrdf );
    //         }
    //     }


    //   /*  if (faceIdx % 2 == 0) {
    //         prd_pt.radiance = make_float3(0.f);
    //     }*/

    //     // prd_pt.radiance = make_float3(1.f, 0.f, 0.f);

    //     prd_pt.directIllumination = false;
    //     prd_pt.normal = shadingNormal_world_ff;
    //     prd_pt.normal_geometric = geometricNormal_world_ff;
    //     prd_pt.origin = hit_point;
    //     prd_pt.hitDist = t_hit;
    //     prd_pt.kd = 1.f;


    // } else {

        // rtPrintf("GLOSSY SHADER!\n");

        // if (prd_pt.kd == 1.f) {
        //     prd_pt.done = true;
        // }

      if (prd_pt.kd == 1.f) {
            prd_pt.done = true;
        }

        float3 hit_point = ray.origin + t_hit * ray.direction;
        float3 specTex = make_float3 ( tex2D ( specular_map, texcoord.x, texcoord.y ) ) * ks;
        float3 diffTex = make_float3 ( tex2D ( diffuse_map, texcoord.x, texcoord.y ) ) * kd;
        float3 geometricNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, geometric_normal ) );
        float3 geometricNormal_world_ff = optix::faceforward ( geometricNormal_world, -ray.direction, geometricNormal_world );
        float3 shadingNormal_world = optix::normalize ( rtTransformNormal ( RT_OBJECT_TO_WORLD, shading_normal ) );
        /* front facing normals are necessary if we look at the primitive from behind */
        float3 shadingNormal_world_ff = optix::faceforward ( shadingNormal_world, -ray.direction, geometricNormal_world );
      
        prd_pt.texCol = specTex;
        // prd_pt.throughput = specTex;
        prd_pt.shininess = fminf(1.f, phong_exp / 10000.f);

        if (prd_pt.done) {
            prd_pt.origin = hit_point;
            prd_pt.hitDist = t_hit;
            prd_pt.normal = shadingNormal_world_ff;
            prd_pt.normal_geometric = geometricNormal_world_ff;
            prd_pt.pdf = 1.f;
            prd_pt.f = make_float3(1.f);
            return;
        }



        // 
           
        float u1 = 0.f;
        float u2 = 0.f;

        
        // prd_pt.kd = 0.f;
        

        if ( useLdSampling_integrator ) {
            unsigned int dimension_u = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_U;
            unsigned int dimension_v = ENUM_DIMENSION_BASE_NUM + prd_pt.depth * ENUM_DIMENSION_BOUNCE_NUM + ENUM_DIMENSION_TERMINATE + ENUM_DIMENSION_BSDF_V;
            u1 = Sobol ( dimension_u, prd_pt.frame, prd_pt.seed );
            u2 = Sobol ( dimension_v, prd_pt.frame, prd_pt.seed );

        } else {
            u1 = rnd ( prd_pt.seed );
            u2 = rnd ( prd_pt.seed );
        }

        TorranceSparrowBRDF sbrdf ( phong_exp, ni );
        // prd_pt.texCol = specTex;
        /* sample from specular distribution */
        prd_pt.f = make_float3 ( sbrdf.sample_f ( -ray.direction, & ( prd_pt.direction ), shadingNormal_world_ff, u1, u2, & ( prd_pt.pdf ) ) );

        prd_pt.dotProduct = fmaxf ( 0.f, optix::dot ( prd_pt.direction, shadingNormal_world_ff ) );

        /* if our ray points below the surface, stop everything
            * this can happen with normal maps
            */
        if ( optix::dot ( prd_pt.direction, geometricNormal_world_ff ) < 0.f ) {
            prd_pt.done = true;
        }

            /* for ray depth < 2, sample all light sources, afterwards choose only one randomly
                */
        // if (!prd_pt.cachingSearch) {    
        // if (RENDER_CAUSTICS || prd_pt.depth < 1) {
        //     if ( prd_pt.depth < 1 ) {
        //         prd_pt.radiance = sampleAllLightSources<TorranceSparrowBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, lightProb_specular, sbrdf );

        //     } else {
        //         prd_pt.radiance = sampleOneLightSource<TorranceSparrowBRDF> ( hit_point, shadingNormal_world_ff, geometricNormal_world_ff, lightProb_specular, sbrdf );
        //     }
        // } 
        // }
       




        prd_pt.directIllumination = true;
        prd_pt.normal = shadingNormal_world_ff;
        prd_pt.normal_geometric = geometricNormal_world_ff;
        prd_pt.origin = hit_point;
        prd_pt.hitDist = t_hit;
        prd_pt.kd = 0.1f;

        
    // }
}

