/*
 * triangle_mesh.cu
 *
 *  Created on: Jul 5, 2011
 *      Author: troth22s
 *
 * This has been taken from the OptiX SDK for easy tri mesh representation!
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <optixu/optixu_aabb_namespace.h>

using namespace optix;

// This is to be plugged into an RTgeometry object to represent
// a triangle mesh with a vertex buffer of triangle soup (triangle list)
// with an interleaved position, normal, texturecoordinate layout.

rtBuffer<float3> vertex_buffer;
rtBuffer<float3> normal_buffer;
rtBuffer<float2> texcoord_buffer;
rtBuffer<int3> vindex_buffer; // position indices
rtBuffer<int3> nindex_buffer; // normal indices
rtBuffer<int3> tindex_buffer; // texcoord indices
rtBuffer<int> findex_buffer; // global face indices

rtBuffer<uint> material_buffer; // per-face material index
rtDeclareVariable ( float3, texcoord, attribute texcoord, );
rtDeclareVariable ( float3, geometric_normal, attribute geometric_normal, );
rtDeclareVariable ( float3, shading_normal, attribute shading_normal, );
rtDeclareVariable ( optix::Ray, ray, rtCurrentRay, );
rtDeclareVariable ( int, faceIdx, attribute faceIdx, );

rtTextureSampler<float4, 2> bump_map;

RT_PROGRAM void mesh_intersect ( int primIdx ) {

    int3 v_idx = vindex_buffer[primIdx];
    float3 p0 = vertex_buffer[v_idx.x];
    float3 p1 = vertex_buffer[v_idx.y];
    float3 p2 = vertex_buffer[v_idx.z];
    int3 n_idx = v_idx;
    int3 t_idx = v_idx;
    // Intersect ray with triangle
    float3 n;
    float t, beta, gamma;

    if ( intersect_triangle ( ray, p0, p1, p2, n, t, beta, gamma ) ) {
        if ( rtPotentialIntersection ( t ) ) {
            faceIdx = findex_buffer[primIdx];
            if ( normal_buffer.size() == 0 || n_idx.x < 0 || n_idx.y < 0 || n_idx.z < 0 ) {
                shading_normal = normalize ( n );

            } else {
                float3 n0 = normal_buffer[n_idx.x];
                float3 n1 = normal_buffer[n_idx.y];
                float3 n2 = normal_buffer[n_idx.z];
                shading_normal = normalize ( n1 * beta + n2 * gamma + n0 * ( 1.0f - beta - gamma ) );
            }

            geometric_normal = normalize ( n );
//          geometric_normal = shading_normal;

            if ( texcoord_buffer.size() == 0 || t_idx.x < 0 || t_idx.y < 0 || t_idx.z < 0 ) {
                texcoord = make_float3 ( 0.0f, 0.0f, 0.0f );

            } else {
                float2 t0 = texcoord_buffer[t_idx.x];
                float2 t1 = texcoord_buffer[t_idx.y];
                float2 t2 = texcoord_buffer[t_idx.z];
                texcoord = make_float3 ( t1 * beta + t2 * gamma + t0 * ( 1.0f - beta - gamma ) );
                float3 bumpNormal = make_float3 ( tex2D ( bump_map, texcoord.x, texcoord.y ) );

//
                if ( fmaxf ( bumpNormal ) > 0.f ) {
//
                    float tMatrix[4] = { t1.x - t0.x, t2.x - t0.x, t1.y - t0.y, t2.y - t0.y };
                    float invDet = 1.f / ( tMatrix[0] * tMatrix[3] - tMatrix[1] * tMatrix[2] );
                    tMatrix[0] = tMatrix[3] * invDet;
                    tMatrix[1] = -tMatrix[1] * invDet;
                    tMatrix[2] = -tMatrix[2] * invDet;
                    tMatrix[3] = tMatrix[0] * invDet;
                    float vMatrix[6] = { p1.x - p0.x, p2.x - p0.x, p1.y - p0.y, p2.y - p0.y, p1.z - p0.z, p2.z - p0.z };
                    float result[6] = { vMatrix[0] *tMatrix[0] + vMatrix[1] *tMatrix[2], vMatrix[0] *tMatrix[1] + vMatrix[1] *tMatrix[3], vMatrix[2] *tMatrix[0] + vMatrix[3] *tMatrix[2], vMatrix[2] *tMatrix[1] + vMatrix[3] *tMatrix[3], vMatrix[4] *tMatrix[0] + vMatrix[5] *tMatrix[2], vMatrix[4] *tMatrix[1] + vMatrix[5] *tMatrix[3] };
                    float3 texU = optix::normalize ( make_float3 ( result[0], result[2], result[4] ) );
                    float3 texV = -optix::normalize ( make_float3 ( result[1], result[3], result[5] ) );
                    float3 texV_tmp = texV;

                    if ( fabsf ( optix::dot ( texU, texV ) ) > 0.f ) {
                        texV_tmp = optix::cross ( texU, geometric_normal );

                        if ( optix::dot ( texV_tmp, texV ) < 0.f ) {
                            texV_tmp = -texV_tmp;
                        }
                    }

                    texV = texV_tmp;
                    bumpNormal = optix::normalize ( bumpNormal * 2.f - 1.f );
                    shading_normal = optix::normalize ( bumpNormal.x * texU + bumpNormal.y * texV + bumpNormal.z * shading_normal );
                    shading_normal = rtTransformNormal ( RT_WORLD_TO_OBJECT, shading_normal ); // only necessary for convention purposes ;)
                }

                // Switch from WRAP_REPEAT to WRAP_MIRROR, which is not explicitly supported by OptiX
                //              unsigned int floor_x = floor(texcoord.x);
                //              if (floor_x % 2) { // mirror!
                //                  texcoord.x = 1.f - (texcoord.x - floor_x);
                //                  texU = -texU;
                //              }
                //              unsigned int floor_y = floor(texcoord.y);
                //              if (floor_y % 2) { // mirror!
                //                  texcoord.y = 1.f - (texcoord.y - floor_y);
                //                  texV = -texV;
                //              }
            }

            rtReportIntersection ( material_buffer[primIdx] );
        }
    }
}

RT_PROGRAM void mesh_bounds ( int primIdx, float result[6] ) {
    const int3 v_idx = vindex_buffer[primIdx];
    const float3 v0 = vertex_buffer[v_idx.x];
    const float3 v1 = vertex_buffer[v_idx.y];
    const float3 v2 = vertex_buffer[v_idx.z];
    const float area = length ( cross ( v1 - v0, v2 - v0 ) );
    optix::Aabb *aabb = ( optix::Aabb * ) result;

    if ( area > 0.0f && !isinf ( area ) ) {
        aabb->m_min = fminf ( fminf ( v0, v1 ), v2 );
        aabb->m_max = fmaxf ( fmaxf ( v0, v1 ), v2 );

    } else {
        aabb->invalidate();
    }
}
