#include <cuda_runtime.h>

#include <RenderManager.h>
#include <PostProcessor.h>
#include <OutputWindow.h>

#include <BufferData_UserTypes.h>

#include <SFML/System/Clock.hpp>
#include "TrialConfig.h"

#include <FreeImagePlus.h>

#define BENCHMARKFRAMES 1
#define HFOV 60

using namespace grip;



enum CamSequenceStep {
    CAM_SEQ_STEP_FILL_CACHE = 0,
    CAM_SEQ_STEP_RENDER = 1,

    NUM_SEQ_STEPS
};



int main ( int argc, char **argv ) {

    FreeImage_Initialise();    
    
    sf::Clock clock;

    trd::Fifar fifar;

	RenderManager renderManager ( argc, argv, &fifar);
    RenderParameters *renderParameters = renderManager.getRenderParameters();
    
	PostProcessor postProcessor ( "../xml/demo.xml", renderManager.getBufferInfo(), renderManager.getAccumulatedDevicePointers() );
    
	unsigned int width = renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_RES_X );
    unsigned int height = renderParameters->getParameterInt ( ENUM_RENDERPARAMETER_RES_Y );
    OutputWindow outputWindow ( width, height, renderManager.getCmdLineParams().displayFactor );


    bool renderCamSequence = false;

    TrialConfig tc = renderManager.tc;

    uint NUM_FIFAR_FRAMES = tc.frames;

    std::string camFile = tc.fifar;
    if (camFile == "") {
        std::cout << "No cam file given!" << std::endl;
    } else {
        std::cout << "Cam file given: " << camFile << std::endl;
        fifar.readFromFile(camFile.c_str());
        fifar.computeFrameDescriptors(NUM_FIFAR_FRAMES);
        // postProcessor.setParameter<dt_uint32_t>("samplesPerLaunch", 1, 1); 
        // postProcessor.setParameter<dt_float_t>("lightfield_writeProb", 1, 1); 
        // postProcessor.setParameter<dt_uint32_t>("minRecursionDepth", 2, 1);
        // postProcessor.setParameter<dt_uint32_t>("maxRecursionDepth", 2, 1);     
        renderCamSequence = true;
    }

    postProcessor.setParameter<dt_uint32_t>("samplesPerLaunch", tc.samples, 1); 
    // postProcessor.setParameter<dt_float_t>("frameLimit", 0, 1); 
    // postProcessor.setParameter<dt_float_t>("maxCachingProbability", tc.caching_probability, 1); 
    postProcessor.setParameter<dt_uint32_t> ("saveImages", 1, 1);
    postProcessor.setParameter<dt_float_t>("envRot_phi", tc.envRot_phi, 1); 

    std::stringstream ss;
    ss << tc.outputPath;
    ss << tc.id << "/log.csv";
    // ss << numPreviousFrames;
    // std::string theString = ss.str();
    // std::string theString2 = ss.str();

    // while ( theString.length() < 6 ) {
    //     theString = "0" + theString;
    //     theString2 = "0" + theString2;
    // }

    // std::stringstream ss2_exr;
    // std::stringstream ss2_png;
    // ss2_exr << p_tc.id << "/exr/";
    // ss2_png << p_tc.id << "/png/";

    
    ofstream logfile;
    logfile.open (ss.str());
    logfile << "frame_index; time_trace; " << std::endl;
    
    



	/* dummy launch for acceleration structure build */
    std::cout << "Building acceleration structure..." << std::flush;
    unsigned int frame = 1;
    unsigned int numPreviousFrames = 0;
    renderManager.render ( logfile, 0, numPreviousFrames, frame, 0, 0, 0 );
	std::cout << clock.getElapsedTime().asSeconds() << std::endl;
    
	BenchmarkValues bmv;
    std::cout << "Rendering started!" << std::endl;

  

    int camSequenceStep = 0;
    int keyFrameIndex = 0;
    bool presample = true;


    while ( outputWindow.isRunning() ) {
        // error fixing :S
        if ( frame == 2 ) { numPreviousFrames = 0; }



        clock.restart();
		
        bmv.time_total = bmv.time_prepare + bmv.time_rendering + bmv.time_postproc + bmv.time_display + bmv.time_benchmarking;

		//std::cout << "bmv.time_total: " << bmv.time_total << std::endl;
		
        // fill render parameters from post processor
        for ( int i = 0; i < ENUM_NUM_RENDERPARAMETERS_INT; ++i ) {
            enum_parameterNames_int parameter = ( enum_parameterNames_int ) i;
            std::string name = renderParameters->getParameterNameGrip ( parameter );

            if ( name == "" ) { continue; }

            renderParameters->setParameterInt ( parameter, postProcessor.getParameter<dt_uint32_t> ( name ) );
        }




        for ( int i = 0; i < ENUM_NUM_RENDERPARAMETERS_FLOAT; ++i ) {
            enum_parameterNames_float parameter = ( enum_parameterNames_float ) i;
            std::string name = renderParameters->getParameterNameGrip ( parameter );

            if ( name == "" ) { continue; }

            renderParameters->setParameterFloat ( parameter, postProcessor.getParameter<dt_float_t> ( name ) );
        }




        float controllerSensitivity_movement = postProcessor.getParameter<dt_float_t> ( "controllerSensitivity_movement" );
        float controllerSensitivity_rotation = postProcessor.getParameter<dt_float_t> ( "controllerSensitivity_rotation" );
        outputWindow.getSixAxis()->setControllerSensitivity_movement ( controllerSensitivity_movement );
        outputWindow.getSixAxis()->setControllerSensitivity_rotation ( controllerSensitivity_rotation );
        
		unsigned int maxSamples = postProcessor.getParameter<dt_uint32_t> ( "maxSamples_pow2" );
		unsigned int calcReconstruction = postProcessor.getParameter<dt_uint32_t> ( "calcReconstruction" );

        if ( maxSamples != -1 ) {	maxSamples = powf ( 2, maxSamples ); }

        bool paramsForPtChanged = postProcessor.parameterSetChanged ( "pathTracing" );

        // just needed for averaging afterwards
        if ( paramsForPtChanged ) { numPreviousFrames = 0; }
        
        if ( numPreviousFrames == 0 ) { bmv.mrays = 0; }



        bool focusChanged = false;
        outputWindow.eventLoop ( focusChanged );
        
        // std::cout << "LF WRITEPROB: " << postProcessor.getParameter<dt_float_t>("lightfield_writeProb") << std::endl;
        uint oldKeyFrameIndex = keyFrameIndex;

        if (renderCamSequence) {
            if (keyFrameIndex >= fifar.getNumFrames()) {
                std::cout << "Finished rendering of sequence!" << std::endl;
                
                // deactivate lf storage

                // activate lf reconstruction

                // restart cam sequence rendering

                // break;
                // template<class FTYPE> void setParameter(const string& p_key, FTYPE p_param, uint32_t p_size) {

                // postProcessor.setParameter<dt_float_t>("lightfield_writeProb", 0, 1);
                // postProcessor.setParameter<dt_float_t>("weight_real", 1, 1);
                // postProcessor.setParameter<dt_uint32_t>("calcReconstruction", 1, 1);
                // postProcessor.setParameter<dt_uint32_t>("minRecursionDepth", 1, 1);
                // postProcessor.setParameter<dt_uint32_t>("maxRecursionDepth", 1, 1);

                // if (LF_STORE_DIRECT_ILLUM) {
                //     // postProcessor.setParameter<dt_float_t>("weight_directLight", 0, 1);                    
                //     postProcessor.setParameter<dt_uint32_t>("samplesPerLaunch", 1, 1); 
                // }

                // renderCamSequence = true;
                postProcessor.setParameter<dt_uint32_t>("saveImages", 0, 1); 
                // postProcessor.getParameter<dt_uint32_t> ( "saveImages" )
                keyFrameIndex = 0;
                presample = false;
            } else {
                std::cout << "getting key frame: " << keyFrameIndex << std::endl;
                trd::FrameDescriptor fd = fifar.getFrame(keyFrameIndex);
                optix::float3 pos = make_float3(fd.pos[0], fd.pos[1], fd.pos[2]);
                optix::float3 up = make_float3(fd.up[0], fd.up[1], fd.up[2]);
                optix::float3 dir = make_float3(fd.dir[0], fd.dir[1], fd.dir[2]);
                optix::float3 lookAt = pos + dir;
                renderManager.getCamera()->setEye(fd.pos[0], fd.pos[1], fd.pos[2]);
                renderManager.getCamera()->setUp(fd.up[0], fd.up[1], fd.up[2]);
                renderManager.getCamera()->setLookAt(lookAt.x, lookAt.y, lookAt.z);
                
                std::cout << renderManager.getCamera()->getEye().x << std::endl;
                if (presample) {
                    keyFrameIndex += tc.render_each;
                } else {
                    keyFrameIndex += tc.render_each;
                }
            }
            // renderManager.getCamera()->setLookAt(make_float3(fd.pos[0], fd.pos[1], fd.pos[2]) + make_float3(fd.dir[0], fd.dir[1], fd.dir[2]));
        } else {
    		renderManager.getCamera()->transform ( outputWindow.getSixAxis(), bmv.time_total );
        }
        renderManager.updateCameraAndParameters();
        

		bool movement = outputWindow.getSixAxis()->movement();

        if ( postProcessor.getParameter<dt_uint32_t> ( "cam_autoFocus" ) ) {
            if ( movement || frame == 1 ) {
                focusChanged = true;
            }
        }

        if ( focusChanged ) {
            // renderManager.refocusCamera ( outputWindow.getFocalPointX(), outputWindow.getFocalPointY() );
            numPreviousFrames = 0;
            focusChanged = false;
        }

        if ( movement || paramsForPtChanged || renderCamSequence) {
            renderManager.setMinDepthMode();
            numPreviousFrames = 0;
        }



        bmv.time_prepare = clock.getElapsedTime().asSeconds();
        clock.restart();

		float lf_writeProb = postProcessor.getParameter<dt_float_t> ( "lightfield_writeProb" );
		
        if ( !postProcessor.getParameter<dt_uint32_t> ( "pauseRendering" ) ) { renderManager.render ( logfile, oldKeyFrameIndex, numPreviousFrames, frame, maxSamples, lf_writeProb, calcReconstruction ); }

        bmv.time_rendering = clock.getElapsedTime().asSeconds();
        bmv.time_benchmarking += bmv.time_rendering;
        clock.restart();

         if ( postProcessor.getParameter<dt_uint32_t> ( "saveImages" ) && oldKeyFrameIndex < NUM_FIFAR_FRAMES) { 
           
            float4* outBuf = renderManager.writeOutbufToHost(); 
            outputWindow.saveImage ( tc, oldKeyFrameIndex, outBuf );
        } else {
            break;
        }
        
		postProcessor.markGripBuffersDirty();
        postProcessor.run();
        
		bmv.mrays = postProcessor.getParameter<dt_uint32_t> ( "numRays" ) / ( bmv.time_benchmarking * 1000000.0 );
        bmv.time_benchmarking = 0.f;
		bmv.time_postproc = clock.getElapsedTime().asSeconds();
        


		clock.restart();
        
		dt_UniversalImage_uint8_4_sptr outputImage = postProcessor.getParameter<dt_UniversalImage_uint8_4_sptr> ( "output_image" );
        outputWindow.setImageData ( outputImage->getDataPointer().get() );
        outputWindow.drawImage();
        outputWindow.drawStats (
            bmv,
            renderManager.getRenderParameters(),
            numPreviousFrames,
            renderManager.getCamera()->getEye().x, renderManager.getCamera()->getEye().y, renderManager.getCamera()->getEye().z,
            renderManager.getCamera()->getDirection().x, renderManager.getCamera()->getDirection().y, renderManager.getCamera()->getDirection().z
        );
        outputWindow.getWindow()->display();




        if (outputWindow.getFifarInfo_addKeyFrame()) {
            std::cout << "ADDING KEYFRAME" << std::endl;
            // create frame descriptor
            trd::FrameDescriptor fd;
            float3 pos = renderManager.getCamera()->getEye();
            float3 up = renderManager.getCamera()->getUp();
            float3 dir = renderManager.getCamera()->getDirection();

            std::cout << "pos: " << pos.x << ", " << pos.y << ", " << pos.z << std::endl;
            std::cout << "up: " << up.x << ", " << up.y << ", " << up.z << std::endl;
            std::cout << "dir: " << dir.x << ", " << dir.y << ", " << dir.z << std::endl;

            fd.pos[0] = pos.x;
            fd.pos[1] = pos.y;
            fd.pos[2] = pos.z;
            fd.up[0] = up.x;
            fd.up[1] = up.y;
            fd.up[2] = up.z;
            fd.dir[0] = dir.x;
            fd.dir[1] = dir.y;
            fd.dir[2] = dir.z;

            fifar.addKeyFrame(fd);
        }


        if (outputWindow.getFifarInfo_writeToDisk()) {
            std::cout << "WRITING KEYFRAMES TO DISK" << std::endl;
            fifar.writeKeyFrameDescriptorsToFile("fifarDescriptors.txt");
        }

        // if ( postProcessor.getParameter<dt_uint32_t> ( "saveImages" ) && frame < NUM_FIFAR_FRAMES)  { outputWindow.saveImage ( frame ); }
        
         if ( postProcessor.parameterSetChanged ( "saveLightfield" ) ) {
 		
 			renderManager.saveLightfield();
 					
 		}

        
  

        bmv.time_display = clock.getElapsedTime().asSeconds();
    }

    return 0;
}
