#ifndef GRID_BUILDER_GPU_HPP_
#define GRID_BUILDER_GPU_HPP_

  #include <utils/cuda_error_handler.h>

void buildGridGPU(uint *d_Cells, gutil::Tri *d_Tris, unsigned int numTris, gutil::BBox uniform_bb, unsigned int min_level);

unsigned long long int popCountGPU(uint *d_Cells, size_t no_cells);

#ifdef __NVCC__

__global__ void pop_count_gpu(unsigned long long int *d_Counter, uint *d_Cells, size_t no_cells) {
    int idx = (blockIdx.x * blockDim.x) + threadIdx.x;

    if(idx >= no_cells) return;

    unsigned int count = __popc(d_Cells[idx]);

    atomicAdd(d_Counter, count);

}

unsigned long long int popCountGPU(uint *d_Cells, size_t no_cells) {
  //i.e. 64 threads per block
  int2 blockSize;
  blockSize.x = 128;
  blockSize.y = 1;

  unsigned long long int *d_count;
  cudaMalloc((void**) &d_count, sizeof(unsigned long long int));
	cudaMemset(d_count, 0, sizeof(unsigned long long int));

  dim3 threadsPerBlock(blockSize.x, blockSize.y);
  dim3 numBlocks(ceil(no_cells / (float)blockSize.x) + 1, 1, 1);

  pop_count_gpu << <numBlocks, threadsPerBlock >> >(d_count, d_Cells, no_cells);


  int retVal = 0;
	cudaMemcpy(&retVal, d_count, sizeof(unsigned long long int), cudaMemcpyDeviceToHost);
  cudaFree(d_count);
	return retVal;

}

__device__ void setCell(uint *d_Cells, unsigned int x, unsigned int y, unsigned int z, uint level, uint *p_globalCounter) {

  unsigned int resolution = round(pow(2, level));

  auto start_idx=0;
  //start index
  // for(int i=min_level; i < level; ++i) {
  //   size_t per_level = pow(2,i);
  //   //32bit = 4*4*2
  //   start_idx += (per_level / 4) * (per_level / 4) * (per_level / 2);
  // }

  size_t bit_blocks_in_x = resolution / 4;
  size_t bit_blocks_in_y = resolution / 4;
  //size_t bit_blocks_in_z = resolution / 2;

  size_t bit_blocks_x = x / 4;
  size_t bit_blocks_y = y / 4;
  size_t bit_blocks_z = z / 2;

  unsigned int bit_set_in_block_x = x % 4;
  unsigned int bit_set_in_block_y = y % 4;
  unsigned int bit_set_in_block_z = z % 2;

  int shift = (bit_set_in_block_z * (4*4)) + (bit_set_in_block_y * 4) + bit_set_in_block_x;
  unsigned int val = 0x1 << shift;
  // if (/**/shift > 31) {
    // printf("shift: %d\n", shift);
  // }

  auto block_idx_lin = (bit_blocks_z * bit_blocks_in_x * bit_blocks_in_y) + (bit_blocks_y * bit_blocks_in_x) + bit_blocks_x;

  unsigned int oldVal = atomicOr(&d_Cells[start_idx + block_idx_lin] , val);
  if ((oldVal & val) == 0) {
    atomicAdd(p_globalCounter, 1);
  }
}

__global__ void build_grid_gpu(uint *d_Cells, gutil::Tri *d_Tris, unsigned int numTris, gutil::vec3 uniform_bb_min,
                                        gutil::vec3 uniform_bb_max,  unsigned int minLevel, uint* p_globalCounter) {
        // bool trisSorted = new bool[tris.size()]
        //for(auto i=0; i < tris.size(); ++i) {
        int triID = (blockIdx.x * blockDim.x) + threadIdx.x;

        if(triID >= numTris) return;

      

        const gutil::Tri tri = load_prim(&(d_Tris[triID]));

        auto v0 = tri.v0;

        auto v1 = v0 - tri.e1;
        auto v2 = v0 + tri.e2;

        // std::cout << "tri v1: " << v1.x << ", " << v1.y << ", " << v1.z << std::endl;
        // std::cout << "tri v2: " << v2.x << ", " << v2.y << ", " << v2.z << std::endl;
        // gutil::BBox uniform_bb(uniform_bb_min, uniform_bb_max);


     

        // if (sub != 11) {
        //   printf("subdivision mismatch! %d, %d\n", sub, min_level);
        // }

        // printf("sub, minLevel: %u, %u\n", sub, min_level);

        // uint cell_res_pow = pow(2, minLevel);
        // uint cell_res_powf = powf(2, minLevel);
        // uint cell_res_fpowf = fpowf(2, minLevel);


        // printf("cell_res: %u, %u\n", cell_res_pow, cell_res_powf);

        uint cell_res = round(pow(2, minLevel));

        gutil::vec3 bb_extent = uniform_bb_max - uniform_bb_min;
        float cell_size = bb_extent.x / cell_res;

        // std::cout << "cell_size: " << cell_size << std::endl;

        gutil::BBox prims_bbox = tri.bbox();

        prims_bbox.min -= uniform_bb_min;
        prims_bbox.max -= uniform_bb_min;

        gutil::vec3 cell_idx_min = prims_bbox.min / gutil::vec3(cell_size);
        gutil::vec3 cell_idx_max = prims_bbox.max / gutil::vec3(cell_size);

        //std::cout << "cell_idx_min: " << cell_idx_min.x << ", " << cell_idx_min.y << "," << cell_idx_min.z <<std::endl;
        //std::cout << "cell_idx_max: " << cell_idx_max.x << ", " << cell_idx_max.y << "," << cell_idx_max.z <<std::endl;

        gutil::uivec3 start_cell((uint)floorf(cell_idx_min.x),
                                 (uint)floorf(cell_idx_min.y),
                                 (uint)floorf(cell_idx_min.z));

        gutil::uivec3 end_cell((uint)ceilf(cell_idx_max.x),
                               (uint)ceilf(cell_idx_max.y),
                               (uint)ceilf(cell_idx_max.z));

        end_cell.x = min(end_cell.x, cell_res - 1);
        end_cell.y = min(end_cell.y, cell_res - 1);
        end_cell.z = min(end_cell.z, cell_res - 1);
        // std::cout << "start_cell: " << start_cell.x << ", " << start_cell.y << ", " << start_cell.z << std::endl;
        // std::cout << "end_cell: " << end_cell.x << ", " << end_cell.y << ", " << end_cell.z << std::endl;
        // std::cout << "prims_bbox.min: " << prims_bbox.min.x << ", " << prims_bbox.min.y << ", " << prims_bbox.min.z << std::endl;

        for(auto x = start_cell.x; x <= end_cell.x; ++x) {
                for(auto y = start_cell.y; y <= end_cell.y; ++y) {
                        for(auto z = start_cell.z; z <= end_cell.z; ++z) {
#define eps 0.0001f
                          // atomicAdd(p_globalCounter, 1);
                                gutil::vec3 minBox(uniform_bb_min.x + x * cell_size - eps,
                                                   uniform_bb_min.y + y * cell_size - eps,
                                                   uniform_bb_min.z + z * cell_size - eps
                                                   );
                                gutil::vec3 maxBox(uniform_bb_min.x + (x+1) * cell_size + eps,
                                                   uniform_bb_min.y + (y+1) * cell_size + eps,
                                                   uniform_bb_min.z + (z+1) * cell_size + eps
                                                   );

                                gutil::BBox cell_box(minBox, maxBox);

                                if(intersect_prim_cell(tri, cell_box)) { //|| intersect_prim_cell(tempTri0, cell_box) || intersect_prim_cell(tempTri1, cell_box)) {
                                        // min_level=1;
                                        //cells.push_back(Cell(x,y,z,sub));
                                        setCell(d_Cells, x,y,z, minLevel, p_globalCounter);
                                }
                        }

                }
        }

}

void buildGridGPU(uint *d_Cells, gutil::Tri *d_Tris, unsigned int numTris, gutil::BBox uniform_bb, unsigned int min_level) {
        //i.e. 64 threads per block
        int2 blockSize;
        blockSize.x = 128;
        blockSize.y = 1;

        std::cout << "numTris: " << numTris;
        std::cout << "uniformbb min: " << uniform_bb.min.x << ", " << uniform_bb.min.y << ", " << uniform_bb.min.z << std::endl;
        std::cout << "uniformbb max: " << uniform_bb.max.x << ", " << uniform_bb.max.y << ", " << uniform_bb.max.z << std::endl;


  CUDA_LAST_ERROR();
        dim3 threadsPerBlock(blockSize.x, blockSize.y);
        dim3 numBlocks(ceil(numTris / (float)blockSize.x), 1, 1);

        uint *globalCounter_d = 0;
        cudaMalloc(&globalCounter_d, sizeof(uint));

          CUDA_LAST_ERROR();
        cudaMemset(globalCounter_d, 0, sizeof(uint));
          CUDA_LAST_ERROR();

        std::cout << "Calling build_grid_gpu with min_level "  << min_level << std::endl;

        build_grid_gpu <<< numBlocks, threadsPerBlock >>>(d_Cells, d_Tris, numTris, uniform_bb.min, uniform_bb.max, min_level, globalCounter_d);
  CUDA_LAST_ERROR();

        uint globalCounter_h;
        cudaMemcpy(&globalCounter_h, globalCounter_d, sizeof(uint), cudaMemcpyDeviceToHost);
        std::cout << "globalCounter: " << globalCounter_h << std::endl;
  CUDA_LAST_ERROR();
}
#endif


#endif
