#ifndef _GL_SHADER_MANAGER_H_
#define _GL_SHADER_MANAGER_H_

#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <iostream>
#include <fstream>
#include <sstream>

// in order to get function prototypes from glext.h, define GL_GLEXT_PROTOTYPES before including glext.h
//#define GL_GLEXT_PROTOTYPES

#ifdef __APPLE__
	#include <GLUT/glut.h>
#else
	#include <GL/glut.h>
#endif


#include <GL/glew.h>



/**
 @brief A class to manage GLSL Shader
 @author Martin Weier
 Created by Martin Weier @ HBRS 2011

 \paragraph <Usage> (Usage) {
    ..
    //init
    //Load vertex, fragment and geometry shader
    aShaderManager.loadFromFile/loadFromMemory(..);
    ...
    aShader.initialize();
    ...
    //draw
    aShader.begin();

    //glcode to draw primitives

    aShader.end();
 }

**/

class glShaderManager{
public:
    enum ShaderType{
        VERTEX_SHADER=0,
        FRAGMENT_SHADER,
        GEOMETRY_SHADER
    };

protected:
    bool mVertexShaderLoaded;
    bool mFragmentShaderLoaded;
    bool mGeometryShaderLoaded;
    bool mIsInitialized;

    GLhandleARB mVertexShader, mFragmentShader, mGeometryShader, mShaderProgObj;

    std::string mFragmentShaderCode;
    std::string mVertexShaderCode;
    std::string mGeometryShaderCode;


protected:

    /**
    *  @brief Loads the content of the file to a string
    *  @param pFilename : a filename
    *  @return string holding the files content
    **/
    const std::string loadFromFile(std::string pFilename){
        std::string shaderCode;

        std::ifstream ifs;
        ifs.open(pFilename.c_str());

        if(!(ifs.is_open())){
            std::cerr << "ShaderManager: Could not open shader file <" << pFilename << ">" << std::endl;
            exit(1);
        }

        if ( ifs )
        {
            std::stringstream buffer;
            buffer << ifs.rdbuf();
            ifs.close();
            shaderCode = buffer.str();
        }

        return shaderCode;
    }



    /**
    *! \brief print the ARB Info Log. Shows current state and compilation results
    *  \param pARBHandle : handle to the ARB object, i.e. the vertex, fragment or geometry shader arb
    **/
    void printShaderInfoLog(GLhandleARB &pARBHandle){
        char infobuffer[1000];
        int infobufferlen = 0;
        glGetInfoLogARB(pARBHandle, 999, &infobufferlen, infobuffer);
        infobuffer[infobufferlen] = 0;
        std::cout << infobuffer << std::endl;
    }




public:

    /*Constructor*/
    glShaderManager() : mVertexShaderLoaded(false), mFragmentShaderLoaded(false),  mGeometryShaderLoaded(false),
                        mIsInitialized(false)/*, mFragmentShaderCode(0), mVertexShaderCode(0), mGeometryShaderCode(0)*/
    {

    }

    /*Destructor*/
    virtual ~glShaderManager(){
        //Cleanup
        if(isVertexShaderLoaded())
            glDeleteShader(mVertexShader);
        if(isFragmentShaderLoaded())
            glDeleteShader(mFragmentShader);
        if(isGeometryShaderLoaded())
            glDeleteShader(mGeometryShader);

        if(mIsInitialized)
            glDeleteProgram(mShaderProgObj);
    }


    void printProgramInfoLog(){
        char infobuffer[1000];
        int infobufferlen = 0;
        glGetProgramInfoLog(mShaderProgObj, 999, &infobufferlen, infobuffer);
        infobuffer[infobufferlen] = 0;
        std::cout << infobuffer << std::endl;
    }


    /**
    * @brief Loads a shader from file. The type must be specified using the enum ShaderType
    * @param pFilename  file to load
    * @param pType      The shader type. See ShaderType enum
    **/
    void loadShaderFromFile(std::string pFilename, ShaderType pType){
            if(pType == VERTEX_SHADER){
                mVertexShader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
                mVertexShaderCode = loadFromFile(pFilename);
                const char * vv = mVertexShaderCode.c_str();
                glShaderSource(mVertexShader, 1, &vv,NULL);
                mVertexShaderLoaded=true;
        	}
        	if(pType == FRAGMENT_SHADER){
                mFragmentShader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
                mFragmentShaderCode = loadFromFile(pFilename);
                const char * ff = mFragmentShaderCode.c_str();
                glShaderSource(mFragmentShader, 1, &ff,NULL);
                mFragmentShaderLoaded=true;
        	}
        	if(pType == GEOMETRY_SHADER){
                mGeometryShader = glCreateShaderObjectARB(GL_GEOMETRY_SHADER_ARB);
                mGeometryShaderCode = loadFromFile(pFilename);
                const char * gg = mGeometryShaderCode.c_str();
                glShaderSource(mGeometryShader, 1, &gg,NULL);
                mGeometryShaderLoaded = true;
        	}
    }

    /**
    * @brief Loads a shader from memory. String contains the Shader code. The type must be specified using the enum ShaderType
    * @param pShaderCode    The shader code as string
    * @param pType      The shader type. See ShaderType enum
    **/
    void loadShaderFromMemory(std::string pShaderCode, ShaderType pType){
         if(pType == VERTEX_SHADER){
                mVertexShader = glCreateShader(GL_VERTEX_SHADER_ARB);
                mVertexShaderCode = pShaderCode;
                const char * vv = mVertexShaderCode.c_str();
                glShaderSource(mVertexShader, 1, &vv,NULL);
                mVertexShaderLoaded=true;
        	}
        	if(pType == FRAGMENT_SHADER){
                mFragmentShader = glCreateShader(GL_FRAGMENT_SHADER_ARB);
                mFragmentShaderCode = pShaderCode;
                const char * ff = mFragmentShaderCode.c_str();
                glShaderSource(mFragmentShader, 1, &ff,NULL);
                mFragmentShaderLoaded=true;
        	}
        	if(pType == GEOMETRY_SHADER){
                mGeometryShader = glCreateShader(GL_GEOMETRY_SHADER_ARB);
                mGeometryShaderCode = pShaderCode;
                const char * gg = mGeometryShaderCode.c_str();
                glShaderSource(mGeometryShader, 1, &gg,NULL);
                mGeometryShaderLoaded = true;
        	}
    }


    inline const bool isVertexShaderLoaded(){
        return mVertexShaderLoaded;
    }

    inline const bool isFragmentShaderLoaded(){
        return mFragmentShaderLoaded;
    }

    inline const bool isGeometryShaderLoaded(){
        return mGeometryShaderLoaded;
    }

    inline GLhandleARB getShaderProgram(){
            return mShaderProgObj;
    }


    inline GLhandleARB getVertexShaderARB() {
        return mVertexShader;
    }


    inline GLhandleARB getFragmentShaderARB() {
        return mFragmentShader;
    }


    inline GLhandleARB getGeometryShaderARB() {
        return mGeometryShader;
    }

    /**
    * @brief Get the location of the uniform variables within the shader
    * @param pLocation Name of the uniform varaiable within the shader
    **/
    GLint getUniformLocation(const char *pLocation){
        return glGetUniformLocationARB(mShaderProgObj, pLocation);
    }


    /**
    * @brief Get the location of the attributes within the shader
    * @param pLocation Name of the attribute within the shader
    **/
    GLint getAttributeLocation(const char *pLocation){
        return glGetAttribLocation(mShaderProgObj, pLocation);
    }


	GLuint getUniformBlockIdx(const char* block_name) {
		return glGetUniformBlockIndex(mShaderProgObj, block_name);
		}

    /**
    * @brief Initializes the shader, i.e compiling and linking
    * Using vertex, fragment and geometry shader code if loaded/provided
    * @param pShowCompilationResults: Print compilation results e.g. compiler and linker errors and warnings
    **/
    inline bool compile(bool pShowCompilationResults=false){
            if(!isVertexShaderLoaded() && !isFragmentShaderLoaded() && !isGeometryShaderLoaded()){
                std::cout << "ShaderManager Warning: No Shader code specified: " << std::endl;
                return false;
            }

            if(isVertexShaderLoaded()){
                glCompileShader(mVertexShader);

                if(pShowCompilationResults){
                    std::cout << "Vertex Shader:" << std::endl;
                    std::cout << mVertexShaderCode << std::endl;
                    std::cout << "Vertex Shader compile result: " << std::endl;
                    printShaderInfoLog(mVertexShader);
                }

                GLint shaderState;
                glGetShaderiv(mVertexShader, GL_COMPILE_STATUS, &shaderState);
                if (!shaderState) {
                    std::cout << "ShaderManager Warning: Failed to compile Vertex Shader" << std::endl;
                    glDeleteShader(mVertexShader);
                    mVertexShaderLoaded = false;
					std::cout << "" << std::endl;
					return false;
                } else {
                    std::cout << "ShaderManager: Vertex Shader compiled succesful" << std::endl;
					std::cout << "" << std::endl;
                }
            }


            if(isGeometryShaderLoaded()){
                glCompileShader(mGeometryShader);

                if(pShowCompilationResults){
                    std::cout << "Geometry Shader:" << std::endl;
                    std::cout << mGeometryShaderCode << std::endl;
                    std::cout << "Geometry Shader compile result: " << std::endl;
                    printShaderInfoLog(mGeometryShader);
                }

                GLint shaderState;
                glGetShaderiv(mGeometryShader, GL_COMPILE_STATUS, &shaderState);
                if (!shaderState) {
                    std::cout << "ShaderManager Warning: Failed to compile Geometry Shader" << std::endl;
                    std::cout << "" << std::endl;
                    glDeleteShader(mGeometryShader);
                    mGeometryShaderLoaded = false;
					std::cout << "" << std::endl;
					return false;
                } else {
                    std::cout << "ShaderManager: Geometry Shader compiled succesful" << std::endl;
					std::cout << "" << std::endl;
                }

            }

           if(isFragmentShaderLoaded()){
                glCompileShader(mFragmentShader);

                if(pShowCompilationResults){
                    std::cout << "Fragment Shader:" << std::endl;
                    std::cout << mFragmentShaderCode << std::endl;
                    std::cout << "Fragment Shader compile result: " << std::endl;
                    printShaderInfoLog(mFragmentShader);
                }

                GLint shaderState;
                glGetShaderiv(mFragmentShader, GL_COMPILE_STATUS, &shaderState);
                if (!shaderState) {
                    std::cout << "ShaderManager Warning: Failed to compile Fragment Shader" << std::endl;
                    std::cout << "" << std::endl;
                    glDeleteShader(mFragmentShader);
                    mFragmentShaderLoaded = false;
					std::cout << "" << std::endl;
					return false;
                } else {
                    std::cout << "ShaderManager: Fragment Shader compiled succesful" << std::endl;
					std::cout << "" << std::endl;
                }

            }


            if(isVertexShaderLoaded() || isFragmentShaderLoaded() || isGeometryShaderLoaded()){
                mShaderProgObj = glCreateProgramObjectARB();
            } else {
                std::cout << "ShaderManager Warning: Could not compile one shader at least" << std::endl;
                return false;
            }

            if(isVertexShaderLoaded())
                glAttachShader(mShaderProgObj,mVertexShader);
            if(isFragmentShaderLoaded())
                glAttachShader(mShaderProgObj,mFragmentShader);
            if(isGeometryShaderLoaded())
                glAttachShader(mShaderProgObj,mGeometryShader);





            return true;
    }

    inline bool link(){
       //Once we have compile the shader we link it
            glLinkProgram(mShaderProgObj);
            GLint linkerState;
            glGetProgramiv(mShaderProgObj, GL_LINK_STATUS, &linkerState);
            if (!linkerState) {
                std::cout << "Shader Manager Error: Failed to link shader program " << std::endl;
                std::cout << "Linker Results: " << std::endl;
                printProgramInfoLog();
                glDeleteProgram(mShaderProgObj);
                return false;
            } else {
                std::cout << "Shader Manager: Linking successful" << std::endl;
            }
  		mIsInitialized = true;

        return true;
    }

    inline void validate(){
            glValidateProgram(mShaderProgObj);
	        printProgramInfoLog();
    }


    /**
    * @brief Setting GL to use the specified shader
    **/
    inline void begin(){
        if(!mIsInitialized){
            std::cout << "ShaderManager: Shader not initialized" << std::endl;
            std::cout << "ShaderManager: Initialize first" << std::endl;
            return;
        }

        //Activate Shader
        glUseProgram(mShaderProgObj);
    }


    /**
    * @brief Tells GL to stop using the shader
    **/
    inline void end(){
        //Set Shader prog back to default
        glUseProgramObjectARB(0);
    }

};


#endif
