/* -*- mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*- */
/*
 *  Main authors:
 *     Christian Schulte <schulte@gecode.org>
 *
 *  Contributing authors:
 *     Stefano Gualandi <stefano.gualandi@gmail.com>
 *
 *  Copyright:
 *     Christian Schulte, 2004
 *     Stefano Gualandi, 2013
 *
 *  Last modified:
 *     $Date: 2017-02-22 04:53:03 +0100 (Wed, 22 Feb 2017) $ by $Author: schulte $
 *     $Revision: 15469 $
 *
 *  This file is part of Gecode, the generic constraint
 *  development environment:
 *     http://www.gecode.org
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include <gecode/driver.hh>
#include <gecode/int.hh>
#include <cssolver/primitive-cs.hpp>
#include <gecode/gist.hh>


using namespace Gecode;



namespace cs {
/**
 * \name %Graph specification for graph coloring
 *
 * The edges are described by an array of integers with even number
 * of elements, terminated by the elements -1,-1.
 * The cliques are described by an array of integers, where the first
 * integer gives the size of the clique, the following elements are
 * nodes for each clique. The cliques are terminated by -1 for clique size
 *
 * \relates GraphColor
 */
//@{
/// %Graph specification
class GraphColorSpec {
public:
const int n_v;    ///< Number of nodes
const int* e;     ///< Edges


GraphColorSpec(const int n_v0, const int* e0)
        : n_v(n_v0), e(e0) {
}
};

/// First example graph: edges
std::vector<int> g1_e;
unsigned int no_of_nodes;


const int g1_e_arr[] = {
        160,184,  192,199,  0,129,  20,80,  8,29,  93,171,
        101,165,  124,193,  2,100,  66,173,  151,191,  164,187,
        3,130,  118,176,  121,184,  25,106,  159,193,  121,123,
        5,62,  97,101,  6,143,  123,163,  19,125,  17,108,
        122,168,  181,184,  25,41,  62,70,  29,103,  48,67,
        46,160,  79,170,  143,152,  38,184,  2,40,  191,195,
        7,196,  62,199,  76,141,  82,166,  36,80,  51,189,
        13,97,  3,192,  90,180,  47,176,  13,172,  92,121,
        50,64,  65,113,  108,123,  26,106,  34,153,  90,123,
        34,39,  116,178,  22,179,  50,61,  84,105,  84,93,
        19,108,  29,59,  63,185,  119,129,  50,177,  80,194,
        13,36,  46,56,  38,144,  82,193,  72,93,  49,95,
        42,155,  117,140,  109,189,  19,35,  31,125,  118,191,
        163,169,  40,167,  91,127,  3,121,  124,149,  40,174,
        30,175,  19,132,  18,165,  34,93,  37,63,  10,55,
        88,95,  76,122,  7,91,  25,141,  29,173,  139,173,
        8,130,  110,158,  81,174,  113,114,  95,182,  136,149,
        5,199,  56,106,  36,120,  133,187,  111,172,  19,34,
        96,197,  32,108,  27,63,  50,188,  20,116,  50,118,
        10,50,  24,172,  86,138,  35,50,  141,153,  98,132,
        70,143,  1,97,  8,160,  37,170,  4,73,  1,94,
        88,146,  59,61,  104,156,  62,172,  117,139,  66,189,
        33,134,  122,169,  95,163,  95,152,  83,140,  110,189,
        147,159,  22,147,  59,173,  30,41,  33,183,  181,187,
        88,105,  93,151,  6,130,  24,30,  84,130,  72,120,
        118,159,  147,189,  122,149,  24,175,  39,169,  164,186,
        93,187,  13,156,  119,176,  73,91,  174,178,  71,198,
        10,134,  30,101,  79,93,  180,187,  1,50,  51,59,
        18,169,  73,153,  1,198,  137,154,  61,106,  80,113,
        48,142,  100,111,  97,133,  82,97,  136,170,  53,134,
        65,177,  7,80,  73,137,  6,70,  115,166,  72,196,
        40,109,  91,101,  2,177,  120,185,  55,65,  72,166,
        104,165,  173,187,  54,71,  3,61,  52,56,  120,149,
        64,72,  42,43,  75,185,  62,68,  108,147,  30,111,
        25,58,  39,93,  75,117,  61,194,  140,153,  80,121,
        93,102,  9,177,  7,163,  17,70,  5,168,  63,178,
        74,160,  148,158,  9,84,  30,76,  63,80,  68,99,
        20,152,  7,182,  7,22,  71,134,  32,100,  107,164,
        23,62,  5,98,  130,192,  65,144,  139,161,  24,124,
        31,47,  29,140,  61,153,  53,109,  20,26,  143,160,
        47,195,  171,172,  185,193,  128,173,  38,96,  14,171,
        176,199,  111,139,  21,54,  80,171,  116,185,  184,199,
        37,88,  62,133,  3,150,  48,109,  46,176,  24,178,
        59,172,  180,198,  64,109,  110,111,  101,146,  66,164,
        5,117,  144,179,  71,126,  166,169,  107,151,  46,85,
        106,139,  27,153,  97,148,  68,185,  17,179,  10,142,
        168,169,  4,46,  113,152,  52,176,  6,38,  22,48,
        20,120,  2,84,  71,85,  91,116,  0,189,  116,197,
        142,147,  33,165,  86,198,  146,149,  152,187,  44,62,
        48,175,  56,150,  63,161,  71,164,  17,171,  19,66,       -1,-1
};


//@}

/**
 * \brief %Example: Clique-based graph coloring
 *
 * \ingroup Example
 *
 */
class GraphColor : public IntMinimizeScript {
private:
const GraphColorSpec &g;
/// Color of nodes
IntVarArray v;
/// Number of colors
IntVar m;
public:
/// Model variants
enum {
        MODEL_NONE, ///< No lower bound
        MODEL_CLIQUE ///< Use maximal clique size as lower bound
};
/// Branching to use for model
enum {
        BRANCH_DEGREE,     ///< Choose variable with largest degree
        BRANCH_SIZE,       ///< Choose variablee with smallest size
        BRANCH_DEGREE_SIZE, ///< Choose variable with largest degree/size
        BRANCH_AFC_SIZE,   ///< Choose variable with largest afc/size
        BRANCH_ACTION_SIZE ///< Choose variable with smallest size/action
};
/// Symmetry variants
enum {
        SYMMETRY_NONE,  ///< Simple symmetry
        SYMMETRY_LDSB   ///< Use LDSB for value symmetry breaking
};
/// The actual model
GraphColor(const SizeOptions& opt)
        : IntMinimizeScript(opt)
        , g(GraphColorSpec(no_of_nodes, g1_e.data())),
        v(*this,g.n_v,0,g.n_v-1),
        m(*this,0,g.n_v-1)

{


        rel(*this, v, IRT_LQ, m);
        for (int i = 0; g.e[i] != -1; i += 2)
                rel(*this, v[g.e[i]], IRT_NQ, v[g.e[i+1]]);

/*
        const int* c = g.c;
        for (int i = *c++; i--; c++)
                rel(*this, v[*c], IRT_EQ, i);
        while (*c != -1) {
                int n = *c;
                IntVarArgs x(n); c++;
                for (int i = n; i--; c++)
                        x[i] = v[*c];
                distinct(*this, x, opt.ipl());
                //if (opt.model() == MODEL_CLIQUE)
                //        rel(*this, m, IRT_GQ, n-1);

        }
*/

        // Branching on the number of colors
        branch(*this, m, INT_VAL_MIN());
        if (opt.symmetry() == SYMMETRY_NONE) {
                // Branching without symmetry breaking
                switch (opt.branching()) {
                case BRANCH_SIZE:
                        branch(*this, v, INT_VAR_SIZE_MIN(), INT_VAL_MIN());
                        break;
                case BRANCH_DEGREE:
                        branch(*this, v, tiebreak(INT_VAR_DEGREE_MAX(),INT_VAR_SIZE_MIN()),
                               INT_VAL_MIN());
                        break;
                case BRANCH_DEGREE_SIZE:
                        branch(*this, v, INT_VAR_DEGREE_SIZE_MAX(), INT_VAL_MIN());
                        break;
                case BRANCH_AFC_SIZE:
                        branch(*this, v, INT_VAR_AFC_SIZE_MAX(opt.decay()), INT_VAL_MIN());
                        break;
                case BRANCH_ACTION_SIZE:
                        branch(*this, v, INT_VAR_ACTION_SIZE_MAX(opt.decay()), INT_VAL_MIN());
                        break;
                }
        } else { // opt.symmetry() == SYMMETRY_LDSB
                // Branching while considering value symmetry breaking
                // (every permutation of color values gives equivalent solutions)
                Symmetries syms;
                syms << ValueSymmetry(IntArgs::create(g.n_v,0));
                switch (opt.branching()) {
                case BRANCH_SIZE:
                        branch(*this, v, INT_VAR_SIZE_MIN(), INT_VAL_MIN(), syms);
                        break;
                case BRANCH_DEGREE:
                        branch(*this, v, tiebreak(INT_VAR_DEGREE_MAX(),
                                                  INT_VAR_SIZE_MIN()),
                               INT_VAL_MIN(), syms);
                        break;
                case BRANCH_DEGREE_SIZE:
                        branch(*this, v, INT_VAR_DEGREE_SIZE_MAX(), INT_VAL_MIN(), syms);
                        break;
                case BRANCH_AFC_SIZE:
                        branch(*this, v, INT_VAR_AFC_SIZE_MAX(opt.decay()),
                               INT_VAL_MIN(), syms);
                        break;
                case BRANCH_ACTION_SIZE:
                        branch(*this, v, INT_VAR_ACTION_SIZE_MAX(opt.decay()),
                               INT_VAL_MIN(), syms);
                        break;
                }
        }

}
/// Cost function
virtual IntVar cost(void) const {
        return m;
}
/// Constructor for cloning \a s
GraphColor(bool share, GraphColor& s) : IntMinimizeScript(share,s), g(s.g) {
        v.update(*this, share, s.v);
        m.update(*this, share, s.m);
}
/// Copying during cloning
virtual Space*
copy(bool share) {
        return new GraphColor(share,*this);
}
/// Print the solution
virtual void
print(std::ostream& os) const {
        os << "\tm = " << m << std::endl
           << "\tv[] = {";
        for (int i = 0; i < v.size(); i++) {
                os << v[i] << ", ";
                if ((i+1) % 15 == 0)
                        os << std::endl << "\t       ";
        }
        os << "};" << std::endl;

        //Gist::dfs(this);
}
};


/** \brief Main-function
 *  \relates GraphColor
 */
int run(tFaceNeighbours &neighbours) {

        uint64_t max_node = 0;
        //Fill data

        for(uint64_t i=0; i<neighbours.size(); ++i) {
          //for(int j=0; j < neighbours[i].size(); ++j) {
          for(const auto &n : neighbours[i]) {
            if(i == n) continue;
            g1_e.push_back(i);
            g1_e.push_back(n);
            max_node = std::max(max_node, i);
            max_node = std::max(max_node, n);
            //std::cout << "Pushin back <" << i << ", " << n << ">" << std::endl;
          }
        }

        g1_e.push_back(-1);
        g1_e.push_back(-1);

        std::cout << "g1_e size: " << g1_e.size() << std::endl;
        std::cout << "max_node: " << max_node << std::endl;
        std::cout << "neighbours.size() " << neighbours.size() << std::endl;

        no_of_nodes = max_node + 1;
        /*for (int i = 0; g1_e_arr[i] != -1; i ++) {
            g1_e.push_back(g1_e_arr[i]);
        }

        g1_e.push_back(-1);
        g1_e.push_back(-1);

        no_of_nodes = 200;//neighbours.size();
        */

        SizeOptions opt("GraphColor");
        opt.ipl(IPL_DOM);
        opt.solutions(0);

        opt.model(GraphColor::MODEL_NONE);
        opt.model(GraphColor::MODEL_NONE, "none",
                  "no lower bound");
        //opt.model(GraphColor::MODEL_CLIQUE, "clique",
        //          "use maximal clique size as lower bound");

        opt.branching(GraphColor::BRANCH_DEGREE);
        opt.branching(GraphColor::BRANCH_DEGREE,      "degree");
        opt.branching(GraphColor::BRANCH_SIZE,        "size");
        opt.branching(GraphColor::BRANCH_DEGREE_SIZE, "degree-size");
        opt.branching(GraphColor::BRANCH_AFC_SIZE,    "afc-size");
        opt.branching(GraphColor::BRANCH_ACTION_SIZE, "action-size");

        opt.symmetry(GraphColor::SYMMETRY_NONE);
        opt.symmetry(GraphColor::SYMMETRY_NONE, "none");
        opt.symmetry(GraphColor::SYMMETRY_LDSB, "ldsb",
                     "use value symmetry breaking");
        /*
        int argc=3;
        char* argv[3];
        argv[0] = new char[88];
        argv[1] = new char[88];
        argv[2] = new char[88];
        std::string s1("./spatialSubdiv");
        std::string s2("-mode");
        std::string s3("gist");

        argv[0] = strcpy(argv[0], s1.c_str());
        argv[1] = strcpy(argv[1], s2.c_str());
        argv[2] = strcpy(argv[2], s3.c_str());
        opt.parse(argc,argv);
        */


        Script::run<GraphColor,BAB,SizeOptions>(opt);


        return 0;
}

} // namespace

// STATISTICS: example-any
