#define GLM_ENABLE_EXPERIMENTAL

#include <iostream>
#include <glm/glm.hpp>
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#define GL_CHECK_ERRORS
#include <utils/glErrorHandler.h>
#include <utils/Fifar.h>
#include <renderer/AssimpDataVBORenderer.h>
#include <options.hpp>
#include <event/GLFWEventHandler.h>
#include <event/Tweakbar.h>
#include <chrono>

#include <grid/grid_builder.h>
#include <cuda/grid_builder_gpu.hpp>

#include <camera.hpp>

#include <cmath>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <vector>

#include <cssolver/primitive-cs.hpp>
#include <cssolver/sub_propagation.hpp>

#include <io/SubDataIO.h>



int WINDOW_WIDTH=1024;
int WINDOW_HEIGHT=1024;

#define GPU_BUILD

#ifdef GPU_BUILD
  #include <cuda_runtime_api.h>
  #include <cuda.h>
  #include <cuda_gl_interop.h>
  #include <utils/cuda_error_handler.h>
#endif

void error_callback(int error, const char* description)
{
        fprintf(stderr, "Error: %s\n", description);
}

/*
static void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
        if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
                glfwSetWindowShouldClose(window, true);
}
*/

void initGL() {
        glewInit();
        // Enable Texturing
        glEnable(GL_TEXTURE_2D);
        // Enable depth test
        glEnable(GL_DEPTH_TEST);
        // Accept fragment if it closer to the camera than the former one
        glDepthFunc(GL_LESS);


/*
        glEnable( GL_BLEND );
        glEnable( GL_COLOR_MATERIAL );
        glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
        glBlendEquation( GL_FUNC_ADD );
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
*/

        printf("OpenGL:\n\tvendor %s\n\trenderer %s\n\tversion %s\n\tshader language %s\n",
               glGetString(GL_VENDOR), glGetString(GL_RENDERER), glGetString(GL_VERSION),
               glGetString(GL_SHADING_LANGUAGE_VERSION));

        // Finding the compute shader extension
        int extCount;
        glGetIntegerv(GL_NUM_EXTENSIONS, &extCount);
        bool found = false;
        for (int i = 0; i < extCount; ++i)
                if (!strcmp((const char*)glGetStringi(GL_EXTENSIONS, i), "GL_ARB_compute_shader")) {
                        printf("Extension \"GL_ARB_compute_shader\" found\n");
                        found = true;
                        break;
                }

        if (!found) {
                fprintf(stderr, "Extension \"GL_ARB_compute_shader\" not found\n");
                exit(14);
        }

}


void drawTexToQuad(int texid) {

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glOrtho(-1.f, 1.f, -1.f, 1.f, -1.0, 1.0);

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();
        glColor3f(1, 1, 1);
        glEnable(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, texid);

        //Draw Quad with texture
        glBegin(GL_QUADS);
        glTexCoord2f(0.f, 0.f);
        glVertex3f(-1, -1, 0);

        glTexCoord2f(1.f, 0.f);
        glVertex3f(1, -1, 0);

        glTexCoord2f(1.f, 1.f);
        glVertex3f(1, 1, 0);

        glTexCoord2f(0.f, 1.f);
        glVertex3f(-1, 1, 0);
        glEnd();

        glPopMatrix();

        glMatrixMode(GL_PROJECTION);
        glPopMatrix();

        glMatrixMode(GL_MODELVIEW);

        glBindTexture(GL_TEXTURE_2D, 0);

}


int main(int args, char* argv[]) {
        std::cout << "Spatial Subdivision to compute active cells for PSH" << std::endl;

        bool show_help = false;

        std::string path_scene;
        std::string path_fifar;
        std::string path_output;
        bool fixedSubLevel;

        int no_of_frames = 100;
        int max_subdiv_depth = 14;

        float fov;

        utils::ArgParser parser(args, argv);
        parser.add_option("help", "h", "Shows this message", show_help, false);
        parser.add_option("windowx", "wx", "Sets the window's resolution in x", WINDOW_WIDTH, 1024, "int");
        parser.add_option("windowy", "wy", "Sets the window's resolution in y", WINDOW_HEIGHT, 1024, "int");
        parser.add_option("fov", "fv", "Camera's fov in degrees", fov, 90.f, "float");

        parser.add_option("scene-file", "s", "Sets the path to a 3d model(obj, 3ds, ...)", path_scene, std::string(""), "path");

        parser.add_option("camera-path", "c", "Sets the path to a camera path", path_fifar, std::string(""), "path");
        parser.add_option("no-frames", "n", "Sets the number of frames to consider", no_of_frames, 100, "int");
        parser.add_option("subdiv-depth", "d", "Sets the maximum subdiv depth", MAX_BASE_LEVEL, 12, "int");

        parser.add_option("output-path", "o", "Sets the output path", path_output, std::string(""), "path");
        parser.add_option("fixedLevel", "f", "Fixed subdivision level", fixedSubLevel, false);



        if (!parser.parse()) {
        	 return EXIT_FAILURE;
        }

        if (show_help) {
        		parser.usage();
        		return EXIT_SUCCESS;
        }


        if (fixedSubLevel) {
            MIN_BASE_LEVEL = MAX_BASE_LEVEL;
        }

        if (!glfwInit())
        {
                // Initialization failed
                std::cerr << "Could not initialize GLFW" << std::endl;
                exit(1);
        }

        glfwSetErrorCallback(error_callback);

        AssimpVBOScene scene;

        GLFWwindow* window = glfwCreateWindow(WINDOW_WIDTH, WINDOW_HEIGHT, "Spatial Subdivision", NULL, NULL);
        if (!window)
        {
                // Window or OpenGL context creation failed
                exit(1);
        }
      //  glfwSetKeyCallback(window, key_callback);
        glfwMakeContextCurrent(window);
        glfwSwapInterval(1);

        initGL();

        if(path_scene.compare("") != 0) {
          scene.load(path_scene);
        } else {
          std::cerr << "Please specify scene file" << std::endl;
          parser.usage();
          return EXIT_FAILURE;
        }

        AssimpDataVBORenderer renderer(scene);

        bool render_video = true;
     //    if(path_fifar.compare("") != 0) {
					// render_video = true;
					// std::cout << "CAMERA PATH: " << path_fifar << std::endl;
					// // if(!camf.load(path_Camera.c_str())) exit(1);
					// fifar.readFromFile(path_fifar.c_str());
					// fifar.computeFrameDescriptors(no_of_frames);
			  // }

        glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
        Camera cam(WINDOW_WIDTH, WINDOW_HEIGHT,  DEG2RADF(fov));
        eventhandler::setupEventHandler(window, cam, WINDOW_WIDTH, WINDOW_HEIGHT);
        cam.update();

        double time_elapsed=0.f;
        float scene_size = 10.f;
        uint frame_count = 0;

        auto step_time_start = std::chrono::high_resolution_clock::now();

        // we currently only need one frame, so limit everything to this
        int numFramesToRender = 1;


        while (!glfwWindowShouldClose(window)) {
                auto start = std::chrono::high_resolution_clock::now();

                if(render_video) {
                							std::cout << "frame: " << frame_count << std::endl;
                							// if(frame_count >= fifar.getNumFrames()) break;
                                            if(frame_count >= numFramesToRender) break;
                							

                       //                      trd::FrameDescriptor fd = fifar.getFrame(frame_count);

                							// // FrameInfo fr = camf.getFrame(frame_count);
                							// std::cout << "" << fd.pos[0] << " " << fd.pos[1] << " " << fd.pos[2] << " " << fd.up[0] << " " << fd.up[1] << " " << fd.up[2] << " "  << fd.dir[0] << " " << fd.dir[1] << " " << fd.dir[2] << " " << fd.numFrames << std::endl;
                							// glm::vec3 eye;
                							// eye.x = fd.pos[0];
                							// eye.y = fd.pos[1];
                							// eye.z = fd.pos[2];
                							// glm::vec3 up;
                							// up.x = fd.up[0];
                							// up.y = fd.up[1];
                							// up.z = fd.up[2];
                							// glm::vec3 dir;
                							// dir.x = fd.dir[0];
                							// dir.y = fd.dir[1];
                							// dir.z = fd.dir[2];
                                            glm::vec3 eye;
                                            eye.x = 0;
                                            eye.y = 0;
                                            eye.z = 0;

                                            glm::vec3 up;
                                            up.x = 0;
                                            up.y = 1;
                                            up.z = 0;

                                            glm::vec3 dir;
                                            dir.x = 0;
                                            dir.y = 0;
                                            dir.z = 1;
                							cam.setCameraWithDir(eye, up, dir);
                							cam.update();
                 } else {
                              eventhandler::processEvents(time_elapsed, scene_size);
                 }


                //std::cout << "Rendering" << std::endl;
                glClearColor(0.0f, 0.0f, 0.0f, 0.0f );
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

                renderer.shaderBegin();

                GLuint atom_dbg_buffer;
                int zero = 0;
                //Atomic Triangle Voxel count
                glGenBuffers(1, &(atom_dbg_buffer));

                glBindBufferBase( GL_ATOMIC_COUNTER_BUFFER, 4, atom_dbg_buffer);
                glBufferData(GL_ATOMIC_COUNTER_BUFFER, sizeof(GLuint), NULL, GL_STATIC_COPY);
                glBufferSubData(GL_ATOMIC_COUNTER_BUFFER, 0, sizeof(GLuint), &zero);

                renderer.setProjectionMatrix(cam.getProjectionMatrix());
                renderer.setViewMatrix(cam.getModelViewMatrix());
                renderer.render();

/*
                glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, atom_dbg_buffer);
                GL_CHECK_ERROR();
                uint* fragcount = (uint *) glMapBufferRange(GL_ATOMIC_COUNTER_BUFFER, 0, sizeof(uint), GL_MAP_READ_BIT);
                GL_CHECK_ERROR();
                std::cout << "Fragment count: " << *fragcount << std::endl;
*/
                renderer.shaderEnd();

                Tweakbar::Instance().draw();
                cam.update();
                glfwSwapBuffers(window);

                glfwPollEvents();
                eventhandler::handleJoystick();
                frame_count++;
                time_elapsed = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - start)).count();
                time_elapsed = time_elapsed <= 0 ? 1 : time_elapsed;
        }


        renderer.shaderBegin();

        glBindBuffer(GL_SHADER_STORAGE_BUFFER, renderer.getGLSubTable());

        GL_CHECK_ERROR();
        uint* sub_table_data = (uint *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER, 0, sizeof(uint) * scene.getNbrOfFaces(), GL_MAP_READ_BIT);
        //float *sub_table_data = (float *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER, 0, sizeof(uint) * mAssimpScene.getNbrOfFaces(), GL_MAP_READ_BIT);
        GL_CHECK_ERROR();

        auto step_time = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - step_time_start)).count();
        std::cout << "Computing and fetching Subs GL took: " << step_time << std::endl;

//        int no_iter = 50;//mAssimpScene.getNbrOfFaces()
//        for(int i=0; i < no_iter; i++) {
//          std::cout << "Table data <" << i << " : " << sub_table_data[i] << ">" << std::endl;
//        }


        //Propagate Subs
        step_time_start = std::chrono::high_resolution_clock::now();
        cs::propagate(sub_table_data, scene);
        step_time = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - step_time_start)).count();
        std::cout << "Propagating Subs took: " << step_time << std::endl;
// #define CPU_BUILD
#ifdef CPU_BUILD
        step_time_start = std::chrono::high_resolution_clock::now();
        uint build_res = pow(2, MAX_BASE_LEVEL);
        std::vector<Cell> cells;
        build_grid(cells, sub_table_data, scene.getRawTriangles(), build_res, scene.getUnitUniformBBox());
        step_time = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - step_time_start)).count();
        std::cout << "Build resolution is:" << build_res << "^3" << std::endl;
        std::cout << "Marking grid cells took: " << step_time << std::endl;


        if(path_output.compare("") != 0) {
          step_time_start = std::chrono::high_resolution_clock::now();
          std::cout << "####Writing data to disk <"<< path_output<<">####" << std::endl;
          std::cout << "No of cells <" << cells.size() << ">" << std::endl;
          io::writeSubData(path_output, sub_table_data, scene.getNbrOfFaces(), cells, WINDOW_WIDTH, WINDOW_HEIGHT, MAX_BASE_LEVEL, scene.getScenesBBox(), scene.getUniformBox());
          step_time = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - step_time_start)).count();
          std::cout << "Writing to disk took: " << step_time << std::endl;
        }
#else   //GPU_BUILD
    std::cout << "Building GPU" << std::endl;
    uint64_t no_cell_blocks=0;

    std::cout << "MAX_BASE_LEVEL: " << MAX_BASE_LEVEL << std::endl;
    // for(int i=MIN_BASE_LEVEL; i <= MAX_BASE_LEVEL; ++i) {
       size_t per_level = pow(2,MAX_BASE_LEVEL);
    //   //32bit = 4*4*2
      no_cell_blocks = (per_level / 4) * (per_level / 4) * (per_level / 2);
    // }

    

    unsigned int* d_cell_data = 0;
    //Create Cell Data on device for all possible level
    cudaMalloc((void**) &d_cell_data, no_cell_blocks * sizeof(unsigned int));
    CUDA_LAST_ERROR();
    cudaMemset(d_cell_data, 0, no_cell_blocks * sizeof(unsigned int));
    CUDA_LAST_ERROR();

    std::cout << "no_cell_block: " << no_cell_blocks << std::endl;
    std::cout << "cells total: " << ((uint64_t) no_cell_blocks) * 32 << std::endl;

  
    //Copy triangle data to device
    //TODO: We already did that e.g. as VBO
    gutil::Tri *d_tris = 0;
    cudaMalloc((void**) &d_tris, sizeof(gutil::Tri) * scene.getRawTriangles().size());
    cudaMemcpy(d_tris, scene.getRawTriangles().data(),  scene.getRawTriangles().size() * sizeof(gutil::Tri), cudaMemcpyHostToDevice) ;
    CUDA_LAST_ERROR();

    //Copy SUBS to device
    //TODO: We already have that from GLSL. However, changing this should not do much
    // unsigned int *d_subs = 0;
    // cudaMalloc((void**) &d_subs, sizeof(unsigned int) * scene.getNbrOfFaces());
    // cudaMemcpy(d_subs, sub_table_data, scene.getNbrOfFaces() * sizeof(unsigned int), cudaMemcpyHostToDevice) ;
    // CUDA_LAST_ERROR();

    step_time_start = std::chrono::high_resolution_clock::now();

    //Build grid
    buildGridGPU(d_cell_data, d_tris, scene.getRawTriangles().size(), scene.getUnitUniformBBox(), MIN_BASE_LEVEL);

    step_time = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - step_time_start)).count();
    std::cout << "buildGridGPU took: " << step_time << std::endl;

    //Marked cells
    auto active_cells = popCountGPU(d_cell_data, no_cell_blocks);
    std::cout << "GPU Popcount active cells: "  << active_cells << std::endl;

    //Copy Cell Data back from device to HOST
    unsigned int *cell_data = (unsigned int*) malloc(no_cell_blocks * sizeof(unsigned int));
    cudaMemcpy(cell_data, d_cell_data,  no_cell_blocks * sizeof(unsigned int), cudaMemcpyDeviceToHost);

    //Output marked cells to disk
    if(path_output.compare("") != 0) {
      step_time_start = std::chrono::high_resolution_clock::now();
      std::cout << "####Writing data to disk <"<< path_output<<">####" << std::endl;
      io::writeHashDataGPU(path_output, sub_table_data, scene.getNbrOfFaces(), cell_data, no_cell_blocks, WINDOW_WIDTH, WINDOW_HEIGHT, MIN_BASE_LEVEL, MAX_BASE_LEVEL, scene.getScenesBBox(), scene.getUniformBox());
      step_time = ((std::chrono::duration<double, std::milli>) (std::chrono::high_resolution_clock::now() - step_time_start)).count();
      std::cout << "Writing to disk took: " << step_time << std::endl;
    }

    cudaFree(d_cell_data);
    cudaFree(d_tris);
    // cudaFree(d_subs);
#endif

        glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);

        renderer.shaderEnd();


        // std::cout << "####Debug Renderer####" << std::endl;
        // renderer.setActiveShader(1);

        // frame_count = 0;

        // while (!glfwWindowShouldClose(window))
        // {

        //   if(eventhandler::event_data.play && render_video) {
        //     std::cout << "frame: " << frame_count << std::endl;
        //     if(frame_count >= fifar.getNumFrames()) frame_count = 0;
        //     trd::FrameDescriptor fd = fifar.getFrame(frame_count);

        //     // FrameInfo fr = camf.getFrame(frame_count);
        //     std::cout << "" << fd.pos[0] << " " << fd.pos[1] << " " << fd.pos[2] << " " << fd.up[0] << " " << fd.up[1] << " " << fd.up[2] << " "  << fd.dir[0] << " " << fd.dir[1] << " " << fd.dir[2] << " " << fd.numFrames << std::endl;
        //     glm::vec3 eye;
        //     eye.x = fd.pos[0];
        //     eye.y = fd.pos[1];
        //     eye.z = fd.pos[2];
        //     glm::vec3 up;
        //     up.x = fd.up[0];
        //     up.y = fd.up[1];
        //     up.z = fd.up[2];
        //     glm::vec3 dir;
        //     dir.x = fd.dir[0];
        //     dir.y = fd.dir[1];
        //     dir.z = fd.dir[2];
        //     cam.setCameraWithDir(eye, up, dir);
        //     cam.update();

        //   }

        //   eventhandler::processEvents(time_elapsed, scene_size);


        //   //std::cout << "Rendering" << std::endl;
        //   glClearColor(0.0f, 0.0f, 0.0f, 0.0f );
        //   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        //   renderer.shaderBegin();

        //   renderer.setProjectionMatrix(cam.getProjectionMatrix());
        //   renderer.setViewMatrix(cam.getModelViewMatrix());
        //   renderer.render();

        //   renderer.shaderEnd();

        //   Tweakbar::Instance().draw();
        //   cam.update();
        //   glfwSwapBuffers(window);

        //   glfwPollEvents();
        //   eventhandler::handleJoystick();

        //   frame_count++;

        // }
        // //cs::run(scene.getPrimNeighRel());

        // glfwDestroyWindow(window);
        // glfwTerminate();

        return EXIT_SUCCESS;
}
